let project = new Project('TooLow');

project.addSources('src');
project.addAssets('res/**/!(*.mp3)', {
	nameBaseDir: 'res',
	destination: '{dir}/{name}',
	name: '{dir}/{name}'
});
project.addAssets('assets/_audio/(*.mp3|*.ogg|*.wav)', {
	nameBaseDir: 'assets/_audio/',
	destination: '{dir}/{name}',
	name: '{dir}/{name}',
	quality: 0.99
});
// project.addAssets('res/kha/!(shaders)', {
// 	nameBaseDir: 'res',
// 	destination: '{name}',
// 	name: '{name}'
// });

// project.addParameter('--debug');
// project.addDefine('kha_no_ogg');

// project.addParameter("--times");
// project.addDefine("macro-times");
project.addDefine('analyzer-optimize');
project.addParameter('-dce full');
project.targetOptions.html5.disableContextMenu = true;
project.addDefine('kha_html5_disable_automatic_size_adjust');
project.addDefine('message-reporting=pretty');
project.addDefine('source-map-content');
// project.windowOptions.width = 100;
// project.windowOptions.height = 100;

project.addLibrary("deepnightLibs");
project.addLibrary("ldtk-haxe-api");

const khmPath = project.addLibrary("khm");
project.addDefine(`khm_path=${khmPath}`);
project.addShaders(`${khmPath}/shaders`);

/** @module fs */
const fs = require("fs");

const res = fs.readFileSync(`${__dirname}/build-resources.hxml`).toString();
const lines = res.split("\n");
lines.filter(s => s.startsWith("--resource")).forEach(s => {
	s = s.replace("--resource ", `--resource ${__dirname}/`);
	console.log(s);
	project.addParameter(s);
});
function parseResDefine(name) {
	let define = lines.find(line => line.includes(`${name}=`));
	if (define == undefined) {
		console.log(`Warning: -D ${name} not found`);
		return;
	}
	define = define.replace("-D ", "");
	define = define.replace(`${name}=`, `${name}=${__dirname}/`);
	console.log(define);
	project.addDefine(define);
}
parseResDefine("res_atlases_folder");
parseResDefine("res_audio_json");
parseResDefine("assets_folder");
parseResDefine("src_folder");

function getWatchPort() {
	const i = process.argv.findIndex(v => v.startsWith("--port"));
	if (i == -1) return 8080;
	const port = parseInt(process.argv[i + 1]);
	if (isNaN(port)) return 8080;
	return port
}
const watchPort = getWatchPort()
if (process.argv.includes("--watch")) {
	const observerPort = watchPort + 2;
	project.addDefine(`khm_observer_port=${observerPort}`);
	const path = require('path');
	const libPath = path.resolve(`${khmPath}/tools`);
	const Observer = require(`${libPath}/observer.js`).Observer;
	const projectDir = path.resolve('.');
	const buildDir = `${projectDir}/build/${platform}`;
	const observer = new Observer(projectDir, buildDir, 'kha.js', observerPort);
}

resolve(project);
