package khm;

import khm.Event.EventKind;
import khm.IGameObject.GameObjectType;
import khm.Text.TextAlign;
import khm.Types.Point;
import khm.internal.BaseObjectUtils;
import khm.math.Degrees;
import khm.math.FastMatrix3;
import khm.utils.DebugRectOverlay;
#if phaser
import phaser.gameobjects.Graphics as PGraphics;
#end

#if kha
using kha.graphics2.GraphicsExtension;
#end

enum GraphicCommand {
	// Begin(isClear:Bool, clearColor:Color);
	Clear(clearColor:Color);
	SetColor(color:Color);
	DrawRect(x:Float, y:Float, w:Float, h:Float, lineW:Float);
	FillRect(x:Float, y:Float, w:Float, h:Float);
	DrawCircle(x:Float, y:Float, radius:Float, lineW:Float);
	FillCircle(x:Float, y:Float, radius:Float);
	DrawString(text:String, x:Float, y:Float);
	// End;
}

class Graphics implements IGameObject {
	#if kha
	public var phaser:Dynamic = {};
	#elseif phaser
	public var phaser:PGraphics;
	#end

	public var isInteractive(default, set) = true;
	public var transformation(get, default) = FastMatrix3.identity();

	@:isVar public var width(get, set) = 0.0;
	@:isVar public var height(get, set) = 0.0;

	inline function get_width():Float {
		return this.width;
	}

	inline function set_width(width:Float):Float {
		return this.width = width;
	}

	inline function get_height():Float {
		return this.height;
	}

	inline function set_height(height:Float):Float {
		return this.height = height;
	}

	public var x(default, set) = 0.0;
	public var y(default, set) = 0.0;
	public var parent:IContainer = null;
	public var alpha(default, set) = 1.0;
	public var angle(default, set):Degrees = 0.0;
	public var visible(default, set) = true;
	public var scaleX(default, set) = 1.0;
	public var scaleY(default, set) = 1.0;
	public var originX(default, set) = 0.5;
	public var originY(default, set) = 0.5;
	public var tint:Color = 0xFFFFFFFF;
	public var state(default, never):ObjectState;
	public var type:GameObjectType = GraphicsType;

	public var font:String;
	public var fontSize = 16;

	public function new():Void {
		#if kha
		#elseif phaser
		phaser = new PGraphics(App._scene);
		#end
	}

	function set_isInteractive(v) {
		isInteractive = v;
		return v;
	}

	function set_x(v)
		return phaser.x = this.x = v;

	function set_y(v)
		return phaser.y = this.y = v;

	function set_alpha(v)
		return phaser.alpha = this.alpha = v;

	function set_angle(v) {
		this.angle = v;
		return phaser.angle = this.angle.toFloat();
	}

	function set_visible(v)
		return phaser.visible = this.visible = v;

	function set_scaleX(v)
		return phaser.scaleX = this.scaleX = v;

	function set_scaleY(v)
		return phaser.scaleY = this.scaleY = v;

	public var scale(get, set):Float;

	@:keep function get_scale():Float {
		return (scaleX + scaleY) / 2;
	}

	@:keep function set_scale(scale:Float):Float {
		scaleX = scale;
		scaleY = scale;
		return scale;
	}

	public var displayWidth(get, set):Float;
	public var displayHeight(get, set):Float;

	inline function get_displayWidth():Float
		return width * scaleX;

	inline function get_displayHeight():Float
		return height * scaleY;

	inline function set_displayWidth(dw:Float):Float {
		scaleX = dw / width;
		return dw;
	}

	inline function set_displayHeight(dh:Float):Float {
		scaleY = dh / height;
		return dh;
	}

	function set_originX(v) {
		originX = v;
		// phaser.setOrigin(originX, originY);
		return v;
	}

	function set_originY(v) {
		originY = v;
		// phaser.setOrigin(originX, originY);
		return v;
	}

	#if kha
	function get_transformation() {
		return transformation;
	}
	#elseif phaser
	var tempMatrix = new phaser.gameobjects.components.TransformMatrix();

	function get_transformation() {
		final m = phaser.getWorldTransformMatrix(tempMatrix);
		// trace(m);
		transformation._00 = m.a;
		transformation._01 = m.b;
		transformation._10 = m.c;
		transformation._11 = m.d;
		transformation._20 = m.e;
		transformation._21 = m.f;

		// final graphics = phaser;
		// graphics.x = x;
		// graphics.y = y;
		// graphics.scaleX = scaleX;
		// graphics.scaleY = scaleY;
		// graphics.rotation = angle.toRad();
		return transformation;
	}
	#end

	final commands:Array<GraphicCommand> = [];

	public function begin(isClear = true, ?clearColor:Color) {
		#if phaser
		if (isClear) {
			clear(clearColor);
			fillRect(0, 0, width, height);
		}
		#end
		for (text in tempTexts.reversed()) {
			text.parent.remove(text);
		}
		// tempTexts.resize(0);
		tempTextId = 0;
		// #end
		commands.resize(0);
		// commands.push(Begin(isClear, clearColor));
	}

	public function clear(?clearColor:Color):Void {
		#if phaser
		phaser.clear();
		if (clearColor != null) {
			var old = color;
			color = clearColor;
			fillRect(0, 0, width, height);
			color = old;
		}
		#end
		// commands.push(Clear(clearColor));
		commands.resize(0);
	}

	public var opacity(get, set):Float;

	function get_opacity()
		return alpha;

	function set_opacity(value)
		return alpha = value;

	public var color(default, set):Color = 0xFFFFFFFF;

	function set_color(color:Color):Color {
		#if phaser
		// alpha = color.A;
		phaser.alpha = color.A;
		final color3 = color.toRgb();
		phaser.fillStyle(color3, color.A);
		phaser.lineStyle(1, color3, color.A);
		this.color = color;
		#end
		commands.push(SetColor(color));
		return color;
	}

	public function drawRect(x:Float, y:Float, w:Float, h:Float, lineW = 1.0) {
		#if phaser
		phaser.lineStyle(lineW, color.toRgb(), color.A);
		phaser.strokeRect(x, y, w, h);
		#end
		commands.push(DrawRect(x, y, w, h, lineW));
	}

	public function drawCircle(x:Float, y:Float, radius:Float, lineW = 1.0) {
		#if phaser
		phaser.lineStyle(lineW, color.toRgb(), color.A);
		phaser.strokeCircle(x, y, radius);
		#end
		commands.push(DrawCircle(x, y, radius, lineW));
	}

	public function fillCircle(x:Float, y:Float, radius:Float) {
		#if phaser
		phaser.fillCircle(x, y, radius);
		#end
		commands.push(FillCircle(x, y, radius));
	}

	public function fillRect(x:Float, y:Float, w:Float, h:Float) {
		#if phaser
		phaser.fillRect(x, y, w, h);
		#end
		commands.push(FillRect(x, y, w, h));
	}

	public function end() {
		for (i in tempTextId...tempTexts.length) {
			final text = tempTexts[i];
			text.alpha = 0;
			if (text.text.length != 0) text.text = "";
		}
		for (i in 0...tempTextId) {
			final text = tempTexts[i];
			text.alpha = 1;
		}
		// commands.push(End);
	}

	final tempTexts:Array<Text> = [];
	var tempTextId = 0;

	public function drawString(text:String, x:Float, y:Float):Text {
		// #if phaser
		var obj = tempTexts[tempTextId];
		tempTextId++;
		if (obj == null) {
			final scene = this.findScene() ?? throw 'target ${this.log()} is not attached to scene';
			obj = scene.addText({
				text: text,
				style: {
					fontFamily: font,
					fontSize: fontSize
				}
			});
			obj.setStroke(Black, 3);
			tempTexts.push(obj);
			DebugRectOverlay.ignoreObject(obj);
		} else {
			obj.setText(text);
		}
		obj.originX = 0;
		obj.originY = 0;
		obj.parent.add(obj);
		obj.x = x;
		obj.y = y;
		// #end

		// commands.push(DrawString(text, x, y));

		return obj;
	}

	#if kha
	public function render(g:kha.graphics2.Graphics) {
		if (!visible || alpha == 0) return;
		g.color = 0xFFFFFFFF;
		pushContainerSpace(g);

		for (command in commands) {
			switch command {
				// case Begin(isClear, clearColor):
				// 	g.begin(isClear, clearColor);
				case Clear(clearColor):
					// g.begin(clearColor);
					// g.clear(clearColor);
				case SetColor(color):
					g.color = color;
				case DrawRect(x, y, w, h, lineW):
					// final local = topLeftLocal();
					// trace(local);
					// g.drawRect(local.x + x, local.y + y, w, h, lineW);
					g.drawRect(x, y, w, h, lineW);
				case FillRect(x, y, w, h):
					g.fillRect(x, y, w, h);
				case DrawCircle(x, y, radius, lineW):
					g.drawCircle(x, y, radius, lineW);
				case FillCircle(x, y, radius):
					g.fillCircle(x, y, radius);
				case DrawString(text, x, y):
					g.drawString(text, x, y);
					// case End:
			}
		}
		// commands.resize(0);

		popContainerSpace(g);
	}

	@:noCompletion var prevAlpha:Float;

	// @:noCompletion var prevScaleQuality:ImageScaleQuality;

	function pushContainerSpace(g:kha.graphics2.Graphics):Void {
		prevAlpha = g.opacity;
		// prevScaleQuality = g.imageScaleQuality;
		g.opacity = g.opacity * alpha;
		// g.imageScaleQuality = scaleQuality;
		updateTransformation();
		g.pushTransformation(transformation);
	}

	function popContainerSpace(g:kha.graphics2.Graphics):Void {
		// g.imageScaleQuality = prevScaleQuality;
		g.opacity = prevAlpha;
		g.popTransformation();
	}

	public function updateTransformation():Void {
		transformation.setFrom(FastMatrix3.identity());
		transformation.setFrom(transformation.multmat(parent.transformation));
		BaseObjectUtils.apply2DTransformation(transformation, this);
	}
	#end
}
