package khm;

import haxe.ds.ReadOnlyArray;
import khm.IGameObject;

interface IContainer extends IGameObject {
	/** Array of children game objects. **/
	final children:Children;
	/** Add a child object at the end of the children list. **/
	function add(obj:IGameObject):Void;
	/** Insert a child object at the specified position of the children list. **/
	function addAt(index:Int, obj:IGameObject):Void;
	/**
		Remove the given object from the children list of the object if it's part of it.
		Object will not be rendered, but not destroyed.
	**/
	function remove(obj:IGameObject):Void;

	/**
		Remove all objects from the children list of the object.
		Object will not be rendered, but not destroyed.
	**/
	function removeAll():Void;
}

@:forward(length, unshift, insert, copy, filter, indexOf, iterator, keyValueIterator,
	join, lastIndexOf, map, slice, contains, toString, splice, resize)
abstract Children(Array<IGameObject>) from Array<IGameObject> to Iterable<IGameObject> {
	public inline function new() {
		this = [];
	}

	@:arrayAccess inline function get(i:Int)
		return this[i];

	@:arrayAccess inline function set(i:Int, obj:IGameObject)
		return this[i] = obj;

	public function push(obj:IGameObject):Int {
		return this.push(obj);
	}

	public function remove(obj:IGameObject):Bool {
		return this.remove(obj);
	}

	public function last():IGameObject {
		return ArrayTools.last(this);
	}

	public inline function sum(getValue:(obj:IGameObject) -> Float):Float {
		return ArrayTools.sum(this, getValue);
	}

	/** Returs all children under this container recursively as flat array. **/
	public function getAll():Array<IGameObject> {
		final allChildren:Array<IGameObject> = [];
		for (obj in this) {
			allChildren.push(obj);
			if (!(obj is IContainer)) continue;
			_getAll(cast obj, allChildren);
		}
		return allChildren;
	}

	function _getAll(obj:IContainer, allChildren:Array<IGameObject>):Void {
		for (obj in obj.children) {
			allChildren.push(obj);
			if (!(obj is IContainer)) continue;
			_getAll(cast obj, allChildren);
		}
	}

	public function flatIter(callback:(obj:IGameObject) -> Void):Void {
		final allChildren = getAll();
		for (obj in allChildren) callback(obj);
	}

	// public function insert(pos:Int, obj:IGameObject):Void {
	// 	return this.insert(pos, obj);
	// }
	// public function unshift(obj:IGameObject):Void {
	// 	this.unshift(obj);
	// }
	// function concat(a:Array<T>):Array<T>;
	// function join(sep:String):String;
	// function pop():Null<IGameObject> {
	// 	final item = this.pop();
	// 	item?.parent?.remove(item);
	// 	return item;
	// }
	// function reverse():Void;
	// function shift():Null<T>;
	// function slice(pos:Int, ?end:Int):Array<T>;
	// function sort(f:T->T->Int):Void;
	// function splice(pos:Int, len:Int):Array<T>;
	// function toString():String;
	// function unshift(x:T):Void;
	// inline function insert(pos:Int, x:T):Void {
	// 	(cast this).splice(pos, 0, x);
	// }
	// inline function contains(x:T):Bool {
	// 	#if (js_es >= 6)
	// 	return (cast this).includes(x);
	// 	#else
	// 	return this.indexOf(x) != -1;
	// 	#end
	// }
	// inline function indexOf(x:T, ?fromIndex:Int):Int {
	// 	return @:privateAccess
	// 		HxOverrides.indexOf(this, x, (fromIndex != null) ? fromIndex : 0);
	// }
	// inline function lastIndexOf(x:T, ?fromIndex:Int):Int {
	// 	return @:privateAccess
	// 		HxOverrides.lastIndexOf(this, x, (fromIndex != null) ? fromIndex : length - 1);
	// }
	// @:pure
	// inline function copy():Array<T> {
	// 	return (cast this).slice();
	// }
	// @:runtime inline function map<S>(f:T->S):Array<S> {
	// 	var result:Array<S> = js.Syntax.construct(Array, length);
	// 	for (i in 0...length) {
	// 		result[i] = f(this[i]);
	// 	}
	// 	return result;
	// }
	// @:runtime inline function filter(f:T->Bool):Array<T> {
	// 	return [for (v in this) if (f(v)) v];
	// }
	// @:runtime inline function iterator():haxe.iterators.ArrayIterator<T> {
	// 	return new haxe.iterators.ArrayIterator(this);
	// }
	// @:runtime inline function keyValueIterator():ArrayKeyValueIterator<T> {
	// 	return new ArrayKeyValueIterator(this);
	// }
	// inline function resize(len:Int):Void {
	// 	this.length = len;
	// }
}
