package khm;

import khm.math.Degrees;
import khm.math.FastMatrix3;

enum abstract GameObjectType(Int) {
	var SceneType;
	var ContainerType;
	var SpriteType;
	var TweenType;
	var TextType;
	var GraphicsType;
	var ParticlesType;
	var WidgetType;

	/** Returns `true` for `ContainerType`, `SceneType` and `WidgetType`. **/
	public function isContainerLike():Bool {
		return switch abstract {
			case SceneType, ContainerType, WidgetType: true;
			case _: false;
		}
	}
}

@:keepSub
@:using(khm.tools.GameObjectTools)
@:using(khm.tools.GameObjectAnimTools)
interface IGameObject {
	/**
		Internal type of this game object.

		For debugging and casting from `IGameObject` interface to other types.
	**/
	var type:GameObjectType;
	var parent:IContainer;
	var transformation(get, null):FastMatrix3;
	var isInteractive(default, set):Bool;
	var width(get, set):Float;
	var height(get, set):Float;
	/** In most cases, this value is a shortcut to `obj.width * obj.scaleX`. **/
	var displayWidth(get, set):Float;
	/** In most cases, this value is a shortcut to `obj.height * obj.scaleY`. **/
	var displayHeight(get, set):Float;
	var x(default, set):Float;
	var y(default, set):Float;
	var alpha(default, set):Float;
	var angle(default, set):Degrees;
	var visible(default, set):Bool;
	/**
		In most cases, returns `(scaleX + scaleY) / 2` on reading,
		and changes both `scaleX` and `scaleY` to same value on writing.
	**/
	var scale(get, set):Float;
	var scaleX(default, set):Float;
	var scaleY(default, set):Float;
	var originX(default, set):Float;
	var originY(default, set):Float;
	/**
		State is object with your data fields and `debug` object inside for internal debug values.

		You can set `obj.state.myField = 1` and access it after with `final myField:Int = obj.state.myField`.

		`obj.state.someField` always returns `Any` type, so you need to cast it before usage.

		If you want to only keep single value per game object in state,
		you can use `state.value` as default field for it.
	**/
	var state(default, never):ObjectState;
	#if kha
	/** This function is called by `obj.parent` render to draw current object in active local space. **/
	function render(g:kha.graphics2.Graphics):Void;
	#end
}
