package khm;

class NineSlice extends Container {
	static final dirs = [
		{x: -1, y: -1}, {x: 0, y: -1}, {x: 1, y: -1},
		{x: -1, y: 0}, {x: 0, y: 0}, {x: 1, y: 0},
		{x: -1, y: 1}, {x: 0, y: 1}, {x: 1, y: 1},
	];

	public final slices:Array<Sprite>;
	public final centerSprite:Sprite;
	public final topLeftSprite:Sprite;
	public final topSprite:Sprite;
	public final topRightSprite:Sprite;
	public final leftSprite:Sprite;
	public final rightSprite:Sprite;
	public final bottomLeftSprite:Sprite;
	public final bottomSprite:Sprite;
	public final bottomRightSprite:Sprite;
	public final leftW:Int;
	public final topH:Int;

	public var tint(get, set):Color;

	function get_tint():Color {
		return centerSprite.tint;
	}

	function set_tint(tint:Color):Color {
		for (sprite in slices) sprite.tint = tint;
		return tint;
	}

	public function new(frame:ResItem, ?leftW:Int, ?topH:Int, ?pos:DebugPosInfos) {
		super();
		this.setPosInfos(pos);

		centerSprite = addSprite(frame);
		final w = Std.int(centerSprite.width);
		final h = Std.int(centerSprite.height);
		leftW ??= Std.int(Math.min(w, h) / 2) - 1;
		topH ??= leftW;
		centerSprite.setCrop(leftW, topH, w - leftW * 2, h - topH * 2);

		final topH = topH == 0 ? h : topH;

		topLeftSprite = addSprite(frame);
		topLeftSprite.setCrop(0, 0, leftW, topH);
		topLeftSprite.setOrigin(1, 1);

		topSprite = addSprite(frame);
		topSprite.setCrop(leftW, 0, w - leftW * 2, topH);
		topSprite.setOrigin(0.5, 1);

		topRightSprite = addSprite(frame);
		topRightSprite.setCrop(w - leftW, 0, leftW, topH);
		topRightSprite.setOrigin(0, 1);

		leftSprite = addSprite(frame);
		leftSprite.setCrop(0, topH, leftW, h - topH * 2);
		leftSprite.setOrigin(1, 0.5);

		rightSprite = addSprite(frame);
		rightSprite.setCrop(w - leftW, topH, leftW, h - topH * 2);
		rightSprite.setOrigin(0, 0.5);

		bottomLeftSprite = addSprite(frame);
		bottomLeftSprite.setCrop(0, h - topH, leftW, topH);
		bottomLeftSprite.setOrigin(1, 0);

		bottomSprite = addSprite(frame);
		bottomSprite.setCrop(leftW, h - topH, w - leftW * 2, topH);
		bottomSprite.setOrigin(0.5, 0);

		bottomRightSprite = addSprite(frame);
		bottomRightSprite.setCrop(w - leftW, h - topH, leftW, topH);
		bottomRightSprite.setOrigin(0, 0);

		this.leftW = leftW;
		this.topH = topH;

		slices = [
			topLeftSprite,
			topSprite,
			topRightSprite,
			leftSprite,
			centerSprite,
			rightSprite,
			bottomLeftSprite,
			bottomSprite,
			bottomRightSprite,
		];

		shrink();
	}

	override function set_originX(v:Float):Float {
		super.set_originX(v);
		updateX();
		return v;
	}

	override function set_originY(v:Float):Float {
		super.set_originY(v);
		updateY();
		return v;
	}

	override function set_width(width:Float):Float {
		if (width < leftW * 2) width = leftW * 2;
		this.width = width;
		updateX();
		return width;
	}

	override function set_height(height:Float):Float {
		if (height < topH * 2) height = topH * 2;
		this.height = height;
		updateY();
		return height;
	}

	function updateX():Void {
		centerSprite.displayWidth = Std.int(width - leftW * 2);
		final cw = centerSprite.displayWidth;

		for (i => sprite in slices) {
			final vx = dirs[i].x;
			sprite.x = vx * cw / 2;
			sprite.x += (0.5 - originX) * width;
		}
		topSprite.displayWidth = width - leftW * 2;
		bottomSprite.displayWidth = width - leftW * 2;
	}

	function updateY():Void {
		centerSprite.displayHeight = Std.int(height - topH * 2);
		final ch = centerSprite.displayHeight;

		for (i => sprite in slices) {
			final vy = dirs[i].y;
			sprite.y = vy * ch / 2;
			sprite.y += (0.5 - originY) * height;
		}
		leftSprite.displayHeight = height - topH * 2;
		rightSprite.displayHeight = height - topH * 2;
	}
}
