package khm;

import khm.DebugPosInfos;
import khm.Graphics;
import khm.Lazyr.PosObjectConfig;

private typedef DebugState = {
	var ?posInfos:DebugPosInfos;
	var ?posContainerLock:Bool;
	var ?posContainerData:PosObjectConfig;
	var ?containerBoundsPreview:Graphics;

	var ?debugRect:Container;
	/** Offset after `shrink`, see `draggableDebug` **/
	var ?useParentOffset:Bool;
	var ?offX:Float;
	var ?offY:Float;
	/** For sprite debug rect codegen **/
	var ?debugChangedScale:Bool;
	var ?debugChangedOrigin:Bool;

	var ?ignoreDebugRect:Bool;
	var ?ignoreDebugPanel:Bool;
	var ?disabledInteractive:Bool;
	/** Phaser warning flag in `sprite.setCrop` **/
	var ?trimCropWarned:Bool;
}

private typedef DefaultObjectState = {
	debug:DebugState,
	_listeners:Map<String, Array<(value:Any) -> Void>>
}

abstract ObjectState(DefaultObjectState) from DefaultObjectState {
	public function new() {
		this = defaultState();
	}

	public inline function init() {
		this ??= defaultState();
	}

	inline function defaultState():DefaultObjectState {
		return {debug: {}, _listeners: []};
	}

	public var value(get, set):Any;

	inline function get_value():Any {
		return get("value");
	}

	inline function set_value(v):Any {
		set("value", v);
		return value;
	}

	public var debug(get, never):DebugState;

	inline function get_debug():DebugState {
		this ??= defaultState();
		this.debug ??= {};
		return this.debug;
	}

	@:op(a.b) inline public function get(key:String):Any {
		this ??= defaultState();
		return Reflect.getProperty(this, key);
	}

	@:op(a.b) inline public function set(key:String, value:Any):Void {
		this ??= defaultState();
		Reflect.setProperty(this, key, value);
		if (this._listeners[key] != null) {
			for (callback in this._listeners[key]) callback(value);
		}
	}

	public function on(key:String, afterSet:(value:Any) -> Void):Void {
		this._listeners[key] ??= [];
		this._listeners[key].push(afterSet);
	}

	/** `Reflect.fields` without `debug` and `_listeners.` **/
	public function getUserFields():Array<String> {
		final fields = Reflect.fields(this);
		fields.remove("debug");
		fields.remove("_listeners");
		return fields;
	}

	public function clear():Void {
		this._listeners?.clear();
		for (field in Reflect.fields(this)) {
			Reflect.setProperty(this, field, null);
		}
	}
}
