package khm;

import haxe.extern.EitherType as Or;
import khm.Event.EventKind;
import khm.IContainer.Children;
import khm.IGameObject.GameObjectType;
import khm.ResItem;
import khm.Types.MinMax;
import khm.Types.Point;
import khm.Types.Range;
import khm.math.Degrees;
import khm.math.FastMatrix3;
import khm.utils.FrameUtils;

using khm.tools.PhaserTools;

#if phaser
import phaser.gameobjects.particles.Particle;
import phaser.gameobjects.particles.ParticleEmitter;
#end

private typedef MinMaxOrFloat = Or<Or<Float, MinMax>, Array<Float>>;

// typedef Particle = phaser.gameobjects.particles.Particle;
// https://newdocs.phaser.io/docs/3.55.2/Phaser.GameObjects.Particles.EmitterOp
private typedef EmitterOp = {
	#if phaser
	var ?onEmit:(particle:Particle) -> Float;
	var ?onUpdate:(particle:Particle) -> Float;
	#end
	var ?start:Float;
	var ?end:Float;
	var ?min:Float;
	var ?max:Float;
}

// private typedef ObjectOrFloat = Or<Or<Float, EmitterOp>, MinMax>;
private typedef ObjectOrFloat = Or<Float, EmitterOp>;

typedef ParticlesConfig = {
	/** `[active]` - Sets {@link ParticleEmitter#active}. **/
	var ?active:Bool;
	/** `[blendMode]` - Sets {@link ParticleEmitter#blendMode}. **/
	var ?blendMode:Dynamic;
	/** `[callbackScope]` - Sets {@link ParticleEmitter#deathCallbackScope} and {@link ParticleEmitter#emitCallbackScope}. **/
	// var ?callbackScope:Dynamic;
	/** `[collideBottom]` - Sets {@link ParticleEmitter#collideBottom}. **/
	// var ?collideBottom:Bool;
	/** `[collideLeft]` - Sets {@link ParticleEmitter#collideLeft}. **/
	// var ?collideLeft:Bool;
	/** `[collideRight]` - Sets {@link ParticleEmitter#collideRight}. **/
	// var ?collideRight:Bool;
	/** `[collideTop]` - Sets {@link ParticleEmitter#collideTop}. **/
	// var ?collideTop:Bool;
	/** `[deathCallback]` - Sets {@link ParticleEmitter#deathCallback}. **/
	var ?deathCallback:(particle:Particle) -> Void;
	/** `[deathCallbackScope]` - Sets {@link ParticleEmitter#deathCallbackScope}. **/
	// var ?deathCallbackScope:Dynamic;
	/** `[emitCallback]` - Sets {@link ParticleEmitter#emitCallback}. **/
	var ?emitCallback:(particle:Particle) -> Void;
	/** `[emitCallbackScope]` - Sets {@link ParticleEmitter#emitCallbackScope}. **/
	// var ?emitCallbackScope:Dynamic;
	/** `[follow]` - Sets {@link ParticleEmitter#follow}. **/
	var ?follow:IGameObject;
	/** `[frequency]` - Sets {@link ParticleEmitter#frequency}. **/
	var ?frequency:Float;
	/** `[gravityX]` - Sets {@link ParticleEmitter#gravityX}. **/
	var ?gravityX:Float;
	/** `[gravityY]` - Sets {@link ParticleEmitter#gravityY}. **/
	var ?gravityY:Float;
	/** `[maxParticles]` - Sets {@link ParticleEmitter#maxParticles}. **/
	var ?maxParticles:Float;
	/** `[name]` - Sets {@link ParticleEmitter#name}. **/
	// var ?name:String;
	/** `[isActive]` - Sets {@link ParticleEmitter#isActive}. **/
	// var ?isActive:Bool;
	/** `[particleBringToTop]` - Sets {@link ParticleEmitter#particleBringToTop}. **/
	// var ?particleBringToTop:Bool;
	/** `[particleClass]` - Sets {@link ParticleEmitter#particleClass}. **/
	// var ?particleClass:phaser.gameobjects.particles.Particle;
	/** `[radial]` - Sets {@link ParticleEmitter#radial}. **/
	// var ?radial:Bool;
	/** `[timeScale]` - Sets {@link ParticleEmitter#timeScale}. **/
	// var ?timeScale:Float;
	/** `[trackVisible]` - Sets {@link ParticleEmitter#trackVisible}. **/
	// var ?trackVisible:Bool;
	/** `[visible]` - Sets {@link ParticleEmitter#visible}. **/
	var ?visible:Bool;
	/** `[accelerationX]` - Sets {@link ParticleEmitter#accelerationX} (emit only). **/
	var ?accelerationX:MinMaxOrFloat;
	/** `[accelerationY]` - Sets {@link ParticleEmitter#accelerationY} (emit only). **/
	var ?accelerationY:MinMaxOrFloat;
	/** `[alpha]` - Sets {@link ParticleEmitter#alpha}. **/
	var ?alpha:ObjectOrFloat;
	/** `[angle]` - Sets {@link ParticleEmitter#angle} (emit only). **/
	var ?angle:MinMaxOrFloat;
	/** `[bounce]` - Sets {@link ParticleEmitter#bounce} (emit only). **/
	var ?bounce:MinMaxOrFloat;
	/** `[delay]` - Sets {@link ParticleEmitter#delay} (emit only). **/
	var ?delay:MinMaxOrFloat;
	/** `[lifespan]` - Sets {@link ParticleEmitter#lifespan} (emit only). **/
	var ?lifespan:MinMaxOrFloat;
	/** `[maxVelocityX]` - Sets {@link ParticleEmitter#maxVelocityX} (emit only). **/
	var ?maxVelocityX:MinMaxOrFloat;
	/** `[maxVelocityY]` - Sets {@link ParticleEmitter#maxVelocityY} (emit only). **/
	var ?maxVelocityY:MinMaxOrFloat;
	/** `[moveToX]` - Sets {@link ParticleEmitter#moveToX} (emit only). **/
	var ?moveToX:MinMaxOrFloat;
	/** `[moveToY]` - Sets {@link ParticleEmitter#moveToY} (emit only). **/
	var ?moveToY:MinMaxOrFloat;
	/** `[quantity]` - Sets {@link ParticleEmitter#quantity} (emit only). **/
	var ?quantity:MinMaxOrFloat;
	/** `[rotate]` - Sets {@link ParticleEmitter#rotate}. **/
	var ?rotate:ObjectOrFloat;
	/** `[scale]` - As {@link ParticleEmitter#setScale}. **/
	var ?scale:ObjectOrFloat;
	/** `[scaleX]` - Sets {@link ParticleEmitter#scaleX}. **/
	var ?scaleX:ObjectOrFloat;
	/** `[scaleY]` - Sets {@link ParticleEmitter#scaleY}. **/
	var ?scaleY:ObjectOrFloat;
	/** `[speed]` - As {@link ParticleEmitter#setSpeed} (emit only). **/
	var ?speed:MinMaxOrFloat;
	/** `[speedX]` - Sets {@link ParticleEmitter#speedX} (emit only). **/
	var ?speedX:MinMaxOrFloat;
	/** `[speedY]` - Sets {@link ParticleEmitter#speedY} (emit only). **/
	var ?speedY:MinMaxOrFloat;
	/** `[tint]` - Sets {@link ParticleEmitter#tint}. **/
	var ?tint:ObjectOrFloat;
	/** `[x]` - Sets {@link ParticleEmitter#x} (emit only). **/
	var ?x:MinMaxOrFloat;
	/** `[y]` - Sets {@link ParticleEmitter#y} (emit only). **/
	var ?y:MinMaxOrFloat;
	/** `[emitZone]` - As {@link ParticleEmitter#setEmitZone}. **/
	// var ?emitZone:Dynamic;
	/** `[deathZone]` - As {@link ParticleEmitter#setDeathZone}. **/
	// var ?deathZone:phaser.types.gameobjects.particles.ParticleEmitterDeathZoneConfig;
	/** `[bounds]` - As {@link ParticleEmitter#setBounds}. **/
	// var ?bounds:Dynamic;
	/** `[followOffset]` - Assigns to {@link ParticleEmitter#followOffset}. **/
	// var ?followOffset:Dynamic;
	/** `[frame]` - Sets {@link ParticleEmitter#frames}. **/
	var ?frames:Array<ResItem>;
	var ?texture:String;
	@:noCompletion
	var ?frame:Or<String, {
		var texture:String;
		/** One or more texture frames. **/
		var frames:Array<String>;
		/** Whether texture frames will be assigned consecutively (true) or at random (false). **/
		var ?cycle:Bool;
		/** The number of consecutive particles receiving each texture frame, when cycle is true. **/
		var ?quantity:Int;
	}>;
	/** `[reserve]` - Creates specified number of inactive particles and adds them to this emitter's pool. {@link ParticleEmitter#reserve}. **/
	// var ?reserve:Float;
}

#if kha
class Particle extends Sprite {
	var life = 1000.0;
	var lifeCurrent = 1000.0;
	var delayCurrent = 0.0;
	var holdCurrent = 0.0;
	var lifeT = 0.0;
	var velocityX = 0.0;
	var velocityY = 0.0;
	var emitter:Particles;
	var accelerationX = 0.0;
	var accelerationY = 0.0;
	var maxVelocityX = 0.0;
	var maxVelocityY = 0.0;
	var bounce = 0.0;

	public function new(emitter:Particles) {
		this.emitter = emitter;
		super(cast {frame: ""});
	}

	public function fire(spawnX:Float, spawnY:Float):Bool {
		// var anim = emitter.getAnim();
		var anim = null;
		if (anim != null) {
			// this.anims.play(anim);
		} else {
			setFrame(emitter.getFrame());
		}
		// if (!this.frame) throw new Error('Particle has no texture frame');

		// Updates particle.x and particle.y during this call
		// emitter.getEmitZone(this);

		// if (x == null) this.x += ops.x.onEmit(this, 'x');
		// else if (ops.x.steps > 0) {
		// 	// EmitterOp is stepped but x was forced (follower?) so use it
		// 	this.x += x + ops.x.onEmit(this, 'x');
		// } else this.x += x;
		// if (y == null) this.y += ops.y.onEmit(this, 'y');
		// else if (ops.y.steps > 0) {
		// 	// EmitterOp is stepped but y was forced (follower?) so use it
		// 	this.y += y + ops.y.onEmit(this, 'y');
		// } else this.y += y;

		final props = emitter.props;
		x = spawnX + props.x.randomInt();
		y = spawnY + props.y.randomInt();
		life = props.lifespan.randomInt();
		lifeCurrent = life;
		lifeT = 0;
		delayCurrent = props.delay.randomInt();
		holdCurrent = props.hold.randomInt();
		scaleX = props.scaleX.lerp(0);
		scaleY = props.scaleY.lerp(0);
		angle = props.rotate.lerp(0);

		// emitter.worldMatrix.transformPoint(this.x, this.y, this.worldPosition);

		// Check we didn't spawn in the middle of a DeathZone
		// if (this.delayCurrent == 0 && emitter.getDeathZone(this)) {
		// 	this.lifeCurrent = 0;
		// 	return false;
		// }

		final sx = props.speedX.randomInt();
		final sy = props.speedY.randomInt();

		if (emitter.radial) {
			final rad = props.angle.randomInt().toRad();
			velocityX = rad.cos() * Math.abs(sx);
			velocityY = rad.sin() * Math.abs(sy);
		} else if (emitter.moveTo) {
			// final mx = ops.moveToX.onEmit(this, 'moveToX');
			// final my = ops.moveToY.onEmit(this, 'moveToY');
			// final lifeS = this.life / 1000;
			// velocityX = (mx - x) / lifeS;
			// velocityY = (my - y) / lifeS;
		} else {
			velocityX = sx;
			velocityY = sy;
		}

		// if (ops.color.active) {
		// 	this.tint = ops.color.onEmit(this, 'tint');
		// } else {
		// 	this.tint = ops.tint.onEmit(this, 'tint');
		// }
		accelerationX = props.accelerationX.randomInt();
		accelerationY = props.accelerationY.randomInt();
		maxVelocityX = props.maxVelocityX.randomInt();
		maxVelocityY = props.maxVelocityY.randomInt();
		bounce = props.bounce.randomInt();
		alpha = props.alpha.lerp(0);

		return true;
	}

	// update returns `true` if the particle is now dead (lifeCurrent <= 0)
	public function update(delta:Float, step:Float):Bool {
		if (lifeCurrent <= 0) {
			// Particle is dead via `Particle.kill` method, or being held
			if (holdCurrent > 0) {
				holdCurrent -= delta;
				return (holdCurrent <= 0);
			} else {
				return true;
			}
		}

		if (this.delayCurrent > 0) {
			this.delayCurrent -= delta;
			return false;
		}

		// todo
		// this.anims.update(0, delta);

		// How far along in life is this particle? (t = 0 to 1)
		var t = 1 - (lifeCurrent / life);
		lifeT = t;

		// x = emitter.x.lerp(t);
		// y = emitter.y.lerp(t);

		// todo
		// if (emitter.moveTo) {
		// 	var mx = ops.moveToX.onUpdate(this, 'moveToX', t, emitter.moveToX);
		// 	var my = ops.moveToY.onUpdate(this, 'moveToY', t, emitter.moveToY);
		// 	var lifeS = this.lifeCurrent / 1000;
		// 	this.velocityX = (mx - this.x) / lifeS;
		// 	this.velocityY = (my - this.y) / lifeS;
		// }

		computeVelocity(emitter, delta, step, t);

		final props = emitter.props;
		scaleX = props.scaleX.lerp(t);
		scaleY = props.scaleY.lerp(t);
		angle = props.rotate.lerp(t);

		// if (emitter.getDeathZone(this)) {
		// 	this.lifeCurrent = 0;
		// 	// No need to go any further, particle has been killed
		// 	return true;
		// }

		alpha = props.alpha.lerp(t);
		alpha = alpha.clamp(0, 1);

		// if (ops.color.active) {
		// 	this.tint = ops.color.onUpdate(this, 'color', t, this.tint);
		// } else {
		// 	this.tint = ops.tint.onUpdate(this, 'tint', t, this.tint);
		// }

		this.lifeCurrent -= delta;

		return (this.lifeCurrent <= 0 && this.holdCurrent <= 0);
	}

	function computeVelocity(emitter:Particles, delta:Float, step:Float, t:Float) {
		var vx = velocityX;
		var vy = velocityY;
		final props = emitter.props;
		final ax = props.accelerationX.lerp(t);
		final ay = props.accelerationY.lerp(t);
		final mx = props.maxVelocityX.lerp(t);
		final my = props.maxVelocityY.lerp(t);
		bounce = props.bounce.lerp(t);

		vx += (props.gravityX * step) + (ax * step);
		vy += (props.gravityY * step) + (ay * step);
		vx = vx.clamp(-mx, mx);
		vy = vy.clamp(-my, my);
		velocityX = vx;
		velocityY = vy;

		// Integrate back in to the position
		x += vx * step;
		y += vy * step;

		// emitter.worldMatrix.transformPoint(this.x, this.y, this.worldPosition);
	}
}

@:structInit
private class EmitterProps {
	public var quantity:MinMax;
	public var gravityX:Float;
	public var gravityY:Float;
	// range is start-end, minmax is random value
	public var accelerationX:MinMax;
	public var accelerationY:MinMax;
	public var alpha:Range;
	public var angle:MinMax;
	public var bounce:MinMax;
	public var delay:MinMax;
	/** Frozen or 'held in place' after it has finished its lifespan for a set number of ms. **/
	public var hold:MinMax;
	public var lifespan:MinMax;
	public var maxVelocityX:MinMax;
	public var maxVelocityY:MinMax;
	public var moveToX:MinMax;
	public var moveToY:MinMax;
	public var rotate:Range;
	public var scaleX:Range;
	public var scaleY:Range;
	public var speedX:MinMax;
	public var speedY:MinMax;
	public var tint:MinMax;
	public var x:MinMax;
	public var y:MinMax;
}

class Particles extends Container {
	public var isActive = true;
	// public var alive:Array<Particle> = [];
	public var dead:Array<Particle> = [];

	public var frequency = 0.0;

	var flowCounter = 0.0;
	var maxParticles = 0;
	var particleBringToTop = true;
	var emitCallback:(particle:Particle) -> Void = null;
	var deathCallback:(particle:Particle) -> Void = null;

	public var radial = true;
	public var moveTo = false;

	public final props:EmitterProps;

	var frames:Array<ResItem>;

	public function new(config:ParticlesConfig) {
		super();
		frequency = config.frequency;
		frames = config.frames;
		emitCallback = config.emitCallback;
		deathCallback = config.deathCallback;

		final scale = toRange(config.scale, {start: 1, end: 1});
		final speed = config.speed ?? {min: 0, max: 0};
		radial = config.speedX == null && config.speedY == null;
		props = {
			quantity: toMinMax(config.quantity, 1),
			gravityX: config.gravityX ?? 0.0,
			gravityY: config.gravityY ?? 0.0,
			accelerationX: toMinMax(config.accelerationX, 0),
			accelerationY: toMinMax(config.accelerationY, 0),
			alpha: toRange(config.alpha, 1),
			angle: toMinMax(config.angle, {
				min: 0,
				max: 360}),
			bounce: toMinMax(config.bounce, 0),
			// color: new EmitterColorOp('color');
			delay: toMinMax(config.delay, 0, true),
			hold: toMinMax(0, 0, true),
			lifespan: toMinMax(config.lifespan, 1000, true),
			maxVelocityX: toMinMax(config.maxVelocityX, 10000),
			maxVelocityY: toMinMax(config.maxVelocityY, 10000),
			moveToX: toMinMax(config.moveToX, 0),
			moveToY: toMinMax(config.moveToY, 0),
			rotate: toRange(config.rotate, {
				start: 0,
				end: 0}),
			scaleX: toRange(config.scaleX, cast scale),
			scaleY: toRange(config.scaleY, cast scale),
			speedX: toMinMax(config.speedX, speed, true),
			speedY: toMinMax(config.speedY, speed, true),
			tint: toMinMax(config.tint, 0xffffff),
			x: toMinMax(config.x, 0),
			y: toMinMax(config.y, 0),
		}

		Event.on((this : Any), BeforeRender, update);
	}

	function toMinMax(v:MinMaxOrFloat, defValue:MinMaxOrFloat, ?emitOnly:Bool):MinMax {
		if (v == null) {
			if (defValue == null) throw "both values are null";
			return toMinMax(defValue, null);
		}
		if (v is Float) return {min: v, max: v};
		// todo not correct, should be arr of values for random
		if (v is Array) return {
			min: v[0],
			max: v[1]
		}
		final v:EmitterOp = cast v;
		if (v.start != null) return {
			min: v.start,
			max: v.end,
		};
		if (v.min != null) return {
			min: v.min,
			max: v.max,
		};
		trace(v);
		throw "value is not Range or MinMax object";
	}

	function toRange(v:ObjectOrFloat, defValue:ObjectOrFloat, ?emitOnly:Bool):Range {
		if (v == null) {
			if (defValue == null) throw "both values are null";
			return toRange(defValue, null);
		}
		if (v is Float) return {start: v, end: v};
		final v:EmitterOp = cast v;
		if (v.start != null) return {
			start: v.start,
			end: v.end,
		};
		if (v.min != null) return {
			start: v.min,
			end: v.max,
		};
		trace(v);
		throw "value is not Range or MinMax object";
	}

	public function setPosition(x, y):Void {}

	public function flow(frequency:Float, count:Int):Void {
		this.frequency = frequency;
		props.quantity = {min: count, max: count};
	}

	public function explode(count:Int, x:Float, y:Float):Void {
		props.quantity = {min: count, max: count};
		emitParticles(count, x, y);
	}

	function update(e:Event):Void {
		preUpdate(App.deltaTime);
	}

	function preUpdate(delta:Float):Void {
		// Scale the delta
		// delta *= timeScale;

		var step = (delta / 1000);

		// if (trackVisible) visible = follow.visible;
		// Any particle processors?
		// var processors = manager.getProcessors();

		final rip:Array<{index:Int, particle:Particle}> = [];

		for (i in 0...children.length) {
			var particle:Particle = cast children[i];

			// update returns `true` if the particle is now dead (lifeCurrent <= 0)
			if (particle.update(delta, step)) { // processors
				rip.push({index: i, particle: particle});
			}
		}

		// Move dead particles to the dead array
		if (rip.length > 0) {
			var i = rip.length;
			while (i-- > 0) {
				final entry = rip[i];
				// Remove from particles array
				// children.splice(entry.index, 1);
				remove(entry.particle);
				// Add to dead array
				dead.push(entry.particle);
				// Callback
				if (deathCallback != null) deathCallback(entry.particle);
				entry.particle.x = 0;
				entry.particle.y = 0;
			}
		}

		if (!isActive) return;

		if (frequency == 0) {
			emitParticles();
		} else if (frequency > 0) {
			// if (delta > 25 || delta < 10) trace(delta, 1000 / 60);
			flowCounter -= Math.min(delta, 1000 / 60);

			if (flowCounter <= 0) {
				emitParticles();
				// counter = frequency - remained from previous delta
				flowCounter = (frequency - Math.abs(flowCounter));
			}
		}
	}

	function emitParticles(?count:Int, x = 0.0, y = 0.0):Array<Particle> {
		if (atLimit()) return null;

		// if (count == null) count = quantity.onEmit();
		count ??= props.quantity.randomInt();

		final follow:Point = null;
		final followOffset:Point = null;
		var followX = (follow != null) ? follow.x + followOffset.x : x;
		var followY = (follow != null) ? follow.y + followOffset.y : y;

		final created:Array<Particle> = [];
		for (i in 0...count) {
			final particle = dead.pop() ?? new Particle(this);
			created.push(particle);

			particle.fire(followX, followY);

			// todo move particles to global cords
			if (particleBringToTop) add(particle);
			else addAt(0, particle);

			if (emitCallback != null) emitCallback(particle);
			if (atLimit()) break;
		}

		return created;
	}

	function atLimit():Bool {
		return (maxParticles > 0 && getParticleCount() == maxParticles);
	}

	function getParticleCount():Int
		return children.length + dead.length;

	public function stop():Void {
		isActive = false;
		#if phaser
		// todo
		emitter.stop();
		#end
	}

	public function setEmitAngle(v:ObjectOrFloat):Void {
		props.angle = toMinMax(v, 0);
	}

	public function getFrame():ResItem {
		return frames[Std.random(frames.length)];
	}
}
#elseif phaser
class Particles implements IContainer {
	#if phaser3_5
	public var phaser:phaser.gameobjects.particles.ParticleEmitterManager;
	#else
	public var phaser:ParticleEmitter;
	#end
	public var emitter:ParticleEmitter;

	public var isInteractive(default, set):Bool;
	public var transformation(get, default) = FastMatrix3.identity();

	@:isVar public var width(get, set) = 0.0;
	@:isVar public var height(get, set) = 0.0;

	inline function get_width():Float {
		return this.width;
	}

	inline function set_width(width:Float):Float {
		return this.width = width;
	}

	inline function get_height():Float {
		return this.height;
	}

	inline function set_height(height:Float):Float {
		return this.height = height;
	}

	public var x(default, set) = 0.0;
	public var y(default, set) = 0.0;
	public var parent:IContainer = null;
	public var alpha(default, set) = 1.0;
	public var angle(default, set):Degrees = 0.0;
	public var visible(default, set) = true;
	public var scaleX(default, set) = 1.0;
	public var scaleY(default, set) = 1.0;
	public var originX(default, set) = 0.5;
	public var originY(default, set) = 0.5;
	public var tint:Color = 0xFFFFFFFF;
	public var state(default, never):ObjectState;
	public var type:GameObjectType = ParticlesType;
	public var scale(get, set):Float;

	public final children:Children = [];

	public function add(obj:IGameObject):Void {
		if (obj.parent != null) obj.parent.remove(obj);
		obj.parent = this;
		children.push(obj);
	};

	public function addAt(index:Int, obj:IGameObject):Void {
		if (obj.parent != null) obj.parent.remove(obj);
		obj.parent = this;
		children.insert(index, obj);
	};

	public function remove(obj:IGameObject):Void {
		children.remove(obj);
	}

	public function removeAll():Void {
		children.resize(0);
	}

	@:keep function get_scale():Float {
		return (scaleX + scaleY) / 2;
	}

	@:keep function set_scale(scale:Float):Float {
		scaleX = scale;
		scaleY = scale;
		return scale;
	}

	public function new(config:ParticlesConfig):Void {
		final s = config;
		var textureId:Null<String> = s.texture;
		if (s.frames?.length > 0) {
			textureId = s.frames[0].texture;
			for (i => id in s.frames) {
				#if debug
				FrameUtils.checkResItem(id);
				#end
				if (id.texture != textureId) {
					trace("All config.frames should have same texture");
					throw 'config.frames[$i].texture (${id.texture}) is not equal to config.frames[0].texture ($textureId)';
				}
			}
		}
		if (textureId == null) {
			var expl = s.frames == null ? "is null" : "is empty";
			throw 'Texture id not found (config.frames $expl, and config.texture is null)';
		}
		if (s.frames != null && s.frame == null) {
			s.frame = {
				texture: null,
				frames: s.frames.map(id -> id.frame),
				cycle: false,
				quantity: 1
			}
			s.frames = null;
		}
		if (isPhaserOlder3_6()) {
			phaser = App._scene.add.particles(textureId);
			emitter = (phaser : Dynamic).createEmitter(config);
		} else {
			emitter = App._scene.add.particles(0, 0, textureId, cast config);
			// emitter is gameobject in 3.60+
			phaser = cast emitter;
		}
		emitter.stop();
	}

	function isPhaserOlder3_6():Bool {
		final pm = untyped Phaser.GameObjects.Particles.ParticleEmitterManager;
		return pm != null;
	}

	function set_isInteractive(v) {
		isInteractive = v;
		return v;
	}

	function set_x(v) {
		emitter.setPosition(v, y);
		return this.x = v;
	}

	function set_y(v) {
		emitter.setPosition(x, v);
		return this.y = v;
	}

	function set_alpha(v) // return phaser.alpha = this.alpha = v; // todo?
		return this.alpha = v;

	function set_angle(v) {
		this.angle = v;
		return phaser.angle = this.angle.toFloat();
	}

	function set_visible(v)
		return phaser.visible = this.visible = v;

	function set_scaleX(v)
		return phaser.scaleX = this.scaleX = v;

	function set_scaleY(v)
		return phaser.scaleY = this.scaleY = v;

	public var displayWidth(get, set):Float;
	public var displayHeight(get, set):Float;

	inline function get_displayWidth():Float
		return width * scaleX;

	inline function get_displayHeight():Float
		return height * scaleY;

	inline function set_displayWidth(dw:Float):Float {
		scaleX = dw / width;
		return dw;
	}

	inline function set_displayHeight(dh:Float):Float {
		scaleY = dh / height;
		return dh;
	}

	function set_originX(v) {
		originX = v; // todo?
		// phaser.setOrigin(originX, originY);
		return v;
	}

	function set_originY(v) {
		originY = v;
		// phaser.setOrigin(originX, originY);
		return v;
	}

	var tempMatrix = new phaser.gameobjects.components.TransformMatrix();

	function get_transformation() {
		final m = phaser.getWorldTransformMatrix(tempMatrix);
		transformation._00 = m.a;
		transformation._01 = m.b;
		transformation._10 = m.c;
		transformation._11 = m.d;
		transformation._20 = m.e;
		transformation._21 = m.f;
		return transformation;
	}

	public function explode(count:Int, x:Float, y:Float) {
		emitter.explode(count, x, y);
	}

	public function flow(frequency:Float, count:Int) {
		emitter.flow(frequency, count);
	}

	public function stop():Void {
		emitter.stop();
	}

	public function setEmitAngle(v:ObjectOrFloat):Void {
		emitter.setAngle(cast v);
	}
}
#end
