package khm;

import khm.Types.Point;
import khm.Types.Rect;
import khm.math.FastMatrix3;
import khm.utils.Collision;

enum abstract PointerButtonKind(Int) {
	var Left = 0;
	var Right = 1;
	var Middle = 2;
}

class Pointer {
	public var id:Int;
	/** Initial x cord of push. **/
	public var startX = 0;
	/** Initial y cord of push. **/
	public var startY = 0;
	public var x = 0;
	public var y = 0;
	/** Latest pointer x speed. **/
	public var moveX = 0;
	/** Latest pointer y speed. **/
	public var moveY = 0;
	/** Button type (for mouse). **/
	public var button = PointerButtonKind.Left;
	/** Is `button` or finger (`id`) is down. **/
	public var isDown = false;
	/** If pointer is a touch surface. **/
	public var isTouch = false;

	/** If pointer is moved from `startX/Y` position . **/
	public var isMoved(get, never):Bool;

	function get_isMoved() {
		return x != startX || y != startY;
	}

	public function new(id:Int):Void {
		this.id = id;
	}

	public function translate(x:Float, y:Float):Pointer {
		this.x += Std.int(x);
		this.y += Std.int(y);
		return this;
	}

	public inline function localPointTo(t:FastMatrix3, offX = 0.0, offY = 0.0):Point {
		// final m = khm.utils.Utils.matrixAffineInverse(t);
		#if debug
		if (t.scaleX * t.scaleY < 0.00001) {
			throw "cannot get local point of matrix with scale close to zero";
		}
		#else
		if (t.scaleX * t.scaleY < 0.00001) {
			trace("cannot get local point of matrix with scale close to zero");
			return {x: 0, y: 0};
		}
		#end
		final m:FastMatrix3 = t.inverse();
		return {
			x: m.scaleX * (x + offX) + m.translateX,
			y: m.scaleY * (y + offY) + m.translateY
		}
		// return {
		// 	x: m.scaleX * (x + offX) + m.skewX * (y + offY) + m.translateX,
		// 	y: m.scaleY * (y + offY) + m.skewY * (x + offX) + m.translateY
		// }
	}

	public inline function toLocal(obj:IGameObject):Pointer {
		// final m:FastMatrix3 = t.inverse(); // todo report typehint cast
		// return {
		// 	x: m.scaleX * (x + offX) + m.skewX * (y + offY) + m.translateX,
		// 	y: m.scaleY * (y + offY) + m.skewY * (x + offX) + m.translateY
		// }
		final p = new Pointer(id);
		p.startX = startX;
		p.startY = startY;
		final localP = localPointTo(obj.transformation);
		p.x = Std.int(localP.x);
		p.y = Std.int(localP.y);
		p.moveX = Std.int(moveX / obj.transformation.scaleX);
		p.moveY = Std.int(moveY / obj.transformation.scaleY);
		p.button = button;
		p.isDown = isDown;
		p.isTouch = isTouch;
		return p;
	}

	public inline function aabb(rect:Rect):Bool {
		return Collision.aabb({
			x: x,
			y: y,
			w: 1,
			h: 1
		}, rect);
	}
}
