package khm;

import khm.IGameObject.GameObjectType;
import khm.Tween.TweenConfig;
import khm.Types.Point;
import khm.internal.BaseObjectUtils;
import khm.math.Degrees;
import khm.math.FastMatrix3;
#if kha
import kha.Image;
import kha.graphics2.Graphics;
import kha.graphics2.ImageScaleQuality;
#elseif phaser
import phaser.gameobjects.RenderTexture as PRenderTexture;
#end

class RenderTexture implements IGameObject {
	#if kha_html5
	public var phaser:Image;
	#elseif phaser
	public var phaser:PRenderTexture;
	#end

	public var isInteractive(default, set) = true;
	public var transformation(get, default) = FastMatrix3.identity();

	@:isVar public var width(get, set) = 0.0;
	@:isVar public var height(get, set) = 0.0;

	inline function get_width():Float {
		return this.width;
	}

	inline function set_width(width:Float):Float {
		return this.width = width;
	}

	inline function get_height():Float {
		return this.height;
	}

	inline function set_height(height:Float):Float {
		return this.height = height;
	}

	public var x(default, set) = 0.0;
	public var y(default, set) = 0.0;
	public var parent:IContainer = null;
	public var alpha(default, set) = 1.0;
	public var angle(default, set):Degrees = 0.0;
	public var visible(default, set) = true;
	public var scaleX(default, set) = 1.0;
	public var scaleY(default, set) = 1.0;
	public var originX(default, set) = 0.0;
	public var originY(default, set) = 0.0;
	public var tint:Color = 0xFFFFFFFF;
	public var state(default, never):ObjectState;
	public var type:GameObjectType = GraphicsType;
	public var scale(get, set):Float;

	@:keep function get_scale():Float {
		return (scaleX + scaleY) / 2;
	}

	@:keep function set_scale(scale:Float):Float {
		scaleX = scale;
		scaleY = scale;
		return scale;
	}

	public function new(width:Int, height:Int, ?pos:DebugPosInfos):Void {
		state.debug.posInfos = pos;
		this.width = width;
		this.height = height;
		#if kha_html5
		phaser = Image.createRenderTarget(width, height);
		#elseif phaser
		phaser = new PRenderTexture(App._scene, 0, 0, width, height);
		#end
	}

	function set_isInteractive(v) {
		isInteractive = v;
		// if (isInteractive) phaser.setInteractive();
		// else phaser.removeInteractive();
		return v;
	}

	function set_x(v) {
		#if kha
		return this.x = v;
		#elseif phaser
		return phaser.x = this.x = v;
		#end
	}

	function set_y(v) {
		#if kha
		return this.y = v;
		#elseif phaser
		return phaser.y = this.y = v;
		#end
	}

	function set_alpha(v) {
		#if kha
		return this.alpha = v;
		#elseif phaser
		return phaser.alpha = this.alpha = v;
		#end
	}

	function set_angle(v) {
		#if kha
		return this.angle = v;
		#elseif phaser
		this.angle = v;
		return phaser.angle = this.angle.toFloat();
		#end
	}

	function set_visible(v) {
		#if kha
		return this.visible = v;
		#elseif phaser
		return phaser.visible = this.visible = v;
		#end
	}

	function set_scaleX(v) {
		#if kha
		return this.scaleX = v;
		#elseif phaser
		return phaser.scaleX = this.scaleX = v;
		#end
	}

	function set_scaleY(v) {
		#if kha
		return this.scaleY = v;
		#elseif phaser
		return phaser.scaleY = this.scaleY = v;
		#end
	}

	public var displayWidth(get, set):Float;
	public var displayHeight(get, set):Float;

	inline function get_displayWidth():Float
		return width * scaleX;

	inline function get_displayHeight():Float
		return height * scaleY;

	inline function set_displayWidth(dw:Float):Float {
		scaleX = dw / width;
		return dw;
	}

	inline function set_displayHeight(dh:Float):Float {
		scaleY = dh / height;
		return dh;
	}

	function set_originX(v) {
		originX = v;
		#if phaser
		phaser.setOrigin(originX, originY);
		#end
		return v;
	}

	function set_originY(v) {
		originY = v;
		#if phaser
		phaser.setOrigin(originX, originY);
		#end
		return v;
	}

	public function fill(color:Color):Void {
		#if kha
		final g = phaser.g2;
		g.begin(true, 0x0);
		g.color = color;
		g.fillRect(0, 0, width, height);
		g.end();
		#elseif phaser
		phaser.fill(color.toRgb(), color.A);
		#end
	}

	public function draw(sprite:Sprite, x:Float, y:Float):Void {
		#if kha
		final g = phaser.g2;
		g.begin(false, 0x0);
		g.color = sprite.tint;
		g.opacity = sprite.alpha;
		final oldX = sprite.x;
		final oldY = sprite.y;
		final visible = sprite.visible;
		sprite.visible = true;
		sprite.x = x;
		sprite.y = y;
		sprite.render(g);
		sprite.x = oldX;
		sprite.y = oldY;
		sprite.visible = visible;
		g.end();
		#elseif phaser
		phaser.draw(sprite.phaser, x, y);
		#end
	}

	public function clear():Void {
		#if kha
		phaser.g2.begin(true, 0x0);
		phaser.g2.end();
		#elseif phaser
		phaser.clear();
		#end
	}

	#if kha
	function get_transformation() {
		return transformation;
	}
	#elseif phaser
	var tempMatrix = new phaser.gameobjects.components.TransformMatrix();

	function get_transformation() {
		final m = phaser.getWorldTransformMatrix(tempMatrix);
		transformation._00 = m.a;
		transformation._01 = m.b;
		transformation._10 = m.c;
		transformation._11 = m.d;
		transformation._20 = m.e;
		transformation._21 = m.f;
		return transformation;
	}
	#end

	#if kha
	public function render(g:kha.graphics2.Graphics) {
		#if kha_html5
		// if (phaser.texture == null) return;
		if (!visible || alpha == 0) return;
		g.color = 0xFFFFFFFF;
		pushContainerSpace(g);

		final local = this.topLeftLocal();
		g.drawImage(phaser, local.x, local.y);

		popContainerSpace(g);
		#end
	}

	@:noCompletion var prevAlpha:Float;

	// @:noCompletion var prevScaleQuality:ImageScaleQuality;

	function pushContainerSpace(g:Graphics):Void {
		prevAlpha = g.opacity;
		// prevScaleQuality = g.imageScaleQuality;
		g.opacity = g.opacity * alpha;
		// g.imageScaleQuality = scaleQuality;
		updateTransformation();
		g.pushTransformation(transformation);
	}

	function popContainerSpace(g:Graphics):Void {
		// g.imageScaleQuality = prevScaleQuality;
		g.opacity = prevAlpha;
		g.popTransformation();
	}

	public function updateTransformation():Void {
		transformation.setFrom(FastMatrix3.identity());
		transformation.setFrom(transformation.multmat(parent.transformation));
		BaseObjectUtils.apply2DTransformation(transformation, this);
	}
	#end
}
