package khm;

import khm.IGameObject.GameObjectType;
import khm.ResItem;
import khm.Types.IRect;
import khm.internal.BaseObjectUtils;
import khm.math.Degrees;
import khm.math.FastMatrix3;
import khm.utils.FrameUtils;

using khm.tools.MathTools;
using khm.tools.PhaserTools;

#if kha
import kha.Assets;
import kha.Color;
import kha.Image;
import kha.graphics2.Graphics;
import khm.utils.Atlas;
#elseif phaser
import khm.Types.ISize;
import phaser.gameobjects.Sprite as PSprite;
#end

class Sprite implements IGameObject {
	public static var alwaysCrop = false;

	#if kha
	public var phaser:Dynamic = {};
	#elseif phaser
	public var phaser:PSprite;
	#end

	public var type:GameObjectType = SpriteType;
	public var isInteractive(default, set) = true;
	public var id(default, null):ResItem;

	@:isVar public var width(get, set) = 0.0;
	@:isVar public var height(get, set) = 0.0;

	inline function get_width():Float {
		return this.width;
	}

	inline function set_width(width:Float):Float {
		return this.width = width;
	}

	inline function get_height():Float {
		return this.height;
	}

	inline function set_height(height:Float):Float {
		return this.height = height;
	}

	public var x(default, set):Float = 0.0;
	public var y(default, set):Float = 0.0;
	public var parent:IContainer = null;
	public var alpha(default, set) = 1.0;
	public var angle(default, set):Degrees = 0.0;
	public var visible(default, set) = true;
	public var scaleX(default, set) = 1.0;
	public var scaleY(default, set) = 1.0;
	public var originX(default, set) = 0.5;
	public var originY(default, set) = 0.5;
	public var tint(default, set):Color = 0xFFFFFFFF;
	public var state(default, never):ObjectState;

	public var flipX(default, set) = false;
	public var flipY(default, set) = false;

	public var frame(get, never):{name:String};

	#if phaser
	// to update cropped sprite origin point
	var cropOffX = 0.0;
	var cropOffY = 0.0;
	#end

	function get_frame():{name:String} {
		return {
			name: id.frame
		}
	}

	#if kha
	var texture:Image;
	var frameData:AtlasFrameData;
	/** Frame position in atlas after trimming. **/
	var srcX(default, null) = 0;
	/** Frame position in atlas after trimming. **/
	var srcY(default, null) = 0;

	/** The width of the frame in atlas after being trimmed. **/
	var realWidth(default, null) = 0;

	/** The height of the frame in atlas after being trimmed. **/
	var realHeight(default, null) = 0;
	#end

	/** How many pixels was trimmed from left side in original texture. **/
	var trimX = 0;
	/** How many pixels was trimmed from top side in original texture. **/
	var trimY = 0;

	public var transformation(get, default) = FastMatrix3.identity();
	public var scale(get, set):Float;

	@:keep function get_scale():Float {
		return (scaleX + scaleY) / 2;
	}

	@:keep function set_scale(scale:Float):Float {
		#if debug
		if (Math.isNaN(scale)) throw scale;
		#end
		scaleX = scale;
		scaleY = scale;
		return scale;
	}

	public var displayWidth(get, set):Float;
	public var displayHeight(get, set):Float;

	inline function get_displayWidth():Float {
		#if debug
		if (Math.isNaN(scaleX)) throw scaleX;
		#end
		return width * scaleX;
	}

	inline function get_displayHeight():Float
		return height * scaleY;

	inline function set_displayWidth(dw:Float):Float {
		scaleX = dw / width;
		return dw;
	}

	inline function set_displayHeight(dh:Float):Float {
		scaleY = dh / height;
		return dh;
	}

	public function new(id:ResItem):Void {
		#if phaser
		phaser = new PSprite(App._scene, 0, 0, id.texture, id.frame);
		#end
		setFrame(id);
	}

	public function setFrame(id:ResItem):Void {
		this.id = id;
		#if debug
		FrameUtils.checkResItem(id);
		#end
		#if kha
		if (id.texture == null) return;
		texture = Atlas.getImage(id);
		// Atlas.hasTexture(id.texture) &&
		if (!Atlas.hasAtlas(id.texture)) {
			realWidth = texture.width;
			realHeight = texture.height;
			width = texture.width;
			height = texture.height;
			return;
		}
		frameData = Atlas.getAtlasFrame(id);
		trimX = frameData.spriteSourceSize.x;
		trimY = frameData.spriteSourceSize.y;
		srcX = frameData.frame.x;
		srcY = frameData.frame.y;
		realWidth = frameData.frame.w;
		realHeight = frameData.frame.h;
		width = frameData.frame.w;
		height = frameData.frame.h;
		if (alwaysCrop) {
			trimX = 0;
			trimY = 0;
		}
		if (frameData.trimmed && !alwaysCrop) {
			width = frameData.sourceSize.w;
			height = frameData.sourceSize.h;
		}
		#elseif phaser
		final ptexture:phaser.textures.Texture = phaser.texture;
		if (ptexture.key != id.texture && id.texture != null) {
			phaser.setTexture(id.texture, id.frame);
		}
		phaser.setFrame(id.frame);
		width = Std.int(phaser.width);
		height = Std.int(phaser.height);
		cropOffX = 0;
		cropOffY = 0;
		updatePhaserOrigin();
		#end
	}

	// public function setFrameName(frame:String):Void {
	// 	if (id.frame != "") {
	// 		if (!App._scene.textures.get(id.texture).has(id.frame)) {
	// 			throw 'Frame "${id.frame}" not found in texture "${id.texture}"';
	// 		}
	// 	}
	// 	phaser.setFrame(frame);
	// 	width = Std.int(phaser.width);
	// 	height = Std.int(phaser.height);
	// }

	public static function fromTile(id:ResItem, rect:IRect, ?pos:DebugPosInfos):Sprite {
		final sprite = new Sprite(id);
		sprite.setPosInfos(pos);
		sprite.setCrop(rect, true);
		return sprite;
	}

	public inline extern overload function setCrop(x:Float, y:Float, w:Float, h:Float, updateSize = true):Void {
		#if debug
		if (w == 0 || h == 0) trace('setCrop size is ${w}x${h}');
		#end
		#if kha
		final rect:IRect = {
			x: Std.int(x),
			y: Std.int(y),
			w: Std.int(w),
			h: Std.int(h),
		}
		if (rect.x + rect.w > frameData.sourceSize.w)
			throw 'tile width is bigger than frame (${rect.x} + ${rect.w} > ${frameData.sourceSize.w})';
		if (rect.y + rect.h > frameData.sourceSize.h)
			throw 'tile height is bigger than frame (${rect.y} + ${rect.h} > ${frameData.sourceSize.h})';

		if (updateSize) {
			width = rect.w;
			height = rect.h;
		}
		final frameData = frameData;
		final sourceX = frameData.spriteSourceSize.x;
		final sourceY = frameData.spriteSourceSize.y;
		final sourceW = frameData.spriteSourceSize.w;
		final sourceH = frameData.spriteSourceSize.h;
		realWidth = Std.int(Math.min(rect.w, sourceW));
		realHeight = Std.int(Math.min(rect.h, sourceH));

		if (rect.x < sourceX) {
			srcX = frameData.frame.x;
			trimX = sourceX - rect.x;
		} else {
			srcX = frameData.frame.x + rect.x - sourceX;
			trimX = 0;
		}
		if (rect.y < sourceY) {
			srcY = frameData.frame.y;
			trimY = sourceY - rect.y;
		} else {
			srcY = frameData.frame.y + rect.y - sourceY;
			trimY = 0;
		}
		#elseif phaser
		phaser.setCrop(x, y, w, h);

		if (updateSize) {
			final cropCenterX = x + w * originX;
			final cropCenterY = y + h * originY;
			final centerX = trimX + this.x + width * originX;
			final centerY = trimY + this.y + height * originY;
			cropOffX = centerX - cropCenterX;
			cropOffY = centerY - cropCenterY;
			set_originX(originX);
			set_originY(originY);

			width = Std.int(w);
			height = Std.int(h);
		}
		#if debug
		final isWarned = this.state.debug.trimCropWarned;
		if (isWarned) return;
		final sourceSize:ISize = phaser.frame.customData.sourceSize;
		if ((sourceSize.w - phaser.frame.cutWidth) +
			(sourceSize.h - phaser.frame.cutHeight) > 5) {
			trace('Id ${id.texture}.${id.frame} is trimmed by atlas generator, so `setCrop` can be incorrect in phaser');
			trace("Texture trimming in image editor is recommended");
			this.state.debug.trimCropWarned = true;
		}
		#end
		#end
	}

	public inline extern overload function setCrop(rect:IRect, updateSize = true):Void {
		setCrop(rect.x, rect.y, rect.w, rect.h, updateSize);
	}

	public function remove():Void {
		if (parent == null) return;
		parent.remove(this);
		// updateTransformation();
	}

	#if kha
	public function render(g:Graphics) {
		if (!visible || alpha == 0) return;
		if (texture == null) {
			var posText = this.posInfosLine() ?? "";
			if (posText.length != 0) posText = '$posText: ';
			throw '${posText}sprite texture is null';
		}
		g.color = tint;
		final prevAlpha = g.opacity;
		g.opacity = g.opacity * alpha;
		updateTransformation();
		g.pushTransformation(transformation);
		final p = this.topLeftLocal();
		g.drawScaledSubImage(
			texture, srcX, srcY, realWidth, realHeight,
			p.x + trimX,
			p.y + trimY,
			realWidth,
			realHeight
		);
		// g.drawRect(p.x, p.y, realWidth, realHeight);
		g.popTransformation();
		g.opacity = prevAlpha;
	}

	public var isPixelPerfect = false;

	public function updateTransformation():Void {
		transformation.setFrom(FastMatrix3.identity());
		if (parent != null) {
			transformation.setFrom(transformation.multmat(parent.transformation));
		}
		if (isPixelPerfect) {
			BaseObjectUtils.apply2DTransformationPixelPerfect(transformation, this);
		} else {
			BaseObjectUtils.apply2DTransformation(transformation, this);
		}
		if (flipX || flipY) {
			transformation.setFrom(transformation.multmat(FastMatrix3.translation(
				width * (flipX ? (1 - originX * 2) : 0),
				height * (flipY ? (1 - originY * 2) : 0),
			)));
			final sx = flipX ? -1 : 1;
			final sy = flipY ? -1 : 1;
			transformation.setFrom(transformation.multmat(FastMatrix3.scale(sx, sy)));
		}
	}
	#end

	/**
		Returns pixel color of local space sprite x/y cords.

		With `originX = 0.5`, `x` arg should be in range `-width / 2 ... width / 2`.

		With `originX = 0`, `x` should be in range `0 ... width`.

		Same for `y` arg, but with `originY` / `height`.

		This method returns `null` if position is out of range, or if texture is trimmed in atlas.
	**/
	public function getPixel(x:Float, y:Float):Null<Color> {
		// normalize to 0,0 as start frame pos
		final localX = Std.int(x + width * originX) - trimX;
		final localY = Std.int(y + height * originY) - trimY;
		#if kha
		if (localX < 0 || localY < 0) return null;
		if (localX > realWidth || localY > realHeight) return null;
		return texture.at(Std.int(srcX + localX), Std.int(srcY + localY));
		#elseif phaser
		final p = App._scene.textures.getPixel(localX, localY, id.texture, id.frame);
		if (p == null) return null;
		return Color.fromBytes(
			Std.int(p.red),
			Std.int(p.green),
			Std.int(p.blue),
			Std.int(p.alpha)
		);
		#end
	}

	public function copy():Sprite {
		return new Sprite(id);
	}

	function set_isInteractive(v) {
		isInteractive = v;
		return v;
	}

	function set_x(v) {
		#if debug
		if (Math.isNaN(v)) throw v;
		#end
		return phaser.x = this.x = v;
	}

	function set_y(v) {
		#if debug
		if (Math.isNaN(v)) throw v;
		#end
		return phaser.y = this.y = v;
	}

	function set_alpha(v)
		return phaser.alpha = this.alpha = v;

	function set_angle(v) {
		this.angle = v;
		return phaser.angle = this.angle.toFloat();
	}

	function set_visible(v)
		return phaser.visible = this.visible = v;

	function set_scaleX(v)
		return phaser.scaleX = this.scaleX = v;

	function set_scaleY(v)
		return phaser.scaleY = this.scaleY = v;

	function set_originX(v) {
		originX = v;
		#if phaser
		updatePhaserOrigin();
		#end
		return v;
	}

	function set_originY(v) {
		originY = v;
		#if phaser
		updatePhaserOrigin();
		#end
		return v;
	}

	#if phaser
	function updatePhaserOrigin():Void {
		phaser.setOrigin(originX - cropOffX / phaser.width, originY - cropOffY / phaser.height);
		phaser.displayOriginX -= cropOffX;
		phaser.displayOriginY -= cropOffY;
	}
	#end

	function set_tint(color:Color):Color {
		this.tint = color;
		#if phaser
		this.phaser.tint = color.toRgb();
		#end
		return color;
	}

	function set_flipX(v) {
		return phaser.flipX = this.flipX = v;
	}

	function set_flipY(v) {
		return phaser.flipY = this.flipY = v;
	}

	#if kha
	function get_transformation() {
		return transformation;
	}
	#elseif phaser
	var tempMatrix = new phaser.gameobjects.components.TransformMatrix();

	function get_transformation() {
		final m = phaser.getWorldTransformMatrix(tempMatrix);
		transformation._00 = m.a;
		transformation._01 = m.b;
		transformation._10 = m.c;
		transformation._11 = m.d;
		transformation._20 = m.e;
		transformation._21 = m.f;
		return transformation;
	}
	#end
}
