package khm.audio;

#if kha
import kha.arrays.Float32Array;
import kha.audio2.Buffer;
import kha.internal.IntBox;
#end
#if cpp
import sys.thread.Mutex;
#end

#if phaser
class Audio {
	public var debug = false;

	public static function mutexLock():Void {}

	public static function mutexUnlock():Void {}

	public var sampleRate = 0;

	public function new() {}
}
#else
class Audio extends AudioGroup {
	public var debug = false;

	#if cpp
	static var mutex:Mutex;
	static var mutexLocked:Bool = false;
	#end

	public static inline function mutexLock():Void {
		#if cpp
		if (mutexLocked) trace("Audio: unlock mutex before locking");
		mutex.acquire();
		mutexLocked = true;
		#end
	}

	public static inline function mutexUnlock():Void {
		#if cpp
		if (!mutexLocked) trace("Audio: lock mutex before unlocking");
		mutex.release();
		mutexLocked = false;
		#end
	}

	public var sampleRate(get, never):Int;

	var _data:Float32Array;

	@:allow(khm.App)
	function new() {
		super();
		#if cpp
		mutex = new Mutex();
		#end

		_data = new Float32Array(512);

		kha.audio2.Audio.audioCallback = mix;
	}

	function mix(outputBufferLength:IntBox, buffer:Buffer):Void {
		// how many samples we need to fill in our buffer to ouput
		final samples = outputBufferLength.value;

		// we need to fill
		if (_data.length < samples) {
			_data = new Float32Array(samples);
		}

		for (i in 0...samples) _data[i] = 0;

		if (!_mute) {
			process(_data, samples);
		}

		// copy data from _data to output buffer
		for (i in 0...samples) {
			// buffer.data.set(buffer.writeLocation, _data[i].clamp(-1.0, 1.0) * _volume);
			buffer.data.set(buffer.writeLocation, _data[i].clamp(-1.0, 1.0));
			buffer.writeLocation += 1;
			if (buffer.writeLocation >= buffer.size) {
				buffer.writeLocation = 0;
			}
		}
	}

	inline function get_sampleRate():Int {
		return kha.audio2.Audio.samplesPerSecond;
	}
}
#end
