package khm.audio;

#if kha
import kha.arrays.Float32Array;
#else
import haxe.io.Float32Array;
#end
import haxe.ds.Vector;

class AudioChannel {
	public var mute(get, set):Bool;
	public var volume(get, set):Float;
	public var pan(get, set):Float;
	public var output(get, null):AudioGroup;

	public var effects(get, null):Array<AudioEffect>;
	public var effectsCount(get, null):Int;

	var _mute:Bool;
	var _volume:Float;
	var _pan:Float;
	var _output:AudioGroup;

	var _l:Float;
	var _r:Float;

	var _effectsCount:Int;
	var _maxEffects:Int;
	var _dirtyEffects:Bool;

	var _effects:Vector<AudioEffect>;
	var _effectsInternal:Vector<AudioEffect>;

	public function new(maxEffects = 8) {
		_mute = false;
		_volume = 1;
		_pan = 0;

		_l = 0.7071;
		_r = 0.7071;

		_effectsCount = 0;
		_maxEffects = maxEffects;
		_dirtyEffects = true;

		_effects = new Vector(_maxEffects);
		_effectsInternal = new Vector(_maxEffects);
	}

	public function addEffect(effect:AudioEffect) {
		Audio.mutexLock();

		if (_effectsCount >= _maxEffects) {
			trace("cant add effect, max effects: " + _maxEffects);
			return;
		}

		if (effect._parent != null) {
			trace("audio effect already in another channel");
			return;
		}

		effect._parent = this;
		_effects[_effectsCount++] = effect;
		_dirtyEffects = true;

		Audio.mutexUnlock();
	}

	public function removeEffect(effect:AudioEffect) {
		Audio.mutexLock();

		for (i in 0..._effectsCount) {
			if (_effects[i] == effect) { // todo: remove rest from _effectsCount and effect
				_effects[i] = _effects[--_effectsCount];
				break;
			}
		}
		_dirtyEffects = true;

		Audio.mutexUnlock();
	}

	public function removeAllEffects() {
		Audio.mutexLock();

		_effectsCount = 0;
		_dirtyEffects = true;

		Audio.mutexUnlock();
	}

	@:noCompletion public function process(data:Float32Array, samples:Int):Void {}

	inline function processEffects(data:Float32Array, samples:Int):Void {
		Audio.mutexLock();

		if (_dirtyEffects) {
			for (i in 0..._effectsCount) {
				_effectsInternal[i] = _effects[i];
			}
			_dirtyEffects = false;
		}
		final count = _effectsCount;

		Audio.mutexUnlock();

		for (i in 0...count) {
			final effect = _effectsInternal[i];
			if (!effect._mute) {
				effect.process(samples, data, App.audio.sampleRate);
			}
		}
	}

	function get_mute():Bool {
		Audio.mutexLock();
		final v = _mute;
		Audio.mutexUnlock();

		return v;
	}

	function set_mute(v:Bool):Bool {
		Audio.mutexLock();
		_mute = v;
		Audio.mutexUnlock();

		return v;
	}

	function get_volume():Float {
		Audio.mutexLock();
		final v = _volume;
		Audio.mutexUnlock();

		return v;
	}

	function set_volume(v:Float):Float {
		Audio.mutexLock();
		_volume = v.clamp(0, 1);
		v = _volume;
		Audio.mutexUnlock();

		return v;
	}

	function get_pan():Float {
		Audio.mutexLock();
		final v = _pan;
		Audio.mutexUnlock();

		return v;
	}

	function set_pan(v:Float):Float {
		Audio.mutexLock();

		_pan = v.clamp(-1, 1);
		_l = Math.cos((_pan + 1) * Math.PI / 4);
		_r = Math.sin((_pan + 1) * Math.PI / 4);
		v = _pan;

		Audio.mutexUnlock();

		return v;
	}

	function get_output():AudioGroup {
		Audio.mutexLock();
		final v = _output;
		Audio.mutexUnlock();

		return v;
	}

	function get_effects():Array<AudioEffect> {
		Audio.mutexLock();
		final arr = [];
		for (i in 0..._effectsCount) {
			arr.push(_effects[i]);
		}
		Audio.mutexUnlock();

		return arr;
	}

	function get_effectsCount():Int {
		Audio.mutexLock();
		final count = _effectsCount;
		Audio.mutexUnlock();

		return count;
	}
}
