package khm.audio;

#if kha
import kha.arrays.Float32Array;
#else
import haxe.io.Float32Array;
#end
import haxe.ds.Vector;

class AudioGroup extends AudioChannel {
	public var channels(get, null):Array<AudioChannel>;
	public var channelsCount(get, null):Int;

	var _channels:Vector<AudioChannel>;
	var _channelsCount:Int;
	var _dirtyChannels:Bool;

	var _maxChannels:Int;
	var _cache:Float32Array;

	var _channelsInternal:Vector<AudioChannel>;

	public function new(maxChannels:Int = 32) {
		super();

		_maxChannels = maxChannels;
		_channelsCount = 0;
		_cache = new Float32Array(512);
		_dirtyChannels = true;

		_channels = new Vector(_maxChannels);
		_channelsInternal = new Vector(_maxChannels);
	}

	public function add(channel:AudioChannel):Void {
		Audio.mutexLock();

		if (channel == this) {
			Audio.mutexUnlock();
			trace("can't add channel to itself");
			return;
		}

		if (channel._output == this) {
			Audio.mutexUnlock();
			trace("channel already added to group");
			return;
		}

		if (_channelsCount >= _maxChannels) {
			Audio.mutexUnlock();
			trace('can\'t add channel, max channels: ${_maxChannels}');
			return;
		}

		if (channel._output != null) {
			channel._output.remove(channel);
		}

		channel._output = this;
		_channels[_channelsCount++] = channel;
		_dirtyChannels = true;

		Audio.mutexUnlock();
	}

	public function remove(channel:AudioChannel):Void {
		Audio.mutexLock();

		if (channel._output == this) {
			channel._output = null;
			var found = false;
			for (i in 0..._channelsCount) {
				if (_channels[i] == channel) { // TODO: remove rest from _channelsInternal and channels
					_channelsCount--;
					_channels[i] = _channels[_channelsCount];
					_channels[_channelsCount] = null;
					found = true;
					break;
				}
			}
			if (!found) {
				trace("can't remove channel ?");
			}
			_dirtyChannels = true;
		} else {
			trace("can't remove channel, it not belong to this group");
		}

		Audio.mutexUnlock();
	}

	public function empty():Void {
		Audio.mutexLock();

		for (i in 0..._maxChannels) {
			_channels[i]._output = null;
			_channels[i] = null;
		}
		_channelsCount = 0;
		_dirtyChannels = true;

		Audio.mutexUnlock();
	}

	override function process(data:Float32Array, bufferSamples:Int) {
		if (_cache.length < bufferSamples) {
			Audio.mutexLock();
			if (App.audio.debug)
				trace('Allocation request in audio thread. cache: ${_cache.length}, samples: $bufferSamples');
			Audio.mutexUnlock();
			_cache = new Float32Array(bufferSamples);
		}

		for (i in 0...bufferSamples) _cache[i] = 0;

		Audio.mutexLock();

		if (_dirtyChannels) {
			var j = 0;
			while (j < _channelsCount) {
				_channelsInternal[j] = _channels[j];
				j++;
			}
			while (j < _maxChannels && _channelsInternal[j] != null) {
				_channelsInternal[j] = null;
				j++;
			}
			_dirtyChannels = false;
		}

		var count = _channelsCount;

		Audio.mutexUnlock();

		for (i in 0...count) {
			if (!_channelsInternal[i]._mute) {
				_channelsInternal[i].process(_cache, bufferSamples);
			}
		}

		processEffects(_cache, bufferSamples);

		var i = 0;
		while (i < bufferSamples) {
			data[i] += _cache[i] * _volume * _l;
			data[i + 1] += _cache[i + 1] * _volume * _r;
			i += 2;
		}
	}

	function get_channels():Array<AudioChannel> {
		Audio.mutexLock();
		final arr:Array<AudioChannel> = [];
		for (i in 0..._channelsCount) arr.push(_channels[i]);
		Audio.mutexUnlock();
		return arr;
	}

	function get_channelsCount():Int {
		Audio.mutexLock();
		final count = _channelsCount;
		Audio.mutexUnlock();
		return count;
	}
}
