package khm.audio.dsp;

class CombFilter {
	public var feedback:Float;
	public var damping(default, set):Float;
	public var length(get, never):Int;

	var buffer:FloatRingBuffer;
	var filter:Float;
	var dampingInv:Float;

	public function new(length:Int, feedback:Float = 0, damping:Float = 0) {
		filter = 0;
		dampingInv = 0;
		this.feedback = feedback;
		this.damping = damping;
		buffer = new FloatRingBuffer(length);
	}

	public function process(input:Float):Float {
		var output = buffer.read();
		filter = output * dampingInv + filter * damping;
		buffer.insert(input + filter * feedback);

		return output;
	}

	function set_damping(v:Float):Float {
		damping = v.clamp(0, 1);
		dampingInv = 1.0 - damping;

		return damping;
	}

	inline function get_length():Int {
		return buffer.length;
	}
}
