package khm.audio.dsp;

#if kha
import kha.arrays.Float32Array;
#else
import haxe.io.Float32Array;
#end

class FloatRingBuffer {
	public var length(get, never):Int;

	var buffer:Float32Array;
	var pos:Int;

	public function new(length:Int) {
		pos = 0;
		buffer = new Float32Array(length);

		for (i in 0...length) buffer[i] = 0; // cpp target not 0 ?
	}

	public inline function insert(v:Float) {
		buffer[pos] = v;
		shift(1);
	}

	public inline function get(i:Int):Float {
		return buffer[modLength(getIndex(i))];
	}

	public inline function set(i:Int, v:Float) {
		buffer[modLength(getIndex(i))] = v;
	}

	public inline function read():Float {
		return buffer[pos];
	}

	public inline function shift(n:Int) {
		pos = modLength(pos + n);
	}

	public function clear() {
		for (i in 0...buffer.length) {
			buffer[i] = 0;
		}

		pos = 0;
	}

	inline function modLength(i:Int):Int {
		return MathLib.mod(i, buffer.length);
	}

	inline function getIndex(i:Int):Int {
		return pos + i;
	}

	inline function get_length():Int {
		return buffer.length;
	}
}
