package khm.internal;

import khm.math.FastMatrix3;

class BaseObjectUtils {
	public static function apply2DTransformation(transformation:FastMatrix3, obj:IGameObject):Void {
		// we need to keep translation-rotation-scale order here
		// but this can also be optimized with matrix mutation if will be needed
		if (obj.x != 0 || obj.y != 0) {
			transformation.setFrom(transformation.multmat(FastMatrix3.translation(obj.x, obj.y)));
		}
		if (obj.angle != 0) {
			transformation.setFrom(transformation.multmat(FastMatrix3.rotation(
				obj.angle.toRad().toFloat()
			)));
		}
		if (obj.scaleX != 1 || obj.scaleY != 1) {
			transformation.setFrom(transformation.multmat(FastMatrix3.scale(obj.scaleX, obj.scaleY)));
		}
	}

	public static function apply2DTransformationPixelPerfect(transformation:FastMatrix3, obj:IGameObject):Void {
		// we need to keep translation-rotation-scale order here
		// but this can also be optimized with matrix mutation if will be needed
		if (obj.x != 0 || obj.y != 0) {
			transformation.setFrom(transformation.multmat(FastMatrix3.translation(
				Std.int(obj.x * transformation.scaleX) / transformation.scaleX,
				Std.int(obj.y * transformation.scaleY) / transformation.scaleY
			)));
		}
		if (obj.angle != 0) {
			transformation.setFrom(transformation.multmat(FastMatrix3.rotation(
				obj.angle.toRad().toFloat()
			)));
		}
		if (obj.scaleX != 1 || obj.scaleY != 1) {
			transformation.setFrom(transformation.multmat(FastMatrix3.scale(obj.scaleX, obj.scaleY)));
		}
	}
}
