#if (macro || sys)
package khm.internal;

import haxe.Json;
import haxe.io.Path;
import sys.FileSystem;
import sys.io.File;

using Lambda;
using StringTools;

class BuildRelease {
	static function main() {
		final mainFile = "main.js";
		final dir = "build";
		final names = FileSystem.readDirectory(dir);
		final scriptPattern = ~/<script src="([^"]+)"><\/script>/g;
		var index = File.getContent('$dir/index.html');
		final releaseDir = '${getDateFolderName()}-${getCurrentDirName()}';
		if (FileSystem.exists(releaseDir)) {
			deleteDirRecursively(releaseDir);
			FileSystem.deleteDirectory(releaseDir);
		}
		FileSystem.rename("release", releaseDir);

		final params = '{
"activeVar": 1,
"var1": ${Json.stringify(Params.vars[0], "\t")},
"var2": ${Json.stringify(Params.vars[1], "\t")},
"var3": ${Json.stringify(Params.vars[2], "\t")},
"var4": ${Json.stringify(Params.vars[3], "\t")},
"var5": ${Json.stringify(Params.vars[4], "\t")}
}';
		// final json = Json.stringify(params, null, "\t");
		final matchParams = ~/([\t ]*)var params = {}/;
		if (matchParams.match(index)) {
			final tabs = matchParams.matched(1);
			final jsonLines = params.split("\n");
			var json = jsonLines.mapi((i, s) -> {
				if (i == 0) return s;
				if (i == jsonLines.length - 1) return tabs + s;
				return tabs + "\t" + s;
			}).join("\n");
			final matchRange = ~/{\n[ \t]+("min.+),\n[ \t]+("max.+),\n[ \t]+("value.+)\n[ \t]+}/g;
			json = matchRange.replace(json, "{$3, $1, $2}");
			final matchArrays = ~/(\[[^][]+\])/g;
			json = matchArrays.map(json, reg -> {
				final arrString = reg.matched(1);
				final noSpaces = ~/[\n\t ]/g.replace(arrString, "");
				return noSpaces.split(",").join(", ");
			});
			index = matchParams.replace(index, '${tabs}var params = $json');
		} else {
			haxe.Log.trace('`var params = {}` is not a empty object. `Params.hx` ignored.', null);
		}

		final soundMainJs = File.getContent('$releaseDir/sound/$mainFile');
		var singleJs = soundMainJs;
		// var singleJsNoSound = File.getContent('$releaseDir/no_sound/$mainFile');
		singleJs = singleJs.replace("window.location.href =", "var _no_rdr =");
		final singleFile = scriptPattern.map(index, r -> {
			var name = r.matched(1);
			#if canvas_only
			if (name.startsWith("phaser")) name = name.replace("phaser", "phaser-canvas");
			#end
			var file = name == mainFile ? soundMainJs : File.getContent('$dir/$name');
			if (name.startsWith("phaser")) {
				file = file.replace("new XMLHttpRequest", "null");
				file = file.replace('new window["XML" + "Http" + "Request"]', "null");
			}
			if (name == mainFile) {
				file = file.replace("window.location.href =", "var _no_rdr =");
			}
			if (name != mainFile) {
				singleJs = '$file\n' + singleJs;
				// singleJsNoSound = '$file\n' + singleJsNoSound;
			}
			return '<script type="text/javascript">\n$file\n</script>';
		});
		File.saveContent('$releaseDir/demo.html', singleFile);

		extractParamsFile(index, '$releaseDir/params.json');

		final indexSingleJs = scriptPattern.map(index, r -> {
			final name = r.matched(1);
			if (name == mainFile) return r.matched(0);
			return '';
		});

		File.saveContent('$releaseDir/sound/demo.html', indexSingleJs);
		File.saveContent('$releaseDir/sound/main.js', singleJs);
		final size = FileSystem.stat('$releaseDir/demo.html').size;
		final mb = Std.int(size / 1024 / 1024 * 1000) / 1000;
		haxe.Log.trace('Release build complete ($mb MiB)', null);

		// File.saveContent('$releaseDir/no_sound/demo.html', indexSingleJs);
		// File.saveContent('$releaseDir/no_sound/main.js', singleJsNoSound);
		// final size = dirSize('$releaseDir/no_sound/');
		// final mb = Std.int(size / 1024 / 1024 * 1000) / 1000;
		// haxe.Log.trace('($mb MiB without sound)', null);
	}

	static function dirSize(path:String):Int {
		var size = 0;
		for (item in FileSystem.readDirectory(path)) {
			if (item.startsWith(".")) continue;
			size += FileSystem.stat('$path/$item').size;
		}
		return size;
	}

	static function deleteDirRecursively(path:String):Void {
		var entries = FileSystem.readDirectory(path);
		for (entry in entries) {
			if (FileSystem.isDirectory(path + '/' + entry)) {
				deleteDirRecursively(path + '/' + entry);
				FileSystem.deleteDirectory(path + '/' + entry);
			} else {
				FileSystem.deleteFile(path + '/' + entry);
			}
		}
	}

	static function getCurrentDirName():String {
		final path = new Path(FileSystem.absolutePath("."));
		return path.file;
	}

	static function getDateFolderName():String {
		final date = Date.now();
		final year = '${date.getFullYear()}';
		final day = '${date.getDate()}'.lpad("0", 2);
		final month = '${date.getMonth() + 1}'.lpad("0", 2);
		return '$day.$month.$year';
	}

	static function extractParamsFile(index:String, savePath:String):Void {
		final lines = index.split("\n");
		final i = lines.findIndex(line -> line.trim().contains("params = {"));
		if (i == -1) {
			trace("line with `params = {` not found in index.html");
			return;
		}
		final json = "{\n" + readFunctionBody(lines, i, "params") + "\n}";
		File.saveContent(savePath, json);
	}

	static function readFunctionBody(lines:Array<String>, num:Int, fieldName:String):String {
		final body = new StringBuf();
		var section:BodySection = Code;
		var level = 0;
		while (num < lines.length) {
			final line = lines[num];
			var lineStart = 0;
			var lineEnd = line.length;
			var i = 0;
			while (i < line.length) {
				final code = line.fastCodeAt(i);
				switch (section) {
					case Code:
						switch (code) {
							case "{".code:
								// start body text from first bracket
								if (level == 0) lineStart = i + 1;
								level++;

							case "}".code:
								level--;
								if (level == 0) {
									lineEnd = i;
									break;
								} else if (level < 0) {
									throw 'Field "$fieldName" closed before been opened';
								}

							case "/".code:
								final next = line.fastCodeAt(i + 1);
								if (next == "/".code) section = SingleComment;
								else if (next == "*".code) section = MultiComment;
								if (section != Code) i++;

							case "'".code:
								section = SingleQuotes;
							case '"'.code:
								section = DoubleQuotes;
							case "`".code:
								section = Backticks;
						}
					case SingleComment:
						if (i == line.length - 1) section = Code;
					case MultiComment:
						if (code == "*".code) {
							final next = line.fastCodeAt(i + 1);
							if (next == "/".code) {
								section = Code;
								i++;
							}
						}
					case SingleQuotes:
						if (code == "\\".code) i++;
						else if (code == "'".code) section = Code;
					case DoubleQuotes:
						if (code == "\\".code) i++;
						else if (code == '"'.code) section = Code;
					case Backticks:
						if (code == "\\".code) i++;
						else if (code == "`".code) section = Code;
				}
				i++;
			}

			// add line break if body already has lines and this is not latest line
			if (body.length > 0 && lineEnd == line.length) body.add("\n");
			body.add(line.substring(lineStart, lineEnd));
			if (level == 0 && section == Code) break;
			num++;
		}
		return body.toString();
	}
}

enum abstract BodySection(Int) {
	var Code;
	var SingleComment;
	var MultiComment;
	var SingleQuotes;
	var DoubleQuotes;
	var Backticks;
}
#end
