package khm.math;

abstract Degrees(Float) from Float {
	public inline function new(f:Float) {
		this = f;
	}

	@:op(-A) function neg():Degrees;

	@:op(A > B) function gt(b:Degrees):Bool;

	@:op(A >= B) function gte(b:Degrees):Bool;

	@:op(A < B) function lt(b:Degrees):Bool;

	@:op(A <= B) function lte(b:Degrees):Bool;

	@:op(A == B) function eq(b:Degrees):Bool;

	@:commutative
	@:op(A + B) static function _add(a:Degrees, b:Float):Degrees;

	@:op(A + B) function add(b:Degrees):Degrees;

	@:commutative
	@:op(A - B) static function _sub(a:Degrees, b:Float):Degrees;

	@:op(A - B) function sub(b:Degrees):Degrees;

	@:commutative
	@:op(A * B) static function mult(a:Degrees, b:Float):Degrees;

	@:commutative
	@:op(A / B) static function div(a:Degrees, b:Float):Degrees;

	@:commutative
	@:op(A % B) static function mod(a:Degrees, b:Float):Degrees;

	/** `y` **/
	public inline function sin():Float {
		return Math.sin(this.toRad().toFloat());
	}

	/** `x` **/
	public inline function cos():Float {
		return Math.cos(this.toRad().toFloat());
	}

	public inline function toFloat():Float {
		return this;
	}

	public inline function toRad():Radians {
		return this.toRad();
	}

	// public inline function toString():String {
	// 	return '$this°';
	// }

	public inline function clamp():Degrees {
		return MathLib.angle.clamp(abstract).toFloat();
		// return this;
	}

	public inline function distTo(b:Degrees):Degrees {
		return MathLib.angle.dist(abstract, b);
	}

	public inline function toFixed(digits:Int = 2):Degrees {
		return this.toFixed(digits);
	}

	public inline function abs():Degrees {
		return this.abs();
	}
}
