package khm.tilemap;

class Tile {
	final tilemap:Tilemap;

	public var obj(default, null):IGameObject;
	public var ix:Int;
	public var iy:Int;

	public var layerId:Int;
	public var id(default, set):Int;

	/** Use this state instead of `obj.state`, so it will be keeped in case of `toContainer` conversion. **/
	public var state(default, never):ObjectState;

	function set_id(id:Int):Int {
		final rect = tilemap.tileset.layers[layerId][id].rect;
		getSprite().setCrop(rect);
		return this.id = id;
	}

	public function new(tilemap:Tilemap, obj:IGameObject, layerId:Int, id:Int, ix:Int, iy:Int) {
		this.tilemap = tilemap;
		this.obj = obj;
		this.layerId = layerId;
		@:bypassAccessor this.id = id;
		this.ix = ix;
		this.iy = iy;
	}

	/** Does not update sprite rect. **/
	public function setIdOnly(id:Int):Void {
		@:bypassAccessor this.id = id;
	}

	public function toContainer():Container {
		final container = getContainer();
		if (container != null) return container;

		final objC = new Container();
		objC.x = obj.x;
		objC.y = obj.y;
		obj.parent.add(objC);
		objC.add(obj);
		obj.x = 0;
		obj.y = 0;
		objC.shrink();
		obj = objC;
		return objC;
	}

	public function getSprite():Sprite {
		if (obj.type == SpriteType) return cast obj;
		if (obj.type == ContainerType) {
			final obj:Container = cast obj;
			final first = obj.first;
			if (first.type == SpriteType) return cast first;
			throw "Sprite is not found as first child in tile obj";
		}
		throw "Sprite not found in tile obj, tile is not a Sprite or Container with Sprite";
	}

	public function getContainer():Null<Container> {
		if (obj.type == ContainerType) return cast obj;
		// throw "obj is not a Container, call `convertToContainer` first";
		return null;
	}

	public function copy():Tile {
		return new Tile(tilemap, obj, layerId, id, ix, iy);
	}
}
