package khm.tilemap;

#if kha
import kha.graphics2.Graphics;
#end

class TilemapLayer extends Container {
	final tilemap:Tilemap;
	final layerId:Int;

	public var w(default, null):Int;
	public var h(default, null):Int;
	public final map:Array<Array<Tile>>;

	public function new(tilemap:Tilemap, layerId:Int, map:Array<Array<Tile>>, ?pos:DebugPosInfos) {
		super();
		this.setPosInfos(pos);
		this.tilemap = tilemap;
		this.layerId = layerId;

		final tileset = tilemap.tileset;
		final l = layerId;
		w = map[0].length;
		h = map.length;
		this.map = map;
		for (iy => row in map) {
			for (ix => tile in row) {
				add(tile.obj);
			}
		}

		setMapTiles();
		setSize(w * tilemap.tileW, h * tilemap.tileH);
	}

	function setMapTiles():Void {
		for (iy in 0...h) {
			for (ix in 0...w) {
				// map[iy][ix] ??= makeTile(tilemap, layerId, 0, ix, iy);
				final tile = map[iy][ix];
				// tile.ix = ix;
				// tile.iy = iy;
				tile.obj.x = tilemap.getTileX(ix);
				tile.obj.y = tilemap.getTileY(iy);
			}
		}
	}

	public static function fromJson(tilemap:Tilemap, layerId:Int, layer:Array<Array<Int>>):TilemapLayer {
		final map = [
			for (iy in 0...layer.length) [
				for (ix in 0...layer[iy].length) {
					final id = layer[iy][ix];
					final tile = makeTile(tilemap, layerId, id, ix, iy);
					tile;
				}
			]
		];
		return new TilemapLayer(tilemap, layerId, map);
	}

	static function makeTile(tilemap:Tilemap, layerId:Int, id:Int, ix:Int, iy:Int):Tile {
		final tilesetLayer = tilemap.tileset.layers[layerId];
		var t = tilesetLayer[id];
		if (id < 0) t = tilesetLayer[0];
		if (t == null) {
			t = tilesetLayer[0];
			throw 'Tile id $id not found for layer $layerId (x: $ix, y: $iy)';
		}
		final sprite = Sprite.fromTile(t.frame, t.rect);
		return new Tile(tilemap, sprite, layerId, id, ix, iy);
	}

	public function copy():TilemapLayer {
		final map:Array<Array<Tile>> = [
			for (iy in 0...h) [
				for (ix in 0...w)
					map[iy][ix].copy()
			]
		];
		return new TilemapLayer(tilemap, layerId, map);
	}

	#if kha
	override function render(g:Graphics):Void {
		if (map == null) return;

		if (!visible || alpha == 0) return;
		g.color = 0xFFFFFFFF;
		pushContainerSpace(g);

		var count = 0;
		for (iy in 0...h) {
			for (ix in 0...w) {
				final tile = map[iy][ix];
				final id:Int = tile.id;
				if (id < 1) continue;
				final p = tile.obj.getGlobalPosition();
				if (p.x + p.w < 0 || p.y + p.h < 0) continue;
				if (p.x > App.w || p.y > App.h) continue;
				tile.obj.render(g);
				count++;
			}
		}

		for (obj in children) {
			final p = obj.getGlobalPosition();
			if (p.x + p.w < 0 || p.y + p.h < 0) continue;
			if (p.x > App.w || p.y > App.h) continue;
			obj.render(g);
			count++;
		}
		popContainerSpace(g);

		// g.fontSize = 50;
		// g.font = kha.Assets.fonts.Karla_Regular ?? return;
		// g.pushTransformation(khm.math.FastMatrix3.identity());
		// g.drawString('${count}', layerId * 80, this.findScene().y);
		// g.popTransformation();
	}
	#end
	// public function extendTop(subLayer:TilemapLayer) {
	// 	w = Std.int(Math.max(w, subLayer.w));
	// 	h += subLayer.h;
	// 	for (iy => row in subLayer.map) {
	// 		map.unshift(row);
	// 		for (ix => tile in row) {
	// 			add(tile.obj);
	// 		}
	// 	}
	// 	setMapTiles();
	// 	setSize(w * tilemap.tileW, h * tilemap.tileH);
	// }
}
