package khm.tilemap;

import khm.ResItem;
import khm.Types.IRect;
import khm.utils.Atlas;

typedef TsProps = {
	?id:Int,
	?frame:ResItem,
	// grid-locked cords
	?tx:Int,
	?ty:Int,
	// or pixel cords
	?x:Int,
	?y:Int,
	?props:Any,
}

typedef TileData = {
	frame:ResItem,
	rect:IRect,
	props:Any,
}

class Tileset {
	public final layers:Array<Map<Int, TileData>> = [];
	public final tileW:Int;
	public final tileH:Int;

	@:allow(khm.tilemap.Tilemap)
	var nextFrame:Null<ResItem> = null;
	var nextX = 0;
	var nextY = 0;

	public function new(tileW:Int, tileH:Int, layers:Array<Array<TsProps>>):Void {
		this.tileW = tileW;
		this.tileH = tileH;
		// fill tileset properties based on prev props
		for (layerId => dataLayer in layers) {
			for (tileId => tile in dataLayer) {
				fillProps(tile, layerId, tileId);
			}
		}
		for (layerId => dataLayer in layers) {
			this.layers.push([]);
			final layer = this.layers[layerId];
			for (tileId => tile in dataLayer) {
				if (layer[tile.id] != null)
					throw 'Tile is already exist (layer $layerId, tile ${tile.id})';
				layer[tile.id] = {
					frame: tile.frame,
					rect: {
						x: tile.x,
						y: tile.y,
						w: tileW,
						h: tileH,
					},
					props: tile.props,
				};
			}
		}
	}

	function fillProps(tile:TsProps, layerId:Int, tileId:Int):Void {
		final frame = tile.frame ?? nextFrame;
		if (frame == null) throw 'Frame is null (layer $layerId, tile ${tile.id})';
		if (tile.id == null) tile.id = tileId;
		tile.frame = frame;
		if (tile.props == null) tile.props = {};
		// reset next cords on frame change
		if (nextFrame != tile.frame) {
			nextFrame = frame;
			nextX = 0;
			nextY = 0;
		}
		initTileCords(tile);
	}

	function initTileCords(tile:TsProps):Void {
		final atlasFrame = Atlas.getAtlasFrame(nextFrame);
		// final rect:IRect = atlasFrame.frame;
		final size = atlasFrame.sourceSize;
		if (tile.x == null && tile.y == null) {
			if (tile.tx != null && tile.ty != null) {
				tile.x = tile.tx * tileW;
				tile.y = tile.ty * tileH;
			} else {
				tile.x = nextX;
				tile.y = nextY;
			}
		}
		// check next x or enter new line
		nextX = tile.x + tileW;
		nextY = tile.y;
		if (nextX > size.w - tileW) {
			nextX = 0;
			nextY += tileH;
		}
	}
}
