package khm.tools;

import khm.Tween.TweenConfig;

class GameObjectAnimTools {
	public static function showAnim(obj:IGameObject, time = 500.0, delay = 0.0, ?onComplete:() ->
		Void, ?p:DebugPosInfos):Tween {
		return new Tween({
			targets: obj,
			alpha: 1,
			delay: delay,
			duration: time,
			onComplete: onComplete,
		}, p);
	}

	public static function hideAnim(obj:IGameObject, time = 500.0, delay = 0.0, ?onComplete:() ->
		Void, ?p:DebugPosInfos):Tween {
		return new Tween({
			targets: obj,
			alpha: 0,
			delay: delay,
			duration: time,
			onComplete: onComplete,
		}, p);
	}

	public static function opacityAnim(obj:IGameObject, alpha:Float, time = 500.0, delay = 0.0, ?onComplete:() ->
		Void, ?p:DebugPosInfos):Tween {
		return new Tween({
			targets: obj,
			alpha: alpha,
			delay: delay,
			duration: time,
			onComplete: onComplete,
		}, p);
	}

	public static function popUpAnim(obj:IGameObject, time = 500.0, delay = 0.0, scale = 1.0, ?onComplete:() ->
		Void, ?p:DebugPosInfos):Tween {
		return new Tween({
			targets: obj,
			alpha: 1,
			scale: scale,
			ease: BackOut,
			delay: delay,
			duration: time,
			onComplete: onComplete,
			onUpdate: e -> {
				for (obj in e.targets) {
					final obj:IGameObject = obj;
					if (obj.alpha > 1) obj.alpha = 1;
				}
			}
		}, p);
	}

	public static function popOutAnim(obj:IGameObject, time = 500.0, delay = 0.0, scale = 0.0, ?onComplete:() ->
		Void, ?p:DebugPosInfos):Tween {
		return new Tween({
			targets: obj,
			alpha: 0,
			scale: scale,
			ease: BackIn,
			delay: delay,
			duration: time,
			onComplete: onComplete,
			onUpdate: e -> {
				for (obj in e.targets) {
					final obj:IGameObject = obj;
					if (obj.alpha > 1) obj.alpha = 1;
				}
			}
		}, p);
	}

	public static function pulseAnim(obj:IGameObject, config:{
		?from:Float,
		to:Float,
		duration:Float,
		?getScale:() -> Float,
		?repeatDelay:Float,
		?delay:Float
	}, ?p:DebugPosInfos):Tween {
		var scale = obj.scale;
		if (config.getScale == null) config.getScale = () -> scale;
		return Tween.counter({
			targets: obj,
			from: config.from ?? 1.0,
			to: config.to,
			ease: SineInOut,
			duration: config.duration,
			repeat: -1,
			yoyo: true,
			delay: config.delay,
			repeatDelay: config.repeatDelay,
			onStart: () -> {
				if (config.delay == 0) return;
				scale = obj.scale;
			},
			onUpdate: tween -> {
				obj.scale = tween.getValue() * config.getScale();
			},
		}, p);
	}

	public static function slideInAnimX(
		obj:IGameObject, offsetX:Float, ease = Ease.BackOut,
		time = 500.0, delay = 0.0, ?onComplete:() -> Void, ?p:DebugPosInfos
	):Tween {
		return slideIn(obj, offsetX, 0, ease, time, delay, onComplete, p);
	}

	public static function slideInAnimY(
		obj:IGameObject, offsetY:Float, ease = Ease.BackOut,
		time = 500.0, delay = 0.0, ?onComplete:() -> Void, ?p:DebugPosInfos
	):Tween {
		return slideIn(obj, 0, offsetY, ease, time, delay, onComplete, p);
	}

	public static function slideOutAnimX(
		obj:IGameObject, offsetX:Float, ease = Ease.BackOut,
		time = 500.0, delay = 0.0, ?onComplete:() -> Void, ?p:DebugPosInfos
	):Tween {
		return slideOut(obj, offsetX, 0, ease, time, delay, onComplete, p);
	}

	public static function slideOutAnimY(
		obj:IGameObject, offsetY:Float, ease = Ease.BackOut,
		time = 500.0, delay = 0.0, ?onComplete:() -> Void, ?p:DebugPosInfos
	):Tween {
		return slideOut(obj, 0, offsetY, ease, time, delay, onComplete, p);
	}

	static function slideIn(
		obj:IGameObject, offsetX:Float, offsetY:Float, ease = Ease.BackOut,
		time = 500.0, delay = 0.0, ?onComplete:() -> Void, ?p:DebugPosInfos
	):Tween {
		final targets:Array<IGameObject> = (obj is Array) ? cast obj : [obj];
		for (obj in targets) {
			obj.x += offsetX;
			obj.y += offsetY;
		}
		var minAlpha = 0.0;
		final config:TweenConfig = {
			targets: targets,
			alpha: 1,
			ease: ease,
			delay: delay,
			duration: time,
			onUpdate: e -> {
				for (obj in e.targets) {
					final obj:IGameObject = obj;
					if (e.ratio > minAlpha) minAlpha = e.ratio;
					obj.alpha = minAlpha.clamp(0, 1);
				}
			},
			onComplete: onComplete,
		};
		if (offsetX != 0) config.x = targets[0].x - offsetX;
		if (offsetY != 0) config.y = targets[0].y - offsetY;
		return new Tween(config, p);
	}

	static function slideOut(
		obj:IGameObject, offsetX:Float, offsetY:Float, ease = Ease.BackOut,
		time = 500.0, delay = 0.0, ?onComplete:() -> Void, ?p:DebugPosInfos
	):Tween {
		final targets:Array<IGameObject> = (obj is Array) ? cast obj : [obj];
		var minAlpha = 0.0;
		final config:TweenConfig = {
			targets: targets,
			alpha: 0,
			ease: ease,
			delay: delay,
			duration: time,
			onUpdate: e -> {
				for (obj in e.targets) {
					final obj:IGameObject = obj;
					if (e.ratio > minAlpha) minAlpha = e.ratio;
					obj.alpha = 1 - minAlpha.clamp(0, 1);
				}
			},
			onComplete: onComplete,
		};
		if (offsetX != 0) config.x = targets[0].x + offsetX;
		if (offsetY != 0) config.y = targets[0].y + offsetY;
		return new Tween(config, p);
	}
}
