package khm.ui;

import khm.Event.PointerEvent;
import khm.Res;

using khm.tools.GameObjectAnimTools;

@:structInit
class IconButtonOptions {
	public var padding = 10.0;
	public var isActive = false;
	public var onClick:(e:PointerEvent) -> Void = null;
	public var size = 50.0;
	public var color:Null<Color> = null;
}

class IconButton extends Widget {
	final config:IconButtonOptions;
	final active:NineSlice;
	final icon:Sprite;

	public function new(frame:ResItem, ?opts:IconButtonOptions, ?pos:DebugPosInfos) {
		super();
		this.setPosInfos(pos);
		opts ??= {};
		config = opts;

		active = new NineSlice(Res.debug.round_rect, 20);
		add(active);

		icon = addSprite(frame);
		if (config.onClick != null) {
			on(PointerDown, config.onClick);

			on(PointerOver, e -> {
				if (config.isActive) return;
				active.removeTweens();
				active.opacityAnim(0.1, 200);
			});
			on(PointerOut, e -> {
				if (config.isActive) return;
				active.removeTweens();
				active.hideAnim(200);
			});
		}
	}

	override function build():Widget {
		icon.setOrigin(0, 0);
		icon.x = config.padding;
		icon.y = config.padding;
		icon.displayWidth = config.size;
		icon.displayHeight = config.size;
		icon.tint = config.color ?? context.theme.primaryColor;

		width = icon.displayWidth + config.padding * 2;
		height = icon.displayHeight + config.padding * 2;

		active.alpha = config.isActive ? 0.15 : 0;
		// if (config.isActive) {
		active.x = icon.x + icon.displayWidth / 2;
		active.y = icon.y + icon.displayHeight / 2;
		active.width = width * 2;
		active.height = height * 2;
		active.scale = 0.5;
		// }

		return this;
	}
}
