package khm.ui;

/**
	The `mainAxisAlign` property determines how `Row` and `Column` can position
	their children on their main axes.

	A `Row` main axis is horizontal, and a `Column` main axis is vertical.
**/
enum abstract MainAxisAlignment(Int) {
	/**
		Place the children as close to the start of the main axis as possible (Left for `Row`, top for `Column`).

		`[AB___]`
	**/
	var Start;

	/**
		Place the children as close to the end of the main axis as possible (Right for `Row`, bottom for `Column`).

		`[___AB]`
	**/
	var End;

	/**
		Place the children as close to the middle of the main axis as possible.

		`[__AB__]`
	**/
	var Center;

	/**
		Place the free space evenly between the children.

		`[A___B]`
	**/
	var SpaceBetween;

	/**
		Place the free space evenly between the children as well as half of that space before and after the first and last child.

		`[_A__B_]`
	**/
	var SpaceAround;

	/**
		Place the free space evenly between the children as well as before and after the first and last child.

		`[_A_B_]`
	**/
	var SpaceEvenly;
}
