package khm.ui;

import khm.Event.PointerEvent;
import khm.Res;

using khm.tools.GameObjectAnimTools;

@:structInit
class TextButtonOptions {
	public var padding = 10.0;
	public var isActive = false;
	public var isDisabled = false;
	public var onClick:(e:PointerEvent) -> Void = null;
}

class TextButton extends Widget {
	final active:NineSlice;
	final text:Text;

	public final config:TextButtonOptions;

	public function new(textString:String, ?opts:TextButtonOptions, ?pos:DebugPosInfos) {
		super();
		this.setPosInfos(pos);
		opts ??= {};
		config = opts;

		active = new NineSlice(Res.debug.round_rect, 20);
		add(active);

		this.text = addText({
			text: textString,
			style: {
				fontFamily: "Helvetica",
				fontSize: 24,
			}
		});
		if (config.isDisabled) {
			text.alpha = 0.5;
		}
		if (config.onClick != null) {
			on(PointerDown, config.onClick);
			on(PointerOver, e -> {
				if (config.isActive) return;
				active.removeTweens();
				active.opacityAnim(0.1, 200);
			});
			on(PointerOut, e -> {
				if (config.isActive) return;
				active.removeTweens();
				active.hideAnim(200);
			});
		}
	}

	override function build():Widget {
		text.setOrigin(0, 0);
		text.x = config.padding;
		text.y = config.padding;

		width = text.width + config.padding * 2;
		height = text.height + config.padding * 2;

		if (parent.toWidget().isExpanded()) {
			width = context.maxW;
			// height = context.maxH;
		}

		active.alpha = config.isActive ? 0.15 : 0;
		// if (config.isActive) {
		active.x = width / 2;
		active.y = height / 2;
		active.width = width * 2;
		active.height = height * 2;
		active.scale = 0.5;
		// }

		return this;
	}

	override function dispose() {
		text.destroy();
		super.dispose();
	}
}
