package khm.ui;

import khm.Types.Rect;
import khm.math.FastMatrix3;

@:using(khm.ui.WidgetTools)
class Widget extends Container {
	public static var maxSize = 99999;

	public var clipBehavior:ClipBehavior = None;
	public var debugClip = false;
	public var flexFactor = 0.0;

	/**
		Build context contains info about latest build for current widget:

		Max available size (`maxW`, `maxH`), setted by `parent` widget `build()` call.

		And cache of last size for current widget, after `build()` call.
	**/
	public final context:BuildContext = {};

	public function new() {
		super();
		type = WidgetType;
		originX = 0;
		originY = 0;
	}

	/**
		Updates widget size and children based on `context` data, mostly by `context.maxW` and `maxH`.

		Child `context` should be updated by `parent.build()` before this `build()` call.
	**/
	public function build():Widget {
		return this;
	}

	/**
		Recursively calls `dispose` on every widget child, which removes object from tree and removes its events.

		Does not remove `this` widget and its events.
	**/
	public function disposeChildren():Void {
		var i = children.length;
		while (i-- > 0) {
			final obj = children[i];
			obj.off();
			final widget = obj.toWidget() ?? {
				remove(obj);
				continue;
			}
			widget.dispose();
		}
	}

	/**
		Recursively calls `dispose` on every widget child, which removes object from tree and removes its events.

		Then cleans and removes `this` widget too. **/
	public function dispose():Void {
		disposeChildren();
		parent.remove(this);
		this.off();
	}

	#if kha
	override function render(g:kha.graphics2.Graphics):Void {
		if (!visible || alpha == 0) return;
		g.color = 0xFFFFFFFF;
		pushContainerSpace(g);
		var hasCrop = false;
		if (clipBehavior == Crop) {
			// g.disableScissor();
			final p = getCropRect();
			if (p.w > 0 && p.h > 0) {
				// if there is something to crop
				// (this optimization needs to also check scroll)
				// if (p.w < width || p.h < height) {
				hasCrop = true;
				g.scissor(
					Std.int(p.x),
					Std.int(p.y),
					Std.int(p.w),
					Std.int(p.h)
				);
				// }
			}
		}
		for (obj in children) obj.render(g);
		if (clipBehavior == Crop && hasCrop) g.disableScissor();
		popContainerSpace(g);
	}
	#end

	inline function getCropRect():Rect {
		if (context.crop == null) {
			return {
				x: 0,
				y: 0,
				w: 0,
				h: 0
			};
		}
		final t = parent != null ? parent.transformation : FastMatrix3.identity();
		final m:FastMatrix3 = t;
		final p:Rect = context.crop;
		final x = m.scaleX * p.x + m.skewX * p.y + m.translateX;
		final y = m.scaleY * p.y + m.skewY * p.x + m.translateY;
		return {
			x: x,
			y: y,
			w: p.w * m.scaleX,
			h: p.h * m.scaleY
		}
	}
}
