package khm.utils;

import khm.Event.DragEvent;
import khm.Event.PointerEvent;
import khm.Lazyr.*;
import khm.Res;
import khm.ui.*;
import khm.ui.Column.ColumnOptions;
import khm.ui.Theme.ThemeData;

using khm.tools.GameObjectAnimTools;

class DebugPanel extends Scene {
	public static var instance:Null<DebugPanel>;

	final scene:Scene;
	var panelH = 200;
	var ui:Widget;
	var selectedTreeObject:Null<IGameObject>;
	var treeColumn:Column;
	var activePanelI = 0;
	final lastLogMessages:Array<{
		args:Array<Dynamic>,
		pos:String
	}> = [];

	public function new(scene:Scene) {
		super();
		instance = this;
		// todo update default scene if active is hidden
		// (also resize it)
		this.scene = scene;
		panelH = keepDebugValue("debugPanel.panelH", () -> panelH);
		activePanelI = keepDebugValue("debugPanel.activePanelI", () -> activePanelI);
	}

	override function onShow() {
		function onTabClick(e:PointerEvent, activeI:Int):Void {
			final target:TextButton = cast e.target;
			final parent:Row = cast target.parent;
			var n = 0;
			for (obj in parent.children) {
				if (obj is TextButton) {
					final obj:TextButton = cast obj;
					obj.config.isActive = n == activeI;
					n++;
				}
			}
			target.config.isActive = true;
			activePanelI = activeI;
			parent.build();
		}
		on(ScenesChange, e -> {
			if (!isAttached) return;
			final preLast = App.scenes[App.scenes.length - 2];
			if (DebugPanel.instance != preLast
				|| DebugRectOverlay.instance != App.scenes.last()) {
				DebugPanel.instance?.bringToTop();
				DebugRectOverlay.instance?.bringToTop();
			}
		});

		final themeData:ThemeData = {
			accentColor: 0xFFFF00FF,
		};
		ui = new Theme({
			theme: themeData,
			child: new Box(Res.debug.panel_bg, 15, {
				child: new Column({
					children: [
						new Row({
							children: [
								Container.sized(5),
								new TextButton("Tree", {
									isActive: activePanelI == 0,
									onClick: e -> {
										onTabClick(e, 0);
										selectObjectInTree(null);
									}
								}),
								Container.sized(5),
								new TextButton("Console", {
									isActive: activePanelI == 1,
									onClick: e -> {
										onTabClick(e, 1);
										addLogItem(null);
									}
								}),
								Container.sized(5),
								new TextButton("Creator", {
									isActive: activePanelI == 2,
									isDisabled: true
									// onClick: e -> onTabClick(e, 2)
								}),
								new Spacer(),
								new IconButton(Res.debug.icon_close, {
									size: 25,
									color: themeData.primaryColorDim,
									onClick: e -> animateHide()
								}),
							]
						}),
						new Divider(Res.debug.line_hor, {
							padding: 5,
							height: 1.5
						}),
						new Expanded(
							treeColumn = new Column({
								scrollable: true,
								children: [],
							})
						),
						new Dragger(Res.debug.panel_drag, {
							onDrag: e -> {
								final dragY = e.p.y - e.localY;
								panelH = Std.int(dragY + 15);
								panelH = panelH.clamp(200, Std.int(App.h * 0.9));
								onResize();
								restrictColumnScrolling(treeColumn);
								scene.children.flatIter(obj -> Event.emitResize(obj));
							},
						}),
					]
				}),
			})
		});
		treeColumn.parent.on(MouseWheel, e -> {
			if (e.isTrackpad && e.eventCount % 5 != 0) return;
			final speedY = e.delta * 25;
			treeColumn.scrollY += speedY;
			restrictColumnScrolling(treeColumn);
			build();
		});
		add(ui);

		off(Resize);
		scene.off(Resize);
		scene.on(Resize, e -> onResize());
		onResize();

		on(KeyDown, e -> {
			if (selectedTreeObject == null) {
				e.propagate = true;
				return;
			}
			switch e.key {
				case Up, Down:
					final dir = e.key == Up ? -1 : 1;
					final target = selectedTreeObject;
					final next = getNextTreeSibling(target, dir);
					if (next == null) return;
					if (next == scene || DebugRectOverlay.isIgnoredObject(next)) return;
					DebugRectOverlay.toggle(next);
				case _:
					e.propagate = true;
			}
		});

		final cropOrStrict:ClipBehavior = #if kha Crop #else Strict #end;
		treeColumn.clipBehavior = cropOrStrict;
		// treeColumn.debugClip = true;
		delayed(1, () -> {
			// instance = null;
			selectObjectInTree(null);
		});

		App.onLog(args -> {
			delayed(1, () -> {
				addLogItem(args);
			});
		});
	}

	function restrictColumnScrolling(column:Column) {
		var max = column.height - column.context.maxH;
		if (max < 0) max = 0;
		column.scrollY = column.scrollY.clamp(0, max);
	}

	function animateHide():Void {
		DebugRectOverlay.destroy();
		this.hideAnim(200, 300);
		this.addTween({
			duration: 500,
			onUpdate: e -> {
				final ratio = 1 - e.ratio;
				final h = panelH;
				height = ratio * h;
				build();
				resizeMainScene();
				scene.children.flatIter(obj -> Event.emitResize(obj));
			},
			onComplete: () -> hide()
		});
	}

	function animateShow():Void {
		show();
		this.showAnim(300);
		this.addTween({
			duration: 500,
			onUpdate: e -> {
				final ratio = e.ratio;
				final h = panelH;
				height = ratio * h;
				build();
				resizeMainScene();
				scene.children.flatIter(obj -> Event.emitResize(obj));
			},
		});
	}

	override function onHide():Void {
		ui.dispose();
		width = 0;
		height = 0;
		resizeMainScene();
		scene.off(Resize);
		scene.on(Resize, e -> scene.onResize());
		this.removeTweens();
		Event.emitResize(null);
		instance = null;
	}

	override function onResize() {
		width = App.w;
		height = panelH;
		build();
		resizeMainScene();
	}

	function build():Void {
		ui.context.maxW = sceneW;
		ui.context.maxH = sceneH;
		ui.build();
	}

	function treeColumnBuild():Void {
		// treeColumn.build();
		final flex = treeColumn.parent;
		// parent needs to re-pos flex
		flex.parent.toWidget().build();
	}

	function resizeMainScene():Void {
		scene.setSize(App.w, App.h - sceneH);
		scene.y = sceneH;
	}

	function getNextTreeSibling(target:IGameObject, dir:Int):IGameObject {
		final isContainerLike = target.type.isContainerLike();
		if (dir == 1) {
			if (isContainerLike) {
				final target:Container = cast target;
				if (target.children.length > 0) return target.first;
			}
		}
		var parent = target.parent ?? return null;
		while (true) {
			final i = parent.children.indexOf(target) + dir;
			final length = parent.children.length;
			if (i == -1) return parent;
			if (i == length) {
				target = parent;
				parent = target.parent ?? return null;
			} else {
				final target = parent.children[i] ?? return null;
				if (dir == -1) {
					final isContainerLike = target.type.isContainerLike();
					if (isContainerLike) {
						final target:Container = cast target;
						if (target.children.length > 0) return target.last;
					}
				}
				return target;
			}
		}
	}

	public static function ignoreObject(obj:IGameObject):Void {
		obj.state.debug.ignoreDebugPanel = true;
	}

	public function selectObjectInTree(target:IGameObject):Void {
		if (activePanelI != 0) return;

		var hasActive = target != null;

		selectedTreeObject = target;

		target ??= scene.first ?? return;

		if (!App.scenes.contains(this)) {
			if (hasActive) animateShow();
			else return;
		}

		final scene = target.findScene() ?? return;
		if (scene == this) return;

		treeColumn.disposeChildren();

		var activeI = 1;
		scene.children.flatIter(obj -> {
			if (obj.parent != null) {
				final parent:Container = cast obj.parent;
				if (parent.name == "DebugRect") return;
				if (parent is NineSlice) return;
				if (parent is khm.tilemap.TilemapLayer) return;
			}
			if (obj.findParent(obj -> obj.state.debug.ignoreDebugPanel) != null) return;
			if (obj.state.debug.ignoreDebugPanel) return;
			final isActive = hasActive && obj == target;
			final row = makeTreeRow(obj, isActive);
			treeColumn.add(row);
			if (isActive) activeI = treeColumn.children.indexOf(row);
		});
		final activeScroll = 36 * (activeI - 1);
		final max = treeColumn.children.length * 36;
		final scroll = activeScroll.clamp(0, max);
		treeColumn.scrollY = scroll;
		restrictColumnScrolling(treeColumn);

		treeColumnBuild();
	}

	function makeTreeRow(obj:IGameObject, isActive:Bool):Row {
		final iconFrame = switch obj.type {
			case SceneType: Res.debug.icon_layers2;
			case ContainerType:
				if (obj is NineSlice) {
					final obj:NineSlice = cast obj;
					obj.centerSprite.id;
				} else Res.debug.icon_layers;
			case SpriteType:
				final sprite:Sprite = cast obj;
				sprite.id;
			case TweenType: Res.debug.round_rect;
			case TextType: Res.debug.round_rect;
			case GraphicsType: Res.debug.round_rect;
			case ParticlesType: Res.debug.round_rect;
			case WidgetType: Res.debug.icon_layers;
		}
		final icon = new IconButton(iconFrame, {
			padding: 0,
			size: 25,
			color: isActive ? ui.context.theme.accentColor : ui.context.theme.primaryColor
		});
		final name = genObjName(obj);
		var posText = obj.posInfosLine() ?? ":???";
		final posTextI = posText.indexOf(":");
		if (posTextI != -1) posText = posText.substr(posTextI);
		final posBtn = new TextButton(posText, {
			padding: 4,
			onClick: e -> DebugRectOverlay.gotoDebugPosition(obj)
		});
		var level = 0;
		obj.findParent(obj -> {
			level++;
			false;
		});

		var children:Array<IGameObject> = [];
		for (i in 0...level - 1) {
			final divider = new Divider(Res.debug.line_hor, {
				padding: 0,
				axis: Vertical,
				height: 36,
				color: ui.context.theme.primaryColor,
				width: 2,
			});
			children.push(divider);
			children.push(
				Container.sized(7)
			);
		}
		children = children.concat([
			icon,
			new Expanded(
				new TextButton(name, {
					padding: 4,
					onClick: e -> {
						DebugRectOverlay.toggle(obj);
					}
				}),
			),
			// new Spacer(),
			posBtn
		]);
		if (level == 1 && obj.state.depth != null) {
			final depthText = new Text('${obj.state.depth}');
			children.insert(-1, depthText);
		}
		final selectionDividerW = 3;
		if (isActive) {
			children.unshift(
				Container.sized(10)
			);
			children.unshift(
				new Divider(Res.debug.line_hor, {
					axis: Vertical,
					color: ui.context.theme.accentColor,
					width: selectionDividerW,
					height: 36
				})
			);
		} else {
			children.unshift(
				Container.sized(10 + selectionDividerW)
			);
		}

		final row = new Row({
			children: children
		});
		row.name = name;
		return row;
	}

	function genObjName(obj:IGameObject):String {
		return switch obj.type {
			case SceneType, ContainerType, WidgetType:
				final obj:Container = cast obj;
				final typeName = obj.type == SceneType ? "scene" : "container";
				var name = obj.name;
				if (name == null) {
					final classPath = Type.getClassName(Type.getClass(obj));
					final className = classPath.split(".").pop();
					if (className == "Container") typeName;
					if (className == "NineSlice" && obj is NineSlice) {
						final obj:NineSlice = cast obj;
						'$className (${genObjName(obj.centerSprite)})';
					} else className;
				} else name;
			case SpriteType:
				final obj:Sprite = cast obj;
				if (obj.id.frame.length == 0) {
					obj.id.texture;
				} else '${obj.id.texture}.${obj.id.frame}';
			case TweenType:
				"Tween";
			case TextType:
				"Text";
			case GraphicsType:
				"Graphics";
			case ParticlesType:
				"Particles";
		}
	}

	function addLogItem(?args:Array<Dynamic>):Void {
		if (args != null) {
			if (lastLogMessages.length > 10) lastLogMessages.pop();
			lastLogMessages.unshift({
				pos: args.shift(),
				args: args,
			});
		}

		if (activePanelI != 1) return;
		if (!App.scenes.contains(this)) return;

		treeColumn.disposeChildren();

		for (log in lastLogMessages) {
			final pos = log.pos;
			final args = log.args;
			treeColumn.add(
				new Row({
					children: [
						new TextButton(pos, {
							onClick: e -> {
								DebugRectOverlay.openVscodePosLine(pos);
							}
						}),
						new TextButton(args.join(", ")),
					],
				})
			);
			treeColumn.first.toWidget().debugClip = true;
		}

		treeColumnBuild();
	}
}

class ColumnWithDragger extends Column {
	public function new(config:ColumnOptions) {
		super(config);
	}
}

@:structInit
class DraggerOptions {
	public var onDrag:(e:DragEvent) -> Void = null;
	public var color:Null<Color> = null;
	public var hitAreaPadding = 40;
}

class Dragger extends Widget {
	public function new(frame:ResItem, config:DraggerOptions, ?pos:DebugPosInfos) {
		super();
		this.setPosInfos(pos);
		final iconC = addContainer();
		final icon = iconC.addSprite(frame);
		icon.setOrigin(0);
		icon.tint = config.color ?? context.theme.primaryColor;
		icon.alpha = 0.5;
		iconC.shrink();
		fit();
		iconC.width += config.hitAreaPadding;
		iconC.height += config.hitAreaPadding;
		if (config.onDrag != null) {
			iconC.on(PointerDown, e -> {
				e.stopImmediatePropagation();
			});
			iconC.on(Drag, config.onDrag);
			iconC.on(DragEnd, config.onDrag);

			iconC.on(PointerOver, e -> {
				icon.removeTweens();
				icon.showAnim(200);
			});
			iconC.on(PointerOut, e -> {
				icon.removeTweens();
				icon.opacityAnim(0.5, 200);
			});
		}
	}
}
