package khm.utils;

import khm.Event.DragEvent;
import khm.Event.KeyEvent;
import khm.Event.PointerEvent;
import khm.Lazyr.PosObjectConfig;
import khm.Lazyr.PosRect;
import khm.utils.macro.Macro;

using khm.tools.PhaserTools;

class DebugRectOverlay extends Scene {
	public static var instance:Null<DebugRectOverlay>;
	public static var useVscodeLinks = true;
	public static var activeDebugRect:DebugRect = null;
	public static var isLogSize = false;
	public static var fontSize = 20;
	public static var enableRightButton = true;

	var scene:Scene;

	public function new(scene:Scene) {
		super();
		instance = this;
		this.scene = scene;
		ignoreObject(this);
		final g = new Graphics();
		g.font = "Courier";
		add(g);
		on(ScenesChange, e -> {
			if (e.scene == this) return;
			if (App.scenes.last() == this) return;
			bringToTop();
		});
		on(BeforeRender, e -> {
			g.begin();
			g.fontSize = fontSize;
			final target = activeDebugRect?.target ?? {
				g.end();
				return;
			}
			final tweensLength = Tween.getAll(target).length;
			final events = Event.getKindsOf(target);

			var hasUserCallbacks = false;
			final userCallbacksText = events.map(kind -> {
				final callbacks = Event.getCallbacks(target, kind);
				final userCallbacks = callbacks.filter(cb -> !isDebugEventCallback(cb));
				final length = userCallbacks.length;
				if (length == 0) return "";
				else hasUserCallbacks = true;
				return '${kind.toString()} ($length)';
			}).filter(s -> s.length > 0).join(", ");

			// var hasDebugCallbacks = false;
			// final debugCallbacksText = events.map(kind -> {
			// 	final callbacks = Event.getCallbacks(target, kind);
			// 	final debugCallbacks = callbacks.filter(isDebugEventCallback);
			// 	final length = debugCallbacks.length;
			// 	if (length == 0) return "";
			// 	else hasDebugCallbacks = true;
			// 	return '${kind.toString()} ($length)';
			// }).filter(s -> s.length > 0).join(", ");

			final maxW = Math.max(scene.sceneW, App.w - scene.x);
			final lines = target.log().chunkSubstr(Std.int(maxW / (fontSize * 0.61)));
			final objData = logObjectData(target);
			if (objData.length > 0) lines.push(objData);
			if (tweensLength > 0) lines.push('tweens: $tweensLength');
			if (hasUserCallbacks) lines.push('user events: $userCallbacksText');
			final state:Dynamic = Reflect.copy(target.state);
			state.debug = null;
			state._listeners = null;
			#if js
			Reflect.deleteField(state, "debug");
			Reflect.deleteField(state, "_listeners");
			#end
			lines.push('state: $state');
			// if (hasDebugCallbacks) lines.push('debug events: $debugCallbacksText');
			var offY = scene.y + 10;
			if (activeDebugRect != null) {
				final target = activeDebugRect.target;
				if (target.type.isContainerLike()) {
					final target:Container = cast target;
					final p = target.getGlobalPosition();
					if (Collision.aabb(p, {
						x: 10,
						y: offY,
						w: App.w,
						h: 100
					})) {
						final newY = p.y + p.h + 10;
						if (newY < App.h * 0.8) offY = newY;
					}
				}
			}
			for (line in lines) {
				final x = scene.x + 10;
				final y = offY;
				final t = g.drawString(line, x, y);
				offY += t.displayHeight;
			}
			g.end();
		});

		var time = App.stamp();
		on(PointerDown, e -> {
			e.propagation(true);
			// enable / disable debug rect
			if (enableRightButton && e.p.button == Right) {
				e.propagation(false);
				scene = pickSceneParent(e.p) ?? {
					trace('scene is not found on x: ${e.p.x}, y: ${e.p.y}');
					trace("all scene cords:");
					for (s in App.scenes) {
						trace('x: ${s.x}, y: ${s.y}, w: ${s.width}, h: ${s.height} | ' + s.log());
					}
					return;
				};
				final target = pickTarget(scene, e.p);
				if (target == null) {
					if (activeDebugRect != null) {
						activeDebugRect.destroy();
						activeDebugRect = null;
					}
					return;
				}
				toggle(target);
				return;
			}

			final activeRect = activeDebugRect ?? return;
			// if rect is active, detect double click for navigation
			final now = App.stamp();
			final isDoubleClick = now - time < 400;
			time = now;
			final isOnRect = Event.hitTest(e.p, activeRect.debugRect);
			if (isOnRect) e.propagation(false);
			if (!isDoubleClick) return;
			e.propagation(false);
			final scopes = getParentsChain(activeRect.target);
			scopes.unshift(activeRect.target);
			for (scope in scopes) {
				final target = pickTarget(scope, e.p);
				if (target == null) continue;
				if (target == activeRect.target) continue;
				toggle(target);
				break;
			}
		});
	}

	function pickTarget(parent:IGameObject, p:Pointer):Null<IGameObject> {
		if (!parent.type.isContainerLike()) return null;
		final parent:Container = cast parent;
		var i = parent.children.length;
		while (i-- > 0) {
			final obj = parent.children[i];
			if (!Event.hitTest(p, obj)) continue;
			if (isIgnoredObject(obj)) continue;
			return obj;
		}
		return null;
	}

	function pickSceneParent(p:Pointer):Scene {
		// trace(App.scenes.map(scene -> scene.log()));
		var i = App.scenes.length;
		while (i-- > 0) {
			final scene = App.scenes[i];
			if (isIgnoredObject(scene)) continue;
			if (!Event.hitTest(p, scene)) continue;
			return scene;
		}
		return scene;
	}

	function isDebugEventCallback(callback:(e:Event) -> Void):Bool {
		final rect = activeDebugRect ?? return false;
		return rect.isDebugEventCallback(callback);
	}

	public static function getParentsChain(obj:IGameObject):Array<IGameObject> {
		final parents:Array<IGameObject> = [];
		while (true) {
			final parent = obj.parent;
			if (parent == null) return parents;
			parents.push(parent);
			obj = parent;
		}
	}

	public static function ignoreObject(obj:IGameObject):Void {
		obj.state.debug.ignoreDebugRect = true;
	}

	public static function isIgnoredObject(obj:IGameObject):Bool {
		return obj.state.debug.ignoreDebugRect ?? false;
	}

	public static function toggle(target:IGameObject):Void {
		final rect = activeDebugRect;
		destroyRect();
		if (rect?.target != target) activeDebugRect = new DebugRect(instance, target);
		if (activeDebugRect == null) {
			DebugPanel.instance?.selectObjectInTree(null);
			return;
		}
		DebugPanel.instance?.selectObjectInTree(target);
	}

	public static function destroy():Void {
		destroyRect();
		DebugPanel.instance?.selectObjectInTree(null);
	}

	static function destroyRect():Void {
		if (activeDebugRect == null) return;
		activeDebugRect.destroy();
		activeDebugRect = null;
	}

	public static dynamic function logObjectData(obj:IGameObject):String {
		// final isLogSize = true;
		var t = if (isLogSize) {
			final x = obj.x.toFixed(1);
			final y = obj.y.toFixed(1);
			final w = obj.width.toFixed(1);
			final h = obj.height.toFixed(1);
			final scaleX = obj.scaleX.toFixed(2);
			final scaleY = obj.scaleY.toFixed(2);
			final scale = scaleX == scaleY ? '$scaleX' : '$scaleX,$scaleY';
			final dw = obj.displayWidth.toFixed(1);
			final dh = obj.displayHeight.toFixed(1);
			'$x,$y ${w}x$h, scale $scale (${dw}x$dh)';
		} else {
			"";
		}
		final scale = obj.scale.toFixed(2);
		final ox = obj.originX.toFixed(2);
		final oy = obj.originY.toFixed(2);
		final isDisabledInteractive = obj.state.debug.disabledInteractive ?? false;
		if (!obj.isInteractive || isDisabledInteractive) t += 'isInteractive: false';
		if (!isLogSize && scale != 1) t += ', scale: $scale';
		if (!obj.visible) t += ', visible: ${obj.visible}';
		if (obj.alpha != 1) t += ', alpha: ${obj.alpha.toFixed(2)}';
		if (obj.originX != 0.5 || obj.originY != 0.5) t += ', origin $ox,$oy';
		if (obj.angle != 0) t += ', angle ${obj.angle.toFloat().toFixed(1)}°';
		if (t.startsWith(", ")) t = t.substr(2);
		return t;
	}

	public static function gotoDebugPosition(sprite:IGameObject):Void {
		final lineText = sprite.posInfosLine() ?? {
			trace('debug pos is unknown | ${sprite.log()}');
			return;
		}
		if (!DebugRectOverlay.useVscodeLinks) {
			trace('pos copied: $lineText | ${sprite.log()}');
			Clipboard.copy(lineText);
		}
		openVscodePosLine(lineText);
	}

	public static function openVscodePosLine(posInfosLine:String):Void {
		// trace(Macro.getDefines());
		final srcFolder = Macro.getSrcDir();
		var path = '$srcFolder/$posInfosLine';
		path = path.replace("/src/src/", "/src/");
		// replace src/khm lib to absolute lib path from define
		if (path.contains("src/khm/")) {
			final endI = path.indexOf("src/khm/");
			final khmPath = Macro.getDefine("khm_path");
			if (khmPath != null) {
				path = khmPath + "/" + path.substr(endI);
			}
		}
		#if js
		final gotoLink = 'vscode://file/$path:${99 + Std.random(100)}';
		js.Browser.window.location.href = gotoLink;
		#else
		trace("vscode links not implemented on this target");
		trace(path);
		Clipboard.copy('$path:${99 + Std.random(100)}');
		#end
	}
}

class DebugRect {
	public final overlay:Scene;
	public final target:IGameObject;
	public final debugRect:Container;

	final disabledInteractives:Array<IGameObject> = [];

	public function new(overlayScene:Scene, target:IGameObject) {
		this.overlay = overlayScene;
		this.target = target;
		final parents = DebugRectOverlay.getParentsChain(target);
		parents.unshift(target);
		for (object in parents) {
			if (object.isInteractive) continue;
			object.state.debug.disabledInteractive = true;
			object.isInteractive = true;
			disabledInteractives.push(object);
		}
		if (isPosContainerRect(target)) {
			debugRect = initPosContainer();
		} else {
			debugRect = initSprite();
		}
	}

	function isPosContainerRect(target:IGameObject):Bool {
		return switch target.type {
			case SceneType, ContainerType, WidgetType:
				if (target.parentContainer() != null) {
					false;
				} else if (target.originX != 0.5 || target.originY != 0.5) {
					false;
				} else {
					true;
				}
			case SpriteType, TextType, GraphicsType, ParticlesType:
				false;
			case TweenType:
				throw "cannot init DebugRect on tween";
		}
	}

	public function isDebugEventCallback(callback:(e:Event) -> Void):Bool {
		final events:Array<(e:Event) -> Void> = [
			(@:privateAccess Event.onDragActivator)
		];
		return events.contains(callback);
	}

	public function destroy():Void {
		debugRect.off();
		for (obj in debugRect.children) obj.off();
		final g:Graphics = cast debugRect.children.find(obj -> obj is Graphics);
		g?.begin();
		debugRect.removeAll();
		debugRect.parent.remove(debugRect);

		for (target in disabledInteractives) {
			final disabledInteractive = target.state.debug.disabledInteractive ?? false;
			if (disabledInteractive) Reflect.deleteField(target.state, "disabledInteractive");
			target.isInteractive = false;
		}
		disabledInteractives.resize(0);
	}

	function initSprite():Container {
		final sprite = target;
		final debugRect = overlay.addContainer();
		debugRect.name = "DebugRect";
		DebugRectOverlay.ignoreObject(debugRect);
		sprite.state.debug.debugRect = debugRect;
		debugRect.on(PointerDown, e -> {
			if (App.keys.get(Alt) && e.p.button == Left) {
				e.stopImmediatePropagation();
				DebugRectOverlay.gotoDebugPosition(sprite);
			} else {
				e.propagate = true;
			}
		});
		initSpriteMovePoint(debugRect);
		final points = [
			for (i in 0...4) new Graphics()
		];
		final posRectG = new Graphics();
		debugRect.add(posRectG);
		debugRect.on(KeyDown, e -> {
			if (!mainRectHotkeys(e)) {
				e.propagate = true;
			}
		});

		for (g in points) {
			g.width = 25;
			g.height = 25;
			debugRect.add(g);
			initSpriteResizePoint(g);
		}

		debugRect.on(Resize, e -> {
			final p = sprite.getGlobalPosition();
			final x = p.x;
			final y = p.y;
			final w = p.w;
			final h = p.h;
			debugRect.x = x + w / 2;
			debugRect.y = y + h / 2;
			debugRect.width = w;
			debugRect.height = h;

			final g = posRectG;
			g.clear();
			g.color = 0xFF2882B0;
			g.drawRect(-w / 2, -h / 2, w, h, 2);

			renderResizePoint(points[0], -w / 2, -h / 2);
			renderResizePoint(points[1], -w / 2, h / 2);
			renderResizePoint(points[2], w / 2, h / 2);
			renderResizePoint(points[3], w / 2, -h / 2);
		});
		Event.emitResize(debugRect);
		return debugRect;
	}

	function mainRectHotkeys(e:KeyEvent):Bool {
		switch e.key {
			case P:
				e.stopImmediatePropagation();
				DebugRectOverlay.gotoDebugPosition(target);
			case I:
				trace(target);
				traceObjectEvents(target);
			case Backspace:
				target.parent.remove(target);
				if (target.type.isContainerLike()) {
					final target:Container = cast target;
					target.children.flatIter(obj -> {
						obj.off();
					});
				}
				target.off();
				DebugRectOverlay.destroy();
			case OpenBracket:
				final child = nextNonIgnoredSibling(target, -1);
				if (child != target) DebugRectOverlay.toggle(child);
			case CloseBracket:
				final child = nextNonIgnoredSibling(target, 1);
				if (child != target) DebugRectOverlay.toggle(child);
			case Return:
				if (!target.type.isContainerLike()) return true;
				final obj:Container = cast target;
				final first = obj.children[0] ?? return true;
				DebugRectOverlay.toggle(first);
			case Escape:
				final parent = target.parent ?? return true;
				if (parent.type == SceneType && parent.parent == null) {
					DebugRectOverlay.destroy();
					return true;
				}
				DebugRectOverlay.toggle(parent);
			case _:
				return false;
		}
		return true;
	}

	function nextNonIgnoredSibling(target:IGameObject, dir:Int):IGameObject {
		final children = target.parent.children;
		var i = children.indexOf(target);
		for (_ in 0...children.length * 2) {
			i += dir;
			i = i.wrapAround(0, children.length - 1);
			if (!DebugRectOverlay.isIgnoredObject(children[i])) break;
		}
		return children[i];
	}

	function traceObjectEvents(target:IGameObject):Void {
		trace("state:", target.state);
		var firstEvent = true;
		final kinds = Event.getKindsOf(target);
		for (kind in kinds) {
			final items = Event.getEventItems(target, kind);
			for (item in items) {
				for (i => cb in item.callbacks) {
					if (isDebugEventCallback(cb)) continue;
					if (firstEvent) {
						trace("events:");
						firstEvent = false;
					}
					final pos = item.pos[i];
					final textPos = '${pos.fileName}:${pos.lineNumber}';
					trace(kind.toString(), textPos, cb);
				}
			}
		}
		var firstDebugEvent = true;
		for (kind in kinds) {
			final items = Event.getEventItems(target, kind);
			for (item in items) {
				for (i => cb in item.callbacks) {
					if (!isDebugEventCallback(cb)) continue;
					if (firstDebugEvent) {
						trace("debug events:");
						firstDebugEvent = false;
					}
					final pos = item.pos[i];
					final textPos = '${pos.fileName}:${pos.lineNumber}';
					trace(kind.toString(), textPos, cb);
				}
			}
		}
	}

	function initSpriteMovePoint(debugRect:Container):Void {
		final sprite = target;
		debugRect.on(Drag, spriteOnDrag);
		debugRect.on(DragEnd, spriteOnDragEnd);
		// to support single click copy too
		// delay to skip activation after double click selection
		overlay.delayed(1, () -> {
			debugRect.on(PointerUp, spriteOnPointerUp);
		});
		debugRect.on(KeyDown, spriteOnKeyDown);
	}

	function spriteOnDrag(e:DragEvent):Void {
		if (App.keys.get(Alt)) return;
		spriteUpdatePos(e);
	}

	function spriteOnDragEnd(e:DragEvent):Void {
		if (App.keys.get(Alt)) return;
		spriteUpdatePos(e);
		copySpritePos(target);
	}

	function spriteOnPointerUp(e:PointerEvent):Void {
		if (App.keys.get(Alt)) return;
		// already copied in DragEnd
		if (e.p.isMoved) return;
		copySpritePos(target);
	}

	function spriteOnKeyDown(e:KeyEvent):Void {
		final sprite = target;
		final align = switch e.key {
			case Zero: Alignement.Center;
			case One: Alignement.TopLeft;
			case Two: Alignement.TopCenter;
			case Three: Alignement.TopRight;
			case Four: Alignement.CenterLeft;
			case Five: Alignement.Center;
			case Six: Alignement.CenterRight;
			case Seven: Alignement.BottomLeft;
			case Eight: Alignement.BottomCenter;
			case Nine: Alignement.BottomRight;
			case _:
				e.propagate = true;
				return;
		}
		e.stopImmediatePropagation();
		sprite.state.debug.debugChangedOrigin = true;
		sprite.setOrigin(align.x, align.y);
		Event.emitResize(debugRect);
	}

	function spriteUpdatePos(e:DragEvent):Void {
		final sprite = target;
		final scene = target.findScene();
		final pos = sprite.parent.toLocalPos(e.dragX, e.dragY);
		sprite.x = pos.x;
		sprite.y = pos.y;
		Event.emitResize(sprite.parent);
		Event.emitResize(debugRect);
	}

	function copySpritePos(sprite:IGameObject):Void {
		var x = sprite.x;
		var y = sprite.y;
		var offX = 0.0;
		var offY = 0.0;
		final parent = sprite.parentContainer();
		final useParentOffset:Bool = sprite.state.debug.useParentOffset ?? true;
		if (parent != null && useParentOffset) {
			// updateContainerSize(parent);
			if (parent.state.debug.offX != null) {
				offX = parent.state.debug.offX;
				offY = parent.state.debug.offY;
			}
		}
		x = Math.round(x - offX);
		y = Math.round(y - offY);
		var name = @:privateAccess Lazyr.genVarName(sprite);
		var text = '\n$name.x = $x;\n$name.y = $y;';
		final withScale = sprite.state.debug.debugChangedScale ?? false;
		if (withScale) {
			final newScale = Std.int(sprite.scale * 100) / 100;
			text += '\n$name.scale = $newScale;';
		}
		final withOrigin = sprite.state.debug.debugChangedOrigin ?? false;
		if (withOrigin) {
			final x = Std.int(sprite.originX * 100) / 100;
			final y = Std.int(sprite.originY * 100) / 100;
			text += '\n$name.setOrigin($x, $y);';
		}
		Clipboard.copy(text);
		trace('$x, $y');
	}

	function initSpriteResizePoint(point:Graphics):Void {
		function updateSize(e:DragEvent):Void {
			final sprite = target;
			final rect = sprite.getGlobalPosition();
			final rectX = rect.x + rect.w * sprite.originX;
			final rectY = rect.y + rect.h * sprite.originY;
			var width = Math.abs(e.pointer.x - rectX) / sprite.parent.scaleX;
			var height = Math.abs(e.pointer.y - rectY) / sprite.parent.scaleY;
			if (width < 1) width = 1;
			if (height < 1) height = 1;
			sprite.displayWidth = width;
			sprite.displayHeight = height;
			// origin 0 = 0.5, 0.1 = 0.1, 0.5 = 0, 0.9 = 0.4, 1 = 0.5
			final diffX = sprite.originX < 0.5 ? 0.5 - sprite.originX : sprite.originX - 0.5;
			final diffY = sprite.originY < 0.5 ? 0.5 - sprite.originY : sprite.originY - 0.5;
			sprite.displayWidth = width * 2 * (1 - diffX);
			sprite.displayHeight = height * 2 * (1 - diffY);
			sprite.scale = Math.max(sprite.scaleX, sprite.scaleY);

			Event.emitResize(debugRect);
		}
		point.on(Drag, e -> {
			updateSize(e);
		});
		point.on(DragEnd, e -> {
			target.state.debug.debugChangedScale = true;
			updateSize(e);
			copySpritePos(target);
		});
	}

	function initPosContainer():Container {
		final sprite:Container = cast target;
		final debugRect = overlay.addContainer();
		debugRect.name = "DebugRect";
		DebugRectOverlay.ignoreObject(debugRect);
		sprite.state.debug.debugRect = debugRect;
		debugRect.on(PointerDown, e -> {
			if (App.keys.get(Alt) && e.p.button == Left) {
				e.stopImmediatePropagation();
				DebugRectOverlay.gotoDebugPosition(sprite);
			}
		});
		initGlobalMovePoint(debugRect);
		final points = [
			for (i in 0...4) new Graphics()
		];
		final posRectG = new Graphics();
		debugRect.add(posRectG);
		debugRect.on(KeyDown, e -> {
			if (!mainRectHotkeys(e)) {
				e.propagate = true;
			}
		});

		for (g in points) {
			g.width = 25;
			g.height = 25;
			debugRect.add(g);
			initGlobalResizePoint(g);
		}

		debugRect.on(Resize, e -> {
			final data = getPosContainerData(sprite);
			final scene = sprite.findScene();
			final rect = scene.isPortrait ? data.portrait : data.landscape;
			final align = rect.align ?? data.align ?? Center;
			final offX = rect.w * (0.5 - align.x) * scene.sceneW;
			final offY = rect.h * (0.5 - align.y) * scene.sceneH;
			final x = rect.x * scene.sceneW + offX;
			final y = rect.y * scene.sceneH + offY;
			final w = rect.w * scene.sceneW;
			final h = rect.h * scene.sceneH;
			debugRect.x = x;
			debugRect.y = y;
			debugRect.width = w;
			debugRect.height = h;
			final gpos = debugRect.toGlobal(scene.x, scene.y);
			final pos = overlay.toLocalPos(gpos.x, gpos.y);
			final x = pos.x;
			final y = pos.y;
			final w = gpos.w;
			final h = gpos.h;
			debugRect.x = pos.x;
			debugRect.y = pos.y;
			debugRect.width = gpos.w;
			debugRect.height = gpos.h;

			final g = posRectG;
			g.begin();
			g.clear();
			g.color = 0x3000AAFF;
			sprite.updateTransformation();
			final b = sprite.getGlobalPosition();
			g.fillRect(
				b.x - pos.x + scene.camera.x,
				b.y - pos.y + scene.camera.y,
				b.w, b.h
			);
			g.color = 0xFF00AAFF;
			g.drawRect(-w / 2, -h / 2, w, h, 2);

			g.drawString('${rect.x.toFixed(2)},${rect.y.toFixed(2)}', b.x + 6, b.y + 6);
			g.end();

			renderResizePoint(points[0], -w / 2, -h / 2);
			renderResizePoint(points[1], -w / 2, h / 2);
			renderResizePoint(points[2], w / 2, h / 2);
			renderResizePoint(points[3], w / 2, -h / 2);
		});

		Event.emitResize(debugRect);
		return debugRect;
	}

	/** Returns percentage cords relative to target scene. **/
	function getPosContainerData(sprite:Container):PosObjectConfig {
		final scene = sprite.findScene();
		var data:PosObjectConfig = sprite.state.debug.posContainerData;
		if (data == null) {
			data = {
				portrait: {
					x: sprite.x / scene.sceneW,
					y: sprite.y / scene.sceneH,
					w: sprite.displayWidth / scene.sceneW,
					h: sprite.displayHeight / scene.sceneH,
				}
			};
		}
		if (!scene.isPortrait) data.landscape ??= {
			x: data.portrait.x,
			y: data.portrait.y,
			w: data.portrait.w,
			h: data.portrait.h
		};
		return data;
	}

	function initGlobalMovePoint(debugRect:Container):Void {
		function updateResizeCallback(sprite:Container, data:PosObjectConfig):Void {
			Lazyr.removeResizeCallback(sprite);
			Lazyr.setResizeCallback(sprite, sprite -> {
				Lazyr.posContainer(sprite, data);
			});
			Event.emitResize(debugRect);
		}

		function updatePos(e:DragEvent):Void {
			final debugRect:Container = cast e.target;
			final sprite:Container = cast target;
			final scene = sprite.findScene();
			final data = getPosContainerData(sprite);
			final rect = scene.isPortrait ? data.portrait : data.landscape;
			final align = rect.align ?? data.align ?? Center;
			final offX = debugRect.displayWidth * (0.5 - align.x);
			final offY = debugRect.displayHeight * (0.5 - align.y);
			final pos = scene.toLocalPos(e.dragX, e.dragY);
			rect.x = (pos.x - offX) / scene.sceneW;
			rect.y = (pos.y - offY) / scene.sceneH;
			sprite.state.debug.posContainerData = data;
			updateResizeCallback(sprite, data);
		}
		debugRect.on(Drag, e -> {
			if (App.keys.get(Alt)) return;
			updatePos(e);
		});

		debugRect.on(DragEnd, e -> {
			if (App.keys.get(Alt)) return;
			updatePos(e);
			final sprite:Container = cast target;
			copyContainerPosConfig(sprite);
		});
		// to support single click copy too
		overlay.delayed(1, () -> {
			debugRect.on(PointerUp, e -> {
				if (App.keys.get(Alt)) return;
				final sprite:Container = cast target;
				copyContainerPosConfig(sprite);
			});
		});

		debugRect.on(KeyDown, e -> {
			final sprite:Container = cast target;
			final align = switch e.key {
				case Zero: Alignement.Center;
				case One: Alignement.TopLeft;
				case Two: Alignement.TopCenter;
				case Three: Alignement.TopRight;
				case Four: Alignement.CenterLeft;
				case Five: Alignement.Center;
				case Six: Alignement.CenterRight;
				case Seven: Alignement.BottomLeft;
				case Eight: Alignement.BottomCenter;
				case Nine: Alignement.BottomRight;
				case _:
					e.propagate = true;
					return;
			}
			e.stopImmediatePropagation();
			final data = getPosContainerData(sprite);
			final scene = sprite.findScene();
			final rect = scene.isPortrait ? data.portrait : data.landscape;
			rect.align = align;
			updateResizeCallback(sprite, data);
		});
	}

	function initGlobalResizePoint(point:Graphics):Void {
		function updateSize(e:DragEvent):Void {
			final sprite:Container = cast target;
			final scene = sprite.findScene();
			var width = Math.abs(e.pointer.x - sprite.x - scene.x) * 2;
			var height = Math.abs(e.pointer.y - sprite.y - scene.y) * 2;
			if (width < 1) width = 1;
			if (height < 1) height = 1;

			final data = getPosContainerData(sprite);
			final rect = scene.isPortrait ? data.portrait : data.landscape;
			final align = rect.align ?? data.align ?? Center;
			final offX = sprite.displayWidth * (0.5 - align.x);
			final offY = sprite.displayHeight * (0.5 - align.y);
			rect.w = width / scene.sceneW + offX / scene.sceneW;
			rect.h = height / scene.sceneH + offY / scene.sceneH;

			sprite.state.debug.posContainerData = data;
			Lazyr.removeResizeCallback(sprite);
			Lazyr.setResizeCallback(sprite, sprite -> {
				Lazyr.posContainer(sprite, data);
			});
			Event.emitResize(debugRect);
		}
		point.on(Drag, e -> {
			updateSize(e);
		});
		point.on(DragEnd, e -> {
			updateSize(e);
			final sprite:Container = cast target;
			copyContainerPosConfig(sprite);
		});
	}

	function renderResizePoint(g:Graphics, x:Float, y:Float):Void {
		final m = g.parent.transformation;
		final scale = m._00;
		g.x = x - 0.5 / scale;
		g.y = y - 0.5 / scale;
		g.clear();
		g.color = 0xFFFFFFFF;
		final size = 12 / scale;
		g.fillRect(-size / 2, -size / 2, size, size);
		g.color = 0xFF000000;
		g.drawRect(-size / 2, -size / 2, size, size);
	}

	function copyContainerPosConfig(sprite:Container):Void {
		final data = getPosContainerData(sprite);
		inline function getAlignText(rect:PosRect):String {
			final align = rect.align ?? data.align ?? null;
			return align == null ? "" : 'align: ${align.toString()},';
		}
		inline function getPaddingText(rect:PosRect):String {
			final padding = rect.padding ?? data.padding ?? null;
			return padding == null ? "" : 'padding: $padding,';
		}
		inline function toFixed(v:Float) {
			return Std.int(v * 100) / 100;
		}
		final landText = data.landscape == null ? "" : '
									landscape: {
										${getAlignText(data.landscape)}
										${getPaddingText(data.landscape)}
										x: ${toFixed(data.landscape.x)},
										y: ${toFixed(data.landscape.y)},
										w: ${toFixed(data.landscape.w)},
										h: ${toFixed(data.landscape.h)}
									}
			'.trim();
		var text = '
				addContainer({
							onResize: sprite -> {
								posContainer(sprite, {
									portrait: {
										${getAlignText(data.portrait)}
										${getPaddingText(data.portrait)}
										x: ${toFixed(data.portrait.x)},
										y: ${toFixed(data.portrait.y)},
										w: ${toFixed(data.portrait.w)},
										h: ${toFixed(data.portrait.h)}
									},
									$landText
								});
							}
						});
			';
		text = text.align();
		text = text.removeEmptyLines();
		text = text.trim();
		Clipboard.copy(text);
		final posText = sprite.posInfosLine();
		if (posText != null) {
			trace(posText, text);
		} else {
			final name = @:privateAccess Lazyr.genVarName(sprite);
			final nameText = 'final $name = ...';
			trace(nameText, text);
		}
	}
}
