package khm.utils;

// https://github.com/photonstorm/phaser/tree/master/src/math/easing
class Easing {
	public static function linear(v:Float):Float {
		return v;
	}

	public static function sineIn(v:Float):Float {
		if (v == 0) {
			return 0;
		} else if (v == 1) {
			return 1;
		} else {
			return 1 - Math.cos(v * Math.PI / 2);
		}
	}

	public static function sineOut(v:Float):Float {
		if (v == 0) {
			return 0;
		} else if (v == 1) {
			return 1;
		} else {
			return Math.sin(v * Math.PI / 2);
		}
	}

	public static function sineInOut(v:Float):Float {
		if (v == 0) {
			return 0;
		} else if (v == 1) {
			return 1;
		} else {
			return 0.5 * (1 - Math.cos(Math.PI * v));
		}
	};

	static final backOvershoot = 1.70158;

	public static function backIn(v:Float):Float {
		return v * v * ((backOvershoot + 1) * v - backOvershoot);
	}

	public static function backOut(v:Float):Float {
		return --v * v * ((backOvershoot + 1) * v + backOvershoot) + 1;
	}

	public static function backInOut(v:Float):Float {
		final s = backOvershoot * 1.525;

		if ((v *= 2) < 1) {
			return 0.5 * (v * v * ((s + 1) * v - s));
		} else {
			return 0.5 * ((v -= 2) * v * ((s + 1) * v + s) + 2);
		}
	}

	public static function bounceIn(v:Float):Float {
		v = 1 - v;
		if (v < 1 / 2.75) {
			return 1 - (7.5625 * v * v);
		} else if (v < 2 / 2.75) {
			return 1 - (7.5625 * (v -= 1.5 / 2.75) * v + 0.75);
		} else if (v < 2.5 / 2.75) {
			return 1 - (7.5625 * (v -= 2.25 / 2.75) * v + 0.9375);
		} else {
			return 1 - (7.5625 * (v -= 2.625 / 2.75) * v + 0.984375);
		}
	}

	public static function bounceOut(v:Float):Float {
		if (v < 1 / 2.75) {
			return 7.5625 * v * v;
		} else if (v < 2 / 2.75) {
			return 7.5625 * (v -= 1.5 / 2.75) * v + 0.75;
		} else if (v < 2.5 / 2.75) {
			return 7.5625 * (v -= 2.25 / 2.75) * v + 0.9375;
		} else {
			return 7.5625 * (v -= 2.625 / 2.75) * v + 0.984375;
		}
	}

	public static function bounceInOut(v:Float):Float {
		var reverse = false;
		if (v < 0.5) {
			v = 1 - (v * 2);
			reverse = true;
		} else {
			v = (v * 2) - 1;
		}
		if (v < 1 / 2.75) {
			v = 7.5625 * v * v;
		} else if (v < 2 / 2.75) {
			v = 7.5625 * (v -= 1.5 / 2.75) * v + 0.75;
		} else if (v < 2.5 / 2.75) {
			v = 7.5625 * (v -= 2.25 / 2.75) * v + 0.9375;
		} else {
			v = 7.5625 * (v -= 2.625 / 2.75) * v + 0.984375;
		}
		if (reverse) {
			return (1 - v) * 0.5;
		} else {
			return v * 0.5 + 0.5;
		}
	}

	public static function cubicIn(v:Float):Float {
		return v * v * v;
	}

	public static function cubicOut(v:Float):Float {
		return --v * v * v + 1;
	}

	public static function cubicInOut(v:Float):Float {
		if ((v *= 2) < 1) {
			return 0.5 * v * v * v;
		} else {
			return 0.5 * ((v -= 2) * v * v + 2);
		}
	}

	public static function quadIn(v:Float):Float {
		return v * v;
	}

	public static function quadOut(v:Float):Float {
		return v * (2 - v);
	}

	public static function quadInOut(v:Float):Float {
		if ((v *= 2) < 1) {
			return 0.5 * v * v;
		} else {
			return -0.5 * (--v * (v - 2) - 1);
		}
	}

	public static function circIn(v:Float):Float {
		return 1 - Math.sqrt(1 - v * v);
	}

	public static function circOut(v:Float):Float {
		return Math.sqrt(1 - (--v * v));
	}

	public static function circInOut(v:Float):Float {
		if ((v *= 2) < 1) {
			return -0.5 * (Math.sqrt(1 - v * v) - 1);
		} else {
			return 0.5 * (Math.sqrt(1 - (v -= 2) * v) + 1);
		}
	}

	public static function quartIn(v:Float):Float {
		return v * v * v * v;
	}

	public static function quartOut(v:Float):Float {
		return 1 - (--v * v * v * v);
	}

	public static function quartInOut(v:Float):Float {
		if ((v *= 2) < 1) {
			return 0.5 * v * v * v * v;
		} else {
			return -0.5 * ((v -= 2) * v * v * v - 2);
		}
	}

	public static function expoIn(v:Float):Float {
		return Math.pow(2, 10 * (v - 1)) - 0.001;
	}

	public static function expoOut(v:Float):Float {
		return 1 - Math.pow(2, -10 * v);
	}

	public static function expoInOut(v:Float):Float {
		if ((v *= 2) < 1) {
			return 0.5 * Math.pow(2, 10 * (v - 1));
		} else {
			return 0.5 * (2 - Math.pow(2, -10 * (v - 1)));
		}
	}

	public static function quintIn(v:Float):Float {
		return v * v * v * v * v;
	}

	public static function quintOut(v:Float):Float {
		return --v * v * v * v * v + 1;
	}

	public static function quintInOut(v:Float):Float {
		if ((v *= 2) < 1) {
			return 0.5 * v * v * v * v * v;
		} else {
			return 0.5 * ((v -= 2) * v * v * v * v + 2);
		}
	}
}
