package khm.utils;

import khm.tools.RegexTools;

class FrameUtils {
	static final matchNum = ~/[0-9]+/g;

	public static function numericIdSort(id:ResItem, id2:ResItem):Int {
		return numericStringSort(id.frame, id2.frame);
	}

	public static function numericStringSort(s1:String, s2:String):Int {
		if (matchNum.match(s1) && matchNum.match(s2)) {
			final m1 = RegexTools.getMatches(matchNum, s1);
			final m2 = RegexTools.getMatches(matchNum, s2);
			for (i in 0...m1.length) {
				if (m1[i] == null || m2[i] == null) continue;
				final n1 = Std.parseInt(m1[i]);
				final n2 = Std.parseInt(m2[i]);
				if (n1 == n2) continue;
				return n1 < n2 ? -1 : 1;
			}
		}
		return s1 < s2 ? -1 : 1;
	}

	public static function getUnusedFrames(ignoreNumSequences = true):Array<ResItem> {
		final matchNumEnd = ~/[0-9]+$/g;
		final matchChars = ~/[^0-9A-z_]/g;
		final startsWithNumber = ~/^[0-9]/g;
		final textures:Array<String> = [];
		final values:Array<ResItem> = [];
		#if kha
		throw "not impl";
		#elseif phaser
		App._scene.textures.each((e, args) -> {
			final name:String = e.key;
			if (!name.startsWith("res")) return;
			textures.push(name);
			final texture = App._scene.textures.get(name);
			for (field in Reflect.fields(texture.frames)) {
				if (field == "__BASE") continue;

				if (ignoreNumSequences && matchNumEnd.match(field)) continue;
				field = matchChars.replace(field, "_").toLowerCase();
				if (startsWithNumber.match(field)) field = '_$field';

				values.push({
					frame: field,
					texture: name
				});
			}
		}, null);
		for (texture in textures) {
			final fields = Type.getClassFields(Type.resolveClass(texture.toUpperCase()));
			for (field in fields) {
				final item = values.find(item -> {
					return item.frame == field && item.texture == texture;
				});
				values.remove(item);
			}
		}
		#end
		return values;
	}

	public static function checkResItem(id:ResItem):Void {
		#if kha
		if (id.frame != "") {
			if (!Atlas.textures.contains(id.texture)) {
				throw 'Texture "${id.texture}" not found (for frame "${id.frame}")';
			}
			final frames = Atlas.atlases[id.texture];
			if (!frames.inlineExists(item -> item.filename == id.frame)) {
				throw 'Frame "${id.frame}" not found in texture "${id.texture}"';
			}
		} else if (id.texture != null) {
			if (!Atlas.textures.contains(id.texture)) {
				throw 'Texture "${id.texture}" not found (for empty frame)';
			}
		}
		#elseif phaser
		if (id.frame != "") {
			if (!App._scene.textures.exists(id.texture)) {
				throw 'Texture "${id.texture}" not found (for frame "${id.frame}")';
			}
			if (!App._scene.textures.get(id.texture).has(id.frame)) {
				throw 'Frame "${id.frame}" not found in texture "${id.texture}"';
			}
		} else if (id.texture != null) {
			if (!App._scene.textures.exists(id.texture)) {
				throw 'Texture "${id.texture}" not found (for empty frame)';
			}
		}
		#end
	}
}
