package khm.utils;

import khm.ResItem;
import khm.Scene;
import khm.Tween;
import khm.Types.Point;

using khm.tools.GameObjectAnimTools;

private typedef ShowPointsConfig = {
	?delay:Float,
	?scale:Float
}

class Hand {
	var handTween:Tween;
	var handTargetId = 0;

	public var handTargets:Array<IGameObject> = [];
	public var handPoints:Array<Point> = [];

	@:noCompletion
	@:deprecated("Use `sprite` instead")
	public var hand(get, never):Sprite;

	inline function get_hand():Sprite {
		return sprite;
	}

	public final sprite:Sprite;

	public function new(id:ResItem, ?parent:Container) {
		sprite = new Sprite(id);
		sprite.setOrigin(0.15, 0.1);
		sprite.alpha = 0;
		if (parent != null) parent.add(sprite);
	}

	public function getTarget():IGameObject {
		return handTargets[handTargetId];
	}

	public function showTargets(?opts:{
		?withTargetScaleAnim:Bool,
		?duration:Float,
		?delay:Float,
		?onUpdate:(tween:TweenEvent) -> Void,
		?angle:Float,
	}):Void {
		if (handTargets.length == 0) throw "handTargets.length is 0";
		for (i => object in handTargets) {
			if (object == null) throw 'handTargets[$i] is null';
		}
		if (opts == null) opts = {};
		opts.withTargetScaleAnim ??= true;
		opts.duration ??= 1200;
		opts.delay ??= 0;
		opts.angle ??= 45;

		handTargetId = 0;
		if (handTween != null) handTween.stop();
		handTween = Tween.counter({
			targets: sprite,
			from: 0,
			to: 1,
			delay: opts.delay,
			duration: opts.duration,
			ease: Ease.SineInOut,
			repeat: -1,
			yoyo: true,
			onUpdate: tween -> {
				final target = handTargets[handTargetId];
				if (target == null) {
					trace("handTargets:", handTargets);
					throw 'handTargets[$handTargetId] is null';
				}
				if (opts.withTargetScaleAnim) {
					if (target.getData("initScale") == null) {
						target.setData("initScale", target.scale);
					}
				}
				sprite.scale = target.scale;
				final value = 1 - tween.getValue();
				final radius = target.displayWidth / 6 * value;
				final offX = opts.angle.toRad().cos() * radius;
				final offY = opts.angle.toRad().sin() * radius;
				sprite.x = target.x + offX;
				sprite.y = target.y + offY;
				if (opts.withTargetScaleAnim) {
					final targetScale:Float = target.getData("initScale") ?? 1;
					target.scale = targetScale + 0.07 * tween.getValue();
				}
				var alpha = (1 - value) * 3;
				if (alpha > 1) alpha = 1;
				sprite.alpha = alpha;
				if (opts.onUpdate != null) opts.onUpdate(tween);
			},
			onRepeat: () -> {
				handTargetId++;
				if (handTargetId >= handTargets.length) handTargetId = 0;
			}
		});
	}

	public function showPoints(?obj:ShowPointsConfig):Void {
		if (handPoints.length == 0) throw "handPoints.length is 0";
		obj ??= {};
		obj.delay ??= 0;
		obj.scale ??= 1;
		handTargetId = 0;
		final target = handPoints[handTargetId];
		sprite.x = target.x;
		sprite.y = target.y;
		sprite.scale = 1.2 * obj.scale;
		if (handTween != null) handTween.stop();
		handTween = sprite.addTween({
			delay: obj.delay,
			duration: 500,
			scale: 1 * obj.scale,
			alpha: 1,
			onComplete: () -> {
				runPointHand(obj);
			}
		});
	}

	function runPointHand(obj:ShowPointsConfig):Void {
		handTargetId++;
		if (handTargetId >= handPoints.length) handTargetId = 0;
		final target = handPoints[handTargetId];
		// hide hand and restart point cycle
		if (target == null) {
			handTween = sprite.hideAnim(500, () -> {
				handTargetId++;
				if (handTargetId >= handPoints.length) handTargetId = 0;
				final target = handPoints[handTargetId];
				sprite.x = target.x;
				sprite.y = target.y;
				sprite.scale = 1.2 * obj.scale;
				handTween = sprite.addTween({
					duration: 300,
					scale: 1 * obj.scale,
					alpha: 1,
					onComplete: () -> runPointHand(obj)
				});
			});
			return;
		}
		handTween = sprite.addTween({
			scale: 1 * obj.scale,
			x: target.x,
			y: target.y,
			duration: 1000,
			onComplete: () -> runPointHand(obj)
		});
	}

	public function hide(duration = 300, ?onComplete:() -> Void):Tween {
		Lazyr.sendAnalytics("Tutor Stop");
		for (target in handTargets) {
			final target:Container = cast target;
			if (target.getData("initScale") == null) continue;
			target.scale = target.getData("initScale");
			target.setData("initScale", null);
		}
		if (handTween != null) handTween.stop();
		handTween = sprite.hideAnim(duration, onComplete);
		return handTween;
	}
}
