package khm.utils;

import haxe.Json;
import haxe.crypto.Base64;
import haxe.io.Bytes;
import khm.utils.macro.Macro;
#if js
import js.Browser;
import js.html.WebSocket;
#end

enum abstract WsEventType(String) {
	var Connected;
	var WriteFile;
	var WriteTextFile;
}

typedef WsEvent = {
	type:WsEventType,
	?connected:{},
	?writeFile:{
		path:String,
		base64:String,
	},
	?writeTextFile:{
		path:String,
		text:String,
	}
}

/**
	Client for khm local server tool to send commands from web build to nodejs.

	For example, you can update files in project folder with `Observer.send`.
**/
class Observer {
	#if js
	static var ws:WebSocket;
	#end

	public static function connect():Void {
		#if (js && kha_live_reload)
		var host = Browser.location.hostname;
		if (host == "") host = "localhost";
		// var port = Std.parseInt(Browser.location.port) ?? 80;
		final port = Macro.getDefine("khm_observer_port") ?? '${8080 + 2}';
		ws = new WebSocket('ws://$host:$port');
		ws.onopen = () -> {
			// trace("onopen");
		}
		// ws.onmessage = () -> Browser.location.reload();
		ws.onmessage = e -> {
			final data:{} = Json.parse(e.data);
			// trace("onmessage", data);
		}
		ws.onerror = e -> {
			trace(e);
		}
		#end
	}

	public static function send(data:WsEvent):Void {
		#if (js && kha_live_reload)
		if (ws == null) return;
		ws.send(Json.stringify(data));
		#end
	}

	public static function writeFile(path:String, bytes:Bytes):Void {
		final data:WsEvent = {
			type: WriteFile,
			writeFile: {
				path: path,
				base64: Base64.encode(bytes)
			}
		}
		send(data);
	}

	public static function writeTextFile(path:String, text:String):Void {
		final data:WsEvent = {
			type: WriteTextFile,
			writeTextFile: {
				path: path,
				text: text
			}
		}
		send(data);
	}

	public static function writeJson(path:String, obj:Any, pretty = false):Void {
		final data:WsEvent = {
			type: WriteTextFile,
			writeTextFile: {
				path: path,
				text: Json.stringify(obj, pretty ? "\t" : null)
			}
		}
		send(data);
	}
}
