package khm.utils;

#if js
import js.Browser.window;
#end

class ParamsUtils {
	public static function read<T>(vars:Array<T>, activeVar:Int):T {
		#if debug
		fillParams(vars);
		validateParams(vars);
		return vars[activeVar - 1];
		#elseif js
		fillParams(windowVarsObjectToArray());
		// get strict type, then cast Dynamic to it
		var type = vars[activeVar - 1];
		type = readWindowParams();
		return type;
		#end
		return vars[activeVar - 1];
	}

	#if js
	static function windowVarsObjectToArray():Array<Any> {
		final params = (window : Dynamic).params;
		final names = Reflect.fields(params).filter(s -> s.startsWith("var"));
		return names.map(name -> Reflect.field(params, name));
	}
	#end

	#if js
	/**
		Reads variant object from `window.params["var" + window.params.activeVar]`
		and do bounds checks for `{value, min, max}` fields
	**/
	public static function readWindowParams():Dynamic {
		var windowParams = (window : Dynamic).params;
		final type = Reflect.field(windowParams, "activeVar");
		if (type == null) {
			trace("activeVar not found in params");
			return {};
		}
		final params = Reflect.field(windowParams, 'var$type');
		if (params == null) {
			trace('var$type not found in params');
			return params;
		}
		validateParams(params);
		return params;
	}

	@:noCompletion public static function readValidParams(?_):Dynamic
		return readWindowParams();
	#end

	public static function validateParams<T>(params:Array<T>):Void {
		for (i => param in params) {
			validateParam(i, param, params[0]);
		}
	}

	public static function fillParams(vars:Array<Any>) {
		if (!(vars is Array)) throw "vars should be array";
		final first = vars[0];
		for (i => param in vars) {
			if (param == null) continue;
			if (i == 0) continue;
			for (name in Reflect.fields(first)) {
				if (Reflect.hasField(param, name)) continue;
				final value = Reflect.field(first, name);
				Reflect.setField(param, name, value);
			}
		}
	}

	static function validateParam<T>(i:Int, params:T, firstVar:T):Void {
		if (params is Array) throw "param is array, but should be param item";
		for (name in Reflect.fields(params)) {
			final field = Reflect.field(params, name);
			if (!Reflect.isObject(field) || Std.isOfType(field, Array)) {
				Reflect.setField(params, name, field);
				continue;
			}
			if (field.value == null) {
				Reflect.setField(params, name, field);
				continue;
			}
			final firstRange = Reflect.field(firstVar, name);
			if (field.min != null) {
				if (field.min != firstRange.min) {
					trace('params[$i].$name.min (${field.min}) is different to params[0].$name.min (${firstRange.min})');
				}
				if (field.value < field.min) {
					trace('params[$i].$name.value (${field.value}) is below min value (${field.min})');
					field.value = field.min;
				}
			}
			if (field.max != null) {
				if (field.max != firstRange.max) {
					trace('params[$i].$name.max (${field.max}) is different to params[0].$name.max (${firstRange.max})');
				}
				if (field.value > field.max) {
					trace('params[$i].$name.value (${field.value}) is above max value (${field.max})');
					field.value = field.max;
				}
			}
			Reflect.setField(params, name, field);
		}
	}
}
