package khm.utils.macro;

import haxe.macro.Context;
import haxe.macro.Expr;
import haxe.macro.Printer;
import haxe.macro.Type;

using haxe.macro.Tools;

class EnumAbstractUtils {
	macro public static function fromString(name:Expr, typePath:Expr, ?defValue:Expr):Expr {
		final type = Context.getType(typePath.toString());
		switch (type.follow()) {
			case TAbstract(_.get() => ab, _) if (isEnumAbstract(ab)):
				final cases:Array<Case> = [];
				for (field in ab.impl.get().statics.get()) {
					if (isEnumAbstractField(field)) {
						cases.push({
							values: [macro $v{field.name}],
							expr: macro cast $i{field.name}
						});
					}
				}
				final e:Expr = {
					expr: ESwitch(name, cases, defValue),
					pos: Context.currentPos()
				}
				// trace(new haxe.macro.Printer().printExpr(e));
				return e;
			default:
				throw new Error(type.toString() + " should be enum abstract", typePath.pos);
		}
	}

	/**
		Generates switch of enum abstract values and returns enum constructor name string
		@param `enumAb` enum abstract constructor (enum instance) or `abstract` keyword
		@param `defaultValue` string for default switch case (`null` by default)
	**/
	macro public static function toString(enumAb:Expr, ?defaultValue:Expr):Expr {
		final type = Context.typeof(enumAb);
		switch (type.follow()) {
			case TAbstract(_.get() => ab, _) if (isEnumAbstract(ab)):
				final cases:Array<Case> = [];
				for (field in ab.impl.get().statics.get()) {
					if (isEnumAbstractField(field)) {
						cases.push({
							values: [macro $i{field.name}],
							expr: macro $v{field.name}
						});
					}
				}
				final e:Expr = {
					expr: ESwitch(enumAb, cases, defaultValue),
					pos: Context.currentPos()
				}
				// trace(new haxe.macro.Printer().printExpr(e));
				return e;
			default:
				throw new Error(type.toString() +
					" should be enum abstract instance", enumAb.pos);
		}
	}

	/**
		Generates switch of enum abstract values and returns enum constructor `Int` index
		@param `enumAb` enum abstract constructor (enum instance) or `abstract` keyword
		@param `defaultValue` `Int` value for default switch case (`null` by default)
	**/
	macro public static function getIndex(enumAb:Expr, ?defaultValue:Expr):Expr {
		final type = Context.typeof(enumAb);
		switch (type.follow()) {
			case TAbstract(_.get() => ab, _) if (isEnumAbstract(ab)):
				final cases:Array<Case> = [];
				var i = 0;
				for (field in ab.impl.get().statics.get()) {
					if (isEnumAbstractField(field)) {
						cases.push({
							values: [macro $i{field.name}],
							expr: macro $v{i}
						});
						i++;
					}
				}
				final e:Expr = {
					expr: ESwitch(enumAb, cases, defaultValue),
					pos: Context.currentPos()
				}
				// trace(new haxe.macro.Printer().printExpr(e));
				return e;
			default:
				throw new Error(type.toString() + " should be enum abstract", enumAb.pos);
		}
	}

	static function isEnumAbstract(ab:AbstractType):Bool {
		return ab.meta.has(":enum");
	}

	static function isEnumAbstractField(field:ClassField):Bool {
		return field.meta.has(":enum") && field.meta.has(":impl");
	}
}
