package khm.utils.macro;

import haxe.io.Path;
import haxe.macro.Context;
import haxe.macro.Expr;
import haxe.macro.Printer;

using haxe.macro.Tools;

class Macro {
	macro public static function getTypedObject(obj:Expr, typePath:Expr):Expr {
		final type = Context.getType(typePath.toString());
		switch (type.follow()) {
			case TAnonymous(_.get() => td):
				final name = obj.toString();
				if (obj.expr.match(EObjectDecl(_))) {
					throw new Error('$name should be passed as reference (inside of variable)', obj.pos);
				}
				final e:Expr = {
					expr: EObjectDecl([
						for (field in td.fields) {
							field: field.name,
							expr: macro $p{['$name', '${field.name}']},
						}
					]),
					pos: Context.currentPos()
				}
				return e;
			default:
				throw new Error(type.toString() + " should be typedef structure", typePath.pos);
		}
	}

	macro public static function getBuildTime():Expr {
		return macro $v{Date.now().toString()};
	}

	macro public static function getSrcDir():Expr {
		final srcPath = Context.getDefines().get("src_folder") ?? "src";
		#if kha
		// absolute path
		if (srcPath != null) return macro $v{srcPath};
		#end
		final cwd = Path.removeTrailingSlashes(Sys.getCwd());
		return macro $v{cwd + "/" + srcPath};
	}

	macro public static function getDefine(id:String):Null<Expr> {
		final def = Context.getDefines().get(id) ?? return macro null;
		return {
			expr: EConst(CString(def)),
			pos: Context.currentPos()
		};
	}

	macro public static function getDefines():Expr {
		return macro $v{Context.getDefines()};
	}

	/**
		Fills all Containers in local scope with their local variable names:
		```
		final fooC = addContainer(...);
		fooC.name = "fooC"; // generated by macro
		final game = new Container(...);
		game.name = "game"; // generated by macro
		```
	**/
	macro public static function fillLocalContainerNames():Expr {
		final exprs:Array<Expr> = [];
		for (item in Context.getLocalTVars()) {
			switch item.t {
				case TInst(t, _):
					// trace(t.get().name); can be faster
					if (t.toString() == "khm.Container") {
						final tvar = item.name;
						exprs.push(macro if ($i{tvar} != null) $i{tvar}.name ??= $v{tvar});
					}
				case _:
			}
		}
		return macro $b{exprs};
	}
}
