package khm.verlet;

import khm.math.Vector2;
import khm.verlet.Body.BodyOptions;

#if kha
using kha.graphics2.GraphicsExtension;
#end

class Circle extends Body {
	// public var radius:Float;
	// public var angle = 0.0;
	public var dot(get, never):Dot;
	public var rotation = 0.0;
	public var prevRotation = 0.0;

	// public var prevRotationVelocity = 0.0;
	public var radius(default, set):Float;

	function set_radius(radius:Float):Float {
		this.radius = radius;
		halfSize.x = radius;
		halfSize.y = radius;
		return radius;
	}

	inline function get_dot():Dot {
		return dots[0];
	}

	public function new(world:World, radius:Float, opts:BodyOptions) {
		@:bypassAccessor this.radius = radius ?? opts.radius;
		opts.type = Circle;
		super(world, opts);
	}

	override function updateBoundingBox():Void {
		final p = dot.pos;
		halfSize.x = radius;
		halfSize.y = radius;
		aabbCenter.x = p.x;
		aabbCenter.y = p.y;
		center.x = p.x;
		center.y = p.y;
	}

	override function projectAxis(axis:Vector2):Void {
		final d = dot.pos.dot(axis);
		min = d - radius;
		max = d + radius;
	}

	#if kha
	override function renderDots(g:kha.graphics2.Graphics) {}

	override function renderConstraints(g:kha.graphics2.Graphics):Void {
		if (!world.renderCircles) return;
		final p = dot.pos;
		final prevPos = dot.prevPos;
		final rest = p.x == prevPos.x && p.y == prevPos.y;
		g.color = rest ? 0xFFD8BABA : 0xFFFF8888;
		g.drawCircle(p.x, p.y, radius);
		if (rotation == 0) return;
		// final rad = angle.toRad();
		final rad = rotation;
		g.drawLine(
			p.x + Math.cos(rad) * (radius - 10),
			p.y + Math.sin(rad) * (radius - 10),
			p.x + Math.cos(rad) * (radius),
			p.y + Math.sin(rad) * (radius),
		);
	}
	#end
}
