package khm.verlet;

import khm.math.Degrees;
import khm.math.Vector2;

class Constraint {
	public final body:Null<Body>;
	public final dotA:Dot;
	public final dotB:Dot;
	public final isEdge:Bool;

	public var stiffness:Float;
	/** Initial distance between dots. **/
	public var dist:Float;
	public var angle(get, never):Degrees;
	public var ignoreMass = false;

	function get_angle():Degrees {
		return MathLib.angle.betweenPoints(dotA.pos, dotB.pos);
	}

	public function new(body:Null<Body>, dotA:Dot, dotB:Dot, stiffness = 1.0, isEdge = false) {
		if (dotA == dotB) throw "Cannot constrain a dot to itself";
		this.body = body;
		this.dotA = dotA;
		this.dotB = dotB;
		this.stiffness = stiffness;
		this.isEdge = isEdge;
		updateInitialDist();
	}

	public inline function updateInitialDist():Void {
		dist = getCurrentDist();
	}

	public inline function getCurrentDist():Float {
		return dotA.pos.sub(dotB.pos).length;
	}

	public function solve():Void {
		if (body?.fixed) return;
		final posA = dotA.pos;
		final posB = dotB.pos;
		// Calculate displacement
		final delta = posA.sub(posB);
		var actualLength = delta.length;
		if (actualLength == 0) actualLength = 0.000001;
		final targetLength = dist;
		final deltaLength = (targetLength - actualLength) * 0.5 * stiffness;
		final dispX = (delta.x / actualLength) * deltaLength;
		final dispY = (delta.y / actualLength) * deltaLength;

		// Calculate displacement ratios based on mass
		final bodyA = dotA.body;
		final bodyB = dotB.body;
		final massA = ignoreMass ? 1 : bodyA.mass;
		final massB = ignoreMass ? 1 : bodyB.mass;
		final invTotalMass = 1 / (massA + massB);
		final ratioA = dotB.fixed ? 1.0 : massB * invTotalMass;
		final ratioB = dotA.fixed ? 1.0 : massA * invTotalMass;

		if (!dotA.fixed) {
			posA.x += dispX * ratioA;
			posA.y += dispY * ratioA;
		}
		if (!dotB.fixed) {
			posB.x -= dispX * ratioB;
			posB.y -= dispY * ratioB;
		}
	}

	public inline function lineLerp(ratio:Float):Vector2 {
		return dotA.pos.add(dotB.pos.sub(dotA.pos).mult(ratio));
	}
}
