package khm.verlet;

import khm.Types.Point;

typedef DotJson = {
	pos:Point,
	prevPos:Point,
	fixed:Bool,
	gravity:Point,
}

typedef ConstraintJson = {
	indexA:Int,
	indexB:Int,
	stiffness:Float,
	isEdge:Bool,
}

typedef BodyJson = {
	dots:Array<DotJson>,
	constraints:Array<ConstraintJson>,
	joints:Array<ConstraintJson>,
	fixed:Bool,
	layers:Int,
	ignoreLayers:Int,
	?state:{}
}

@:access(khm.verlet)
class VerletJson {
	public static function dot(dot:Dot):DotJson {
		return {
			pos: dot.pos,
			prevPos: dot.prevPos,
			fixed: dot.fixed,
			gravity: dot.gravity
		};
	}

	public static function constraint(c:Constraint):ConstraintJson {
		return {
			indexA: c.body.dots.indexOf(c.dotA),
			indexB: c.body.dots.indexOf(c.dotB),
			stiffness: c.stiffness,
			isEdge: c.isEdge
		};
	}

	public static function body(body:Body):BodyJson {
		return {
			dots: body.dots.map(d -> dot(d)),
			constraints: body.constraints.map(c -> constraint(c)),
			joints: body.joints.map(c -> constraint(c)),
			fixed: body.fixed,
			layers: body.layers.toInt(),
			ignoreLayers: body.ignoreLayers.toInt()
		};
	}
}
