package khm.verlet.editor;

class Line extends Scene {
	final editor:VerletEditor;
	var createdDot:Dot = null;
	var mergeToDot:Dot = null;

	public function new(editor:VerletEditor) {
		this.editor = editor;
		this.name = 'Line';
		super();
	}

	override function onShow():Void {
		on(PointerDown, e -> {
			e.propagate = true;
			final p = e.p.toLocal(editor.scene);
			switch p.button {
				case Left:
					editor.updateClosestDot(p);
					createdDot = createDot(p);

					mergeToDot = editor.updateClosestDot(p, 20, [createdDot]);
					if (mergeToDot != null) {
						createdDot.body.mergeDot(createdDot, mergeToDot);
						editor.selectedDots.resize(0);
						mergeToDot = null;
						createdDot = null;
					}
				case Right:
					editor.updateClosestDot(p);
					if (editor.closestDot != null) {
						editor.toggleDotSelection(editor.closestDot);
					}
				case _:
			}
		});
		on(PointerMove, e -> {
			e.propagate = true;
			final p = e.p.toLocal(editor.scene);
			if (!p.isDown) {
				editor.updateClosestDot(p);
				return;
			}
			if (createdDot == null) return;
			createdDot.movePoint({
				x: p.x,
				y: p.y,
			});
			mergeToDot = editor.updateClosestDot(p, 20, [createdDot]);
		});
		on(PointerUp, e -> {
			e.propagate = true;
			editor.selectedDots.resize(0);
			if (createdDot == null) return;
			if (mergeToDot != null) {
				createdDot.body.mergeDot(createdDot, mergeToDot);
				createdDot = mergeToDot;
				mergeToDot = null;
				return;
			}

			// editor.triangulateClosedBody(createdDot.body);

			editor.selectedDots.push(createdDot);
			createdDot = null;
		});
	}

	override function onHide() {
		off();
		createdDot = null;
		mergeToDot = null;
	}

	function createDot(p:Pointer):Dot {
		final selectedDots = editor.selectedDots;
		final first = selectedDots[0];
		final body = first?.body ?? {
			trace("new body");
			final b = editor.world.addBody({
				dots: [],
				constraints: [],
			});
			b.fixed = true;
			trace(b);
			b;
		};
		final isFixed = first?.fixed ?? true;
		final dot = body.addDot(p.x, p.y, isFixed);
		for (selectedDot in selectedDots) {
			body.addConstraint(dot, selectedDot, 1, true);
		}
		body.updateBoundingBox();
		selectedDots.resize(0);
		return dot;
	}
}
