package khm.verlet.editor;

class RectObj extends Scene {
	final editor:VerletEditor;
	final dragDots:Array<Dot> = [];

	public function new(editor:VerletEditor) {
		this.editor = editor;
		this.name = 'RectObj';
		super();
	}

	override function onShow():Void {
		on(PointerDown, e -> {
			e.propagate = true;
			final p = e.p.toLocal(editor.scene);
			dragDots.resize(0);
			final dragDot = editor.updateClosestDot(p);
			if (dragDot == null) { // try to select full body
				final closestBody = editor.world.bodies.findMin(body -> {
					return body.center.sub({x: p.x, y: p.y}).length;
				}, 300);
				if (closestBody == null) return;
				for (dot in closestBody.dots) {
					editor.toggleDotSelection(dot);
				}
			} else { // select closest dots
				dragDots.push(dragDot);
				if (!App.keys.get(Shift)) {
					for (dot in editor.getClosestDots(dragDot)) dragDots.push(dot);
				}
			}
		});
		on(PointerMove, e -> {
			e.propagate = true;
			final p = e.p.toLocal(editor.scene);
			if (dragDots.length == 0) {
				editor.updateClosestDot(p);
			}
			drag(p);
		});
		on(PointerUp, e -> {
			e.propagate = true;
			final p = e.p;
			final firstDragDot = dragDots[0];
			// click on dot for selection
			if (!p.isMoved && firstDragDot != null) {
				final isSel = editor.toggleDotSelection(firstDragDot);
				// if (isSel) editor.setTool(LineTool);
			}
			// remove selection from dots after move
			if (p.isMoved && editor.selectedDots.length > 0) {
				editor.selectedDots.resize(0);
			}
			dragDots.resize(0);
		});
		on(Update, e -> {
			if (dragDots.length == 0) return;
			final p = App.pointers[0].toLocal(editor.scene);
			drag(p);
		});
	}

	override function onHide() {
		off();
		dragDots.resize(0);
	}

	function drag(p:Pointer):Void {
		if (!p.isDown) return;
		if (!p.isMoved) return;

		if (editor.selectedDots.length > 0) {
			moveSelectedDots(p);
			return;
		}
		for (dot in dragDots) {
			final forceDrag = dot.body.fixed ? 1 : 20;
			final s = dot.body.mass.clamp(0.1, 100) * forceDrag;
			dot.movePoint({
				x: dot.pos.x + (p.x - dot.pos.x) / s,
				y: dot.pos.y + (p.y - dot.pos.y) / s,
			});
		}
	}

	function moveSelectedDots(p:Pointer):Void {
		for (dot in editor.selectedDots) {
			// final forceDrag = dot.body.fixed ? 1 : 20;
			// final s = dot.body.mass.clamp(0.1, 100) * forceDrag;
			dot.movePoint({
				x: dot.pos.x + p.moveX,
				y: dot.pos.y + p.moveY,
			});
			dot.speedX = 0;
			dot.speedY = 0;
		}
	}
}
