package;

import haxe.Json;
import haxe.io.Path;
import js.Node.console;
import js.Node;
import js.lib.Promise;
import js.node.Buffer;
import js.node.ChildProcess;
import sys.FileSystem;
import sys.io.File;

using Lambda;
using StringTools;

private typedef AudiospriteConfig = {
	?output:String,
	?path:String,
	?export:String,
	?format:String,
	?autoplay:String,
	?loop:Array<String>,
	?silence:Int,
	?gap:Int,
	?minlength:Int,
	?bitrate:Int,
	?vbr:Int,
	?samplerate:Int,
	?channels:Int,
	?rawparts:String,
	?ignorerounding:Int,
	?logger:{debug:() -> Void, info:() -> Void, log:() -> Void}
}

extern class AudispriteModule {
	@:selfCall
	function audiosprite(
		files:Array<String>,
		?opts:AudiospriteConfig,
		?callback:(err:{}, obj:{}) -> Void
	):Void;
}

class Auditor {
	static function main() {
		final audiospriteModule:AudispriteModule = try {
			final buffer:Buffer = ChildProcess.execSync("npm root --location=global");
			final path = buffer.toString().trim();
			Node.require(path + "/audiosprite/audiosprite.js");
		} catch (err) {
			console.log("Install audiosprite globally first: check out README.md");
			Node.process.exit(1);
			return;
		}
		var projectPath = haxe.macro.Compiler.getDefine("projectPath");
		final assetsPath = '$projectPath/assets/_audio';
		if (!FileSystem.exists(assetsPath)) {
			console.log('Folder ${assetsPath.replace(projectPath, "")} not found');
			return;
		}

		final audiosprite = audiospriteModule.audiosprite;
		final formats = ["mp3", "ogg", "wav", "ac3", "m4a", "mp4"];

		final files = FileSystem.readDirectory(assetsPath).filter(name -> {
			if (name.startsWith(".")) return false;
			if (FileSystem.isDirectory('$assetsPath/$name')) return false;
			final ext = Path.extension(name);
			return formats.contains(ext);
		}).map(name -> '$assetsPath/$name');

		final names:Array<String> = files.map(path -> getFileName(path));
		final loops:Array<String> = [];

		console.log("Tracks names: " + names.join(", "));

		for (name in names) {
			if (name.startsWith("music_") || name == "music") loops.push(name);
		}
		if (loops.length == 0) console.log("No looped tracks (files with 'music_' prefix)")
		else console.log("Looped tracks: " + loops.join(", "));

		audiosprite(files, {
			output: '$projectPath/res/audio',
			loop: loops,
			export: "mp3",
			path: '$projectPath/res',
			bitrate: 64
		}, (err, json) -> {
			if (err != null) {
				console.log("Error: " + err);
				return;
			}
			File.saveContent('$projectPath/res/audio.json', Json.stringify(json));
			final stat = FileSystem.stat('$projectPath/res/audio.mp3');
			final kb = Std.int(stat.size / 1024);
			console.log('\nres/audio.mp3 updated ($kb KiB)');
			console.log("\n");
		});
	}

	static function getFileName(path:String):String {
		final noExtPath = Path.withoutExtension(path);
		final name = Path.withoutDirectory(noExtPath);
		return name;
	}
}
