package;

import haxe.Json;
import haxe.crypto.Base64;
import haxe.crypto.Md5;
// import haxe.io.Bytes;
import haxe.io.Bytes;
import haxe.io.Path;
import js.Node.console;
import js.Node;
import js.lib.Promise;
import js.lib.Uint8Array;
import js.node.Buffer;
import js.node.ChildProcess;
import js.node.Fs.Fs;
import sys.FileSystem;
import sys.io.File;

using Lambda;
using StringTools;

extern class Ttf2Woff {
	@:selfCall
	function ttf2woff(input:Uint8Array, ?options:{metadata:Uint8Array}):Uint8Array;
}

class Fontier {
	static function main() {
		new Fontier();
	}

	final assetsPath:String;
	final ttfModule:Ttf2Woff;

	public function new() {
		final projectPath = haxe.macro.Compiler.getDefine("projectPath");
		assetsPath = '$projectPath/assets/_fonts';
		ttfModule = try {
			final buffer:Buffer = ChildProcess.execSync("npm root --location=global");
			final path = buffer.toString().trim();
			Node.require(path + "/ttf2woff");
		} catch (err) {
			console.log("Install ttf2woff globally first: check out README.md");
			Node.process.exit(1);
			return;
		}
		if (!FileSystem.exists(assetsPath)) {
			trace('Folder ${assetsPath.replace(projectPath, "")} not found');
			return;
		}
		var fonts = FileSystem.readDirectory(assetsPath);
		fonts = fonts.filter(name -> {
			if (name.startsWith(".")) return false;
			return name.endsWith(".ttf");
		});

		insertFontsToHtml(fonts, '$projectPath/build/index.html');
		var khaPath = '$projectPath/res/kha/index.html';
		if (!FileSystem.exists(khaPath)) khaPath = '$projectPath/res/index.html';
		insertFontsToHtml(fonts, khaPath);
	}

	function insertFontsToHtml(fonts:Array<String>, path:String):Void {
		if (!FileSystem.exists(path)) {
			trace('$path not found');
			return;
		}
		var index = File.getContent(path);
		// File.saveContent('$dir/index_prev.html', index);

		final matchFonts = ~/\n^[\t ]*@font-face {\n.+\n.+\n[\t ]*}[\t ]*/gm;
		final matchFontDivs = ~/\n^[\t ]*<div style="font-family.*<\/div>[\t ]*/gm;
		final matchFontName = ~/font-family: ?['"](.+)['"]/g;
		index = matchFonts.map(index, reg -> {
			final block = reg.matched(0);
			if (matchFontName.match(block)) {
				haxe.Log.trace('Font removed: ${matchFontName.matched(1)}', null);
			}
			return "";
		});
		index = matchFontDivs.replace(index, "");
		haxe.Log.trace('', null);

		final lines = index.split("\n");

		final styleEndI = lines.findIndex(item -> item.contains("</style>"));

		for (i => font in fonts) {
			final name = Path.withoutExtension(font);
			final file = Fs.readFileSync('$assetsPath/$font');
			final out = ttfModule.ttf2woff(file);
			final base64 = Base64.encode(Bytes.ofData(out.buffer));
			// final base64 = base64.substr(0, 10);
			final block = '		@font-face {
			font-family: \'$name\';
			src: url(\'data:font/woff2;charset=utf-8;base64,$base64\');
		}';
			lines.insert(styleEndI + i, block);
		}
		final bodyStartI = lines.findIndex(item -> item.contains("<body>"));
		for (i => font in fonts) {
			final name = Path.withoutExtension(font);
			final div = '	<div style="font-family:\'$name\';visibility: hidden;position: relative;height: 0;">text</div>';
			lines.insert(bodyStartI + 1 + i, div);
			haxe.Log.trace('Font added: $name', null);
		}

		File.saveContent(path, lines.join("\n"));
		haxe.Log.trace('\n$path updated', null);
	}
}
