package;

import sys.FileSystem;
import sys.io.File;

using StringTools;

class Migrant {
	static function main() {
		var projectPath = haxe.macro.Compiler.getDefine("projectPath");
		var filePath = '$projectPath/src/Game.hx';
		if (!FileSystem.exists(filePath)) {
			filePath = '$projectPath/src/game/Game.hx';
		}
		if (FileSystem.exists(filePath)) {
			final file = File.getContent(filePath);
			final newFile = migrateLazyrGame(file);
			if (file == newFile) trace('No changes in $filePath');
			else {
				ensureDir('$projectPath/migrated');
				final backupPath = '$projectPath/migrated/Game.hx';
				backupFile(filePath, backupPath);
				File.saveContent(filePath, newFile);
				trace('$filePath updated');
			}
		}

		final filePath = '$projectPath/build-resources.hxml';
		if (FileSystem.exists(filePath)) {
			final file = File.getContent(filePath);
			final newFile = migrateResHxml(file);
			if (file == newFile) trace('No changes in $filePath');
			else {
				ensureDir('$projectPath/migrated');
				final backupPath = '$projectPath/migrated/build-resources.hxml';
				backupFile(filePath, backupPath);
				File.saveContent(filePath, newFile);
				trace('$filePath updated');
			}
		}
	}

	static function migrateLazyrGame(file:String):String {
		file = file.replace("import Res.res;", "import khm.Res;\nimport khm.Res.res;");
		file = file.replace("import Res.ResItem;", "import khm.ResItem;");
		file = file.replace("import lazyr.Res.ResItem;", "import khm.ResItem;");
		file = file.replace("import Res.", "import khm.Res.");
		file = file.replace("import lazyr.App.Pointer", "import khm.Pointer");

		file = file.replace("using Lambda;\n", "");
		file = file.replace("using StringTools;\n", "");
		file = file.replace("using lazyr.MathExtension;\n", "");
		file = file.replace("lazyr.PhaserTools", "khm.tools.PhaserTools");

		file = file.replace("onImagesLoaded: () -> {", "onImagesLoaded: config -> {");
		file = file.replace("Lazyr.onResize();", "// Lazyr.onResize();");
		final dragTo = ", e -> {
			final p = e.p;
			final x = e.dragX;
			final y = e.dragY;
		".rtrim();
		final pointerTo = ", e -> {
			final p = e.p;
			final x = e.localX;
			final y = e.localY;
		".rtrim();
		file = file.replace("Drag, (p, x, y) -> {", 'Drag $dragTo');
		file = file.replace("DragEnd, (p, x, y) -> {", 'DragEnd $dragTo');
		file = file.replace("PointerDown, (p, x, y) -> {", 'PointerDown $pointerTo');
		file = file.replace("PointerMove, (p, x, y) -> {", 'PointerMove $pointerTo');
		file = file.replace("PointerUp, (p, x, y) -> {", 'PointerUp $pointerTo');

		var lines = file.split("\n");

		var frameRemovals:Array<String> = [];
		var lineI = lines.length;
		while (lineI-- > 0) {
			final line = lines[lineI];
			if (!line.contains("addSprite({")) continue;
			final block = readBlock(lines, lineI, line.indexOf("{")).block;
			final objLines = block.split("\n").length;

			for (subI in lineI...lineI + objLines) {
				final subLine = lines[subI];
				if (!subLine.contains("frame: {")) continue;
				final block = readBlock(lines, subI, subLine.indexOf("{")).block;
				lines[lineI] = lines[lineI].replace("addSprite({", 'addSprite($block, {');
				final tabs = subLine.substr(0, subLine.indexOf("frame"));
				frameRemovals.push('${tabs}frame: $block,\n');
				break;
			}
		}
		file = lines.join("\n");
		for (rm in frameRemovals) file = file.replace(rm, "");

		final addSpriteReg = ~/addSprite\({\n[\t]*container: ([^ ,\n]+),\n[\t]*frame: ([^ ,\n]+),?\n/g;
		file = addSpriteReg.replace(file, "$1.addSprite($2, {\n");
		final addSpriteReg = ~/addSprite\({\n[\t]*frame: ([^ ,\n]+),?\n([\t]*onResize:)/g;
		file = addSpriteReg.replace(file, "addSprite($1, {\n$2");
		file = ~/, {[ \t\n]*}\);/g.replace(file, ");");

		file = file.replace("phaser.sound.BaseSound", "khm.Sound");
		file = file.replace("lazyr.", "khm.");

		file = file.replace("add(new DebugRectOverlay(this));", "new DebugRectOverlay(this).show();");
		file = file.replace("screenW", "sceneW");
		file = file.replace("screenH", "sceneH");
		if (file.contains("Utils.loadBase64Resources")) {
			file = addImportAtStart(file, "khm.utils.Base64Loader");
			file = file.replace("Utils.loadBase64Resources", "Base64Loader.loadBase64Resources");
		}
		if (file.contains("Audio.") && !file.contains("Audio;")) {
			file = addImportAtStart(file, "khm.AudioRes as Audio");
		}
		if (file.contains("final tools = LazyrTools;")) {
			file = file.replace("final tools = LazyrTools;", "final tools:LazyrTools;");
			file = ~/super\(\);\n[\t ]*load\(\);/.replace(file, "super();\ntools = new LazyrTools(this);\nload();");
		}
		file = file.replace("khm.utils.Macro", "khm.utils.macro.Macro");

		return file;
	}

	static function addImportAtStart(file:String, importLine:String):String {
		importLine = 'import $importLine;';
		if (file.startsWith("package")) {
			final lines = file.split("\n");
			lines.insert(1, importLine);
			file = lines.join("\n");
		} else {
			file = '$importLine\n$file';
		}
		return file;
	}

	static function migrateResHxml(file:String):String {
		file = file.replace("-D audiores", "-D res_audio_json");
		file = file.replace("-D res_folder", "-D assets_folder");
		if (!file.contains("src_folder")) file += "-D src_folder=src\n";
		return file;
	}

	public static function backupFile(fromPath:String, toPath:String):Void {
		if (FileSystem.exists(toPath)) {
			FileSystem.deleteFile(toPath);
		}
		File.copy(fromPath, toPath);
	}

	public static function ensureDir(path:String):Void {
		if (!FileSystem.exists(path)) FileSystem.createDirectory(path);
	}

	public static function readBlock(lines:Array<String>, lineI:Int, charI = 0):{line:Int, charI:Int, block:String} {
		final body = new StringBuf();
		var section:BodySection = Code;
		var level = 0;
		// if (charI == 0) charI = lines[lineI].indexOf("{");
		while (lineI < lines.length) {
			final line = lines[lineI];
			// trace(lineI, "line:", line);
			var lineStart = 0;
			var lineEnd = line.length;
			charI = 0;
			while (charI < line.length) {
				final code = line.fastCodeAt(charI);
				switch (section) {
					case Code:
						switch (code) {
							case "{".code:
								if (level == 0) lineStart = charI;
								level++;
							// trace("{", lineI + 1, charI, level);

							case "}".code:
								level--;
								// trace(level);
								if (level == 0) {
									lineEnd = charI + 1;
									break;
								} else if (level < 0) {
									throw 'object ($lineI, $charI) closed before been opened';
								}

							case "/".code:
								final next = line.fastCodeAt(charI + 1);
								if (next == "/".code) section = SingleComment;
								else if (next == "*".code) section = MultiComment;
								if (section != Code) charI++;

							case "'".code:
								section = SingleQuotes;
							case '"'.code:
								section = DoubleQuotes;
							case "`".code:
								section = Backticks;
						}
					case SingleComment:
						if (charI == line.length - 1) section = Code;
					case MultiComment:
						if (code == "*".code) {
							final next = line.fastCodeAt(charI + 1);
							if (next == "/".code) {
								section = Code;
								charI++;
							}
						}
					case SingleQuotes:
						if (code == "\\".code) charI++;
						else if (code == "'".code) section = Code;
					case DoubleQuotes:
						if (code == "\\".code) charI++;
						else if (code == '"'.code) section = Code;
					case Backticks:
						if (code == "\\".code) charI++;
						else if (code == "`".code) section = Code;
				}
				charI++;
			}

			body.add(line.substring(lineStart, lineEnd));
			if (lineEnd == line.length) body.add("\n");

			if (level == 0 && section == Code) break;
			lineI++;
		}
		return {
			line: lineI,
			charI: charI,
			block: body.toString()
		};
	}
}

enum BodySection {
	Code;
	SingleComment;
	MultiComment;
	SingleQuotes;
	DoubleQuotes;
	Backticks;
}
