import game.Game;
import game.Loader;
import khm.App;
#if js
import js.Browser.document;
import js.Browser.window;
import js.html.CanvasElement;
#end
#if kha
import kha.System;
#end

class Main {
	static function main() {
		#if js
		(window : Dynamic).createGame = createGame;
		#else
		createGame();
		#end
	}

	static function createGame():Void {
		// final onStart = () -> new game.Game();
		final onStart = () -> {
			final loader = new Loader(loader -> {
				loader.hide();
				final game = new Game();
				game.show();
			});
			return loader;
		}
		#if kha
		createCanvas("canvas-game");
		setFullWindowCanvas();

		System.start({
			title: "Example",
			width: 600,
			height: 800,
			audio: {allowMobileWebAudio: true}
		}, window -> {
			App.init({
				onStart: onStart,
			});
		});
		#elseif phaser
		App.init({
			onStart: onStart,
		});

		initAdStuff();
		if (window.onresize == null) window.onresize = onWindowResize;
		if (window.onorientationchange == null) window.onorientationchange = onWindowResize;
		#end
	}

	static function createCanvas(parentId:String):Void {
		#if kha_html5
		final size = getSize();
		final parent = document.getElementById(parentId);
		final canvas = document.createCanvasElement();
		canvas.id = kha.Macros.canvasId();
		canvas.width = size.width;
		canvas.height = size.height;
		// canvas.width = Std.int(size.width * window.devicePixelRatio);
		// canvas.height = Std.int(size.height * window.devicePixelRatio);
		// canvas.style.width = size.width + "px";
		// canvas.style.height = size.height + "px";
		canvas.style.display = "block";
		canvas.tabIndex = -1;
		parent.appendChild(canvas);
		// canvas.style = "display: block; width: 100%; height: 100%;";
		#end
	}

	static function setFullWindowCanvas():Void {
		#if kha_html5
		document.documentElement.style.padding = "0";
		document.documentElement.style.margin = "0";
		document.body.style.padding = "0";
		document.body.style.margin = "0";
		final canvas:CanvasElement = cast document.getElementById(kha.Macros.canvasId());
		canvas.style.display = "block";
		final resize = function() {
			final size = inline App.getWindowSize();
			final newW = Std.int(size.width * window.devicePixelRatio);
			final newH = Std.int(size.height * window.devicePixelRatio);
			if (canvas.width != newW) canvas.width = newW;
			if (canvas.height != newH) canvas.height = newH;
			if (canvas.style.width == "") {
				canvas.style.width = "100%";
				canvas.style.height = "100%";
			}
		}
		window.onresize = resize;
		resize();
		#end
	}

	#if phaser
	static function initAdStuff():Void {
		final game = App.phaserGame;
		final w:Dynamic = window;
		w.game = game;
		w.game.pauseGame = function() {
			game.scene.pause("globalScene");
		}
		w.game.resumeGame = function() {
			game.scene.resume("globalScene");
		}
		w.game.sizeChanged = onWindowResize;
	}
	#end

	#if js
	public static function getSize():{width:Int, height:Int} {
		return (window : Dynamic).getSize();
	}
	#end

	#if js
	public static function onWindowResize():Void {
		#if kha_html5
		final canvas:CanvasElement = cast document.getElementById(kha.Macros.canvasId());
		#elseif phaser
		final game = App.phaserGame ?? return;
		if (!game.isBooted) return;
		final canvas = game.canvas;
		#end
		final size = getSize();
		final width = size.width * window.devicePixelRatio;
		final height = size.height * window.devicePixelRatio;
		#if debug
		if (canvas.style.width == "") {
			canvas.style.width = "100%";
			canvas.style.height = "100%";
		}
		#else
		canvas.style.width = size.width + "px";
		canvas.style.height = size.height + "px";
		#end
		#if kha
		canvas.width = Std.int(width);
		canvas.height = Std.int(height);
		#elseif phaser
		game.scale.resize(width, height);
		#end
	}
	#end
}
