#if js
import js.Browser.document;
import js.Browser.window;
#end

class Utils {
	#if js
	public static function initMobileDebugger():Void {
		window.onerror = (msg:String, url:String, line:Int, column:Int, error:Dynamic) -> {
			window.alert('$line:$column: $msg');
			return false;
		}
	}

	public static function initErudaConsole():Void {
		var script = document.createScriptElement();
		script.src = "https://cdn.jsdelivr.net/npm/eruda";
		document.body.querySelector("#canvas-game").append(script);
		script.onload = () -> (window : Dynamic).eruda.init();
	}
	#end

	public static function seededShuffle<T>(arr:Array<T>):Void {
		for (i => a in arr) {
			final range = seedRandom();
			final n = Std.int((arr.length - 1) * range);
			final b = arr[n];
			arr[i] = b;
			arr[n] = a;
		}
	}

	static var randomInstance:Random = null;
	public static var _randomSeed = 0;

	public static function seedRandom(?seed:Int):Float {
		randomInstance ??= new Random(_randomSeed);
		if (seed != null) {
			randomInstance.seed = seed;
		}
		return randomInstance.GetFloat();
	}

	public static function measure<T>(f:Void->T, ?pos:haxe.PosInfos):String {
		var t0 = haxe.Timer.stamp();
		f();
		final t = (haxe.Timer.stamp() - t0) + "s";
		return t;
	}
}

// Small Fast Counter 32 (sfc32) PRNG implementation
// sfc32 is public domain: http://pracrand.sourceforge.net/license.txt
// Implementation derived from: https://github.com/bryc/code/blob/master/jshash/PRNGs.md
// which is also public domain

/**
	Random number generator

	Please use this one instead of the native Haxe one to
	keep consistency between different platforms.
**/
private class Random {
	public var seed:Int;

	var a:Int;
	var b:Int;
	var c:Int;

	public function new(seed:Int):Void {
		this.seed = seed;
		a = 0x36aef51a;
		b = 0x21d4b3eb;
		c = 0xf2517abf;
		// Immediately skip a few possibly poor results the easy way
		for (i in 0...15) {
			this.Get();
		}
	}

	public function Get():Int {
		var t = (a + b | 0) + seed | 0;
		seed = seed + 1 | 0;
		a = b ^ b >>> 9;
		b = c + (c << 3) | 0;
		c = c << 21 | c >>> 11;
		c = c + t | 0;
		return t & 0x7fffffff;
	}

	public function GetFloat():Float {
		return Get() / 0x7fffffff;
	}

	public function GetUpTo(max:Int):Int {
		return Get() % (max + 1);
	}

	public function GetIn(min:Int, max:Int):Int {
		return Get() % (max + 1 - min) + min;
	}

	public function GetFloatIn(min:Float, max:Float):Float {
		return min + GetFloat() * (max - min);
	}

	public static var instance:Random;

	public static function init(seed:Int):Void {
		instance = new Random(seed);
	}

	public static function get():Int {
		return instance.Get();
	}

	public static function getFloat():Float {
		return instance.GetFloat();
	}

	public static function getUpTo(max:Int):Int {
		return instance.GetUpTo(max);
	}

	public static function getIn(min:Int, max:Int):Int {
		return instance.GetIn(min, max);
	}

	public static function getFloatIn(min:Float, max:Float):Float {
		return min + instance.GetFloat() * (max - min);
	}
}
