package game;

import game.Anim.FrameConfig;
import khm.Container;
import khm.Lazyr;
import khm.Res.res;
import khm.Res;
import khm.Sprite;
import khm.Tween;

class Cursor extends Container {
	public var inSelectionMode = false;

	final sprite:Sprite;
	final ui:GameUiScene;

	public final anim:Anim;

	public function new(ui:GameUiScene) {
		super();
		this.ui = ui;
		setSize(8, 8);
		sprite = addSprite(res.cursor);
		add(sprite);

		final frames:Array<FrameConfig> = [
			{set: "idle", delay: 3, loop: false},
			{set: "show"},
			{},
			{},
			{},
			{},
			{},
			{},
			{set: "idleActive", id: 7},
			{set: "hide", id: 7},
			{id: 6},
			{id: 5},
			{id: 4},
			{id: 3},
			{id: 2},
			{id: 1},
			{id: 0},
			{set: "click", id: 7, atEnd: "idleActive"},
			{id: 8},
			{id: 9},
			{id: 8},
			{set: "error", id: 9, atEnd: "idleActive"},
			{id: 10},
			{id: 11},
			{id: 10},
			{set: "playActive", loop: true, id: 12},
			{},
			{},
			{id: 13},
			{id: 12},
			{set: "play", id: 15},
			{},
			{},
			{id: 16},
			{id: 15},
			{
				set: "playHide",
				id: 18,
				atEnd: "idle",
				loop: false,
				// delay: 30
			},
			{id: 19},
			{id: 20},
			{id: 2},
			{id: 1},
			{id: 0},
		];
		anim = new Anim(frames);
		anim.setType("idle");
		setFrame();
	}

	function setFrame():Void {
		sprite.setCrop({
			x: anim.frameI * 16,
			y: 0,
			w: 16,
			h: 16
		});
	}

	public function enterSelectionMode() {
		inSelectionMode = true;
		anim.setType("show");
	}

	public function exitSelectionMode(withPlay:Bool) {
		inSelectionMode = false;
		if (withPlay) anim.setType("playHide");
		else anim.setType("hide");
	}

	public function selectAnim():Void {
		anim.setType("click");
	}

	public function errorAnim() {
		anim.setType("error");
	}

	public function update() {
		anim.playType();
		setFrame();
	}

	public function playAnim() {
		anim.setType("play");
	}
}
