package game;

import khm.Container;
import khm.IGameObject;
import khm.Lazyr;
import khm.MathLib;
import khm.Particles;
import khm.Res.res;
import khm.Res;
import khm.Sprite;
import khm.Types.IPoint;
import khm.Types.Point;
import khm.Types.Rect;
import khm.tilemap.Tile;
import khm.tilemap.Tilemap;
import khm.utils.Collision;

class Elevator extends Container {
	final game:Game;
	final tilemap:Tilemap;

	public final levelId:Int;

	public final platform:Platform;
	public final moveH:Int;
	public var isActive = false;
	public var isUsed = false;

	var lift_wheel_small:Sprite;
	var lift_wheel_big:Sprite;

	public var lift:Sprite;

	public function new(game:Game, tilemap:Tilemap, ix:Int, iy:Int, levelId:Int, isUsed:Bool, moveH = 40) {
		super();
		this.game = game;
		this.tilemap = tilemap;
		this.levelId = levelId;
		this.isUsed = isUsed;
		this.moveH = moveH;

		isPixelPerfect = true;
		final layer = tilemap.tileset.layers[1];
		for (i in 0...4) {
			if (isUsed) continue;
			final block = tilemap.getTile(1, ix + i, iy);
			block.id = TileType.ElevatorBlock;
			block.getState().elevator = this;
		}
		final movePoints:Array<IPoint> = [
			{
				x: ix,
				y: iy
			}, {
				x: ix,
				y: iy - moveH
			},
		];
		if (isUsed) movePoints.pop();
		platform = new Platform("liftPlatform", game, tilemap, {
			movePoints: movePoints,
			speed: 0.3,
			pathMode: OneDir,
			isUsedLift: isUsed,
			hasPathLine: false
		});
		if (!isUsed) platform.y += 1;
		platform.width *= 4;
		platform.visible = false;
		game.addPlatform(platform);

		final lift_bg_bottom = addSprite(res.lift_bg_bottom);
		final lift_bg_top = addSprite(res.lift_bg_bottom);
		lift_bg_top.flipY = true;
		lift_wheel_small = addSprite(res.lift_wheel_small);
		lift_wheel_big = addSprite(res.lift_wheel_big);
		lift = addSprite(res.lift);
		// lift.alpha = 0;
		shrink();
		lift_bg_bottom.x = -5;
		lift_bg_bottom.y = 42;
		lift_wheel_small.x = -30 - 5;
		lift_wheel_small.y = 37;
		lift_wheel_big.x = 24;
		lift_wheel_big.y = 30;
		lift_bg_bottom.visible = !isUsed;
		lift_wheel_big.visible = !isUsed;
		lift_wheel_small.visible = !isUsed;
		// lift_bg_bottom.isPixelPerfect = true;
		// lift_wheel_small.isPixelPerfect = true;
		// lift_wheel_big.isPixelPerfect = true;
		// lift.isPixelPerfect = true;
		x = (ix + 2) * tilemap.tileW + 3;
		y = (iy - 2) * tilemap.tileH + 2;
		final ropeTilesLength = moveH + 4;
		final dir = isUsed ? 1 : -1;
		for (i in 0...ropeTilesLength) {
			final ropeLeft = addSprite(res.lift_rope);
			ropeLeft.x = lift_bg_bottom.x - 25;
			ropeLeft.y = lift_bg_bottom.y - i * tilemap.tileH - 6;
			final ropeRight = addSprite(res.lift_rope);
			ropeRight.x = lift_bg_bottom.x + 29;
			ropeRight.y = lift_bg_bottom.y - i * tilemap.tileH - 6;
			if (isUsed) {
				ropeLeft.y = lift_bg_bottom.y + (i - 2) * tilemap.tileH - 6;
				ropeRight.y = lift_bg_bottom.y + (i - 2) * tilemap.tileH - 6;
			}
			addAt(0, ropeLeft);
			addAt(0, ropeRight);
			// if (isUsed)
		}
		lift_bg_top.x = lift_bg_bottom.x;
		lift_bg_top.y = lift_bg_bottom.y - (ropeTilesLength * tilemap.tileH) + 6;
		if (isUsed) lift_bg_top.y = lift_bg_bottom.y - (3 * tilemap.tileH) + 6;
	}

	public function activate():Void {
		if (isUsed || isActive) return;
		isActive = true;
		isUsed = true;
		lift_wheel_big.addTween({
			duration: 10000,
			angle: 360 * 10
		});
		lift_wheel_small.addTween({
			duration: 10000,
			angle: 360 * 10
		});

		delayed(500, () -> {
			platform.isActive = true;
			platform.onUpdate = () -> {
				this.updateTransformation();
				final pos = platform.toLocal(this);
				lift.y = pos.y - 30;
				lift.updateTransformation();
			}
		});
	}
}
