package game;

import kha.graphics2.Graphics;
import kha.input.Mouse;
import khm.App;
import khm.AudioRes;
import khm.Color;
import khm.Container;
import khm.Event;
import khm.IGameObject;
import khm.Lazyr.*;
import khm.Lazyr;
import khm.LazyrTools;
import khm.MathLib;
import khm.Res.Res;
import khm.Res.res;
import khm.ResItem;
import khm.Settings;
import khm.Sound;
import khm.Sprite;
import khm.Text;
import khm.Tween;
import khm.Types.IPoint;
import khm.Types.Point;
import khm.Types.Rect;
import khm.audio.effects.Filter;
import khm.math.Degrees;
import khm.tilemap.Tilemap;
import khm.tilemap.Tileset;
import khm.utils.Collision;
import khm.utils.DebugPanel;
import khm.utils.DebugRectOverlay;
import khm.utils.macro.Macro;

using khm.tools.GameObjectAnimTools;
using khm.tools.PhaserTools;

typedef Save = {
	version:Int,
	level:Int,
	checkpointI:Int,
}

class Game extends khm.Scene {
	final tools:LazyrTools;
	var clickSound:Null<String> = null;
	var onWin:() -> Void;
	var onLose:() -> Void;
	var isLose = false;
	var isWin = false;
	var progress = 0;
	var music:khm.Sound;
	final tilemaps:Array<Tilemap> = [];
	var tilemap:Tilemap;
	var player:Player;
	final ui:GameUiScene;
	var selector:Selector;
	var selCameraOffset:Point = {x: 0, y: 0};

	public var inSelectionMode(default, null) = false;
	public var timeRatio = 1.0;

	final checkpoints:Array<Checkpoint> = [];

	public final platforms:Array<Platform> = [];
	public final tileButtons:Array<TileButton> = [];
	public final elevators:Array<Elevator> = [];

	var save:Save;
	var tileset:Tileset;
	var ldtk:Ldtk;

	public function new():Void {
		super();
		tools = new LazyrTools(this);

		App.defaultScaleQuality = Low;

		ui = new GameUiScene();

		final defSave:Save = {
			version: 1,
			level: 0,
			checkpointI: -1
		}
		Settings.init(defSave);
		save = Settings.read();

		var tracks = [AudioRes.music_ney, AudioRes.music_cobra];
		var activeTrack = 0;
		var loadNextTrack = false;
		// music = addAudioSprite(AudioRes.music_end);
		music = addAudioSprite(AudioRes.music_ney);
		addTimeEvent({
			delay: 500,
			callback: () -> {
				if (loadNextTrack) return;
				if (music.isFinished) {
					activeTrack++;
					activeTrack = activeTrack.wrapAround(0, tracks.length - 1);
					music = addAudioSprite(tracks[activeTrack]);
				}
			}
		});
		// final music = new Sound(AudioRes.music_ney);
		// music.play();

		#if debug
		onDebugKey(One, () -> onWin());
		onDebugKey(Two, () -> onLose());
		onDebugKey(Q, () -> trace(player.obj));
		onDebugKey(Nine, () -> {
			App.debugBoundsMode++;
			App.debugBoundsMode = App.debugBoundsMode % 3;
		});
		onDebugKey(Eight, () -> App.debugInputOrder = !App.debugInputOrder);
		App.debugBoundsMode = keepDebugValue("debugBoundsMode", () -> App.debugBoundsMode);
		onDebugKey(L, () -> Event.traceLeakedListeners());
		// new DebugPanel(this).show();
		final debugOverlay = new DebugRectOverlay(this).show();
		DebugRectOverlay.enableRightButton = false;
		#end

		onWin = () -> {
			final end = new EndScreen().show();
			end.alpha = 0;
			end.showAnim(1000, 3500, () -> {
			});
			delayed(4000, () -> {
				this.addTween({
					duration: 1000,
					onUpdate: e -> {
						final ratio = e.ratio;
						music.volume = ratio.lerp(music.volume, 0);
					}
				});
				final music2 = addAudioSprite(AudioRes.music_end);
			});
		}

		camera.originX = 0.5;
		camera.originY = 0.5;
		isUnlimitedSize = true;

		App.bgColor = 0xFF8b93af;

		var l1:Array<TsProps> = [
			{frame: res.tileset},
			// {tx: 0, ty: 0}, // 1
		];
		l1 = l1.concat([
			for (i in 0...300) {}
		]);
		var l2:Array<TsProps> = [
			{tx: 0, ty: 0},
			{props: {collide: true}}, // 1
			{props: {collide: true}},
			{props: {collide: true}},
			{props: {collide: true}},
			{props: {collide: true}}, // 5
			{props: {collide: true}},
			{props: {collide: true}},
			{props: {collide: true}},
			{props: {collide: true}},
			{props: {collide: true}}, // 10
			{props: {collide: true}},
		];
		l2 = l2.concat([
			for (i in 0...300) {}
		]);

		tileset = new Tileset(16, 16, [
			l1,
			l2,
		]);

		tilemap = new Tilemap(tileset);
		// tilemap.alpha = 0.5;

		#if js
		untyped window.player = player;
		#end

		ldtk = new Ldtk();

		// todo debug
		// save.level = 0;
		// save.checkpointI = 3;
		// save.level = 0;
		// save.checkpointI = -1;

		if (save.level == 2) {
			save.level = 0;
			save.checkpointI = -1;
		}
		Settings.set(save);
		trace(save);

		loadMap(save.level);

		selector = new Selector(this, tilemap);

		player = new Player(this, tilemap);
		spawnPlayer(save.level);

		on(BeforeRender, e -> {
			final p = App.pointers[0] ?? return;
			final p = p.toLocal(ui);
			ui.cursor.x = p.x;
			ui.cursor.y = p.y;
			ui.cursor.scale = tilemap.scale;
		});

		on(KeyDown, e -> {
			switch e.key {
				case R: restart();
				case M: music.volume = music.volume == 1 ? 0 : 1;
				case Slash:
					player.isCanFly = !player.isCanFly;
				case _:
			}
		});

		on(PointerDown, e -> {
			final mouse = Mouse.get(0);
			if (mouse != null) mouse.hideSystemCursor();
			final p = e.p;
			if (selector.isPlaying) return;
			if (e.p.button != Left) {
				exitSelection();
				return;
			}
			if (!ui.cursor.inSelectionMode) {
				enterSelection();
				// do not click in activation
				return;
			} else {
				ui.cursor.selectAnim();
			}
			final p = p.toLocal(tilemap);
			final rect:Rect = {
				x: p.x,
				y: p.y,
				w: 1,
				h: 1
			};
			final tiles = Col.getOvelappingTiles(rect, tilemap, 1);
			final tile = tiles[0] ?? {
				ui.cursor.errorAnim();
				return;
			}
			if (!tile.isActivatableType()) {
				ui.cursor.errorAnim();
				return;
			}
			final particles = tilemap.addParticles({
				frames: [Res.res.pixel_particle],
				emitCallback: particle -> {
					particle.tint = 0xFFFFFFFF;
					final ang:Degrees = Std.random(360);
					particle.x += ang.cos() * 5;
					particle.y += ang.sin() * 5;
				},
				speed: {min: 20, max: 5},
				lifespan: 500,
				alpha: {start: 1, end: 0},
			});
			particles.explode(5, p.x, p.y);
			selector.addTile(tile);
		});
	}

	final timeCallbacks:Array<Tween> = [];

	public function delayedTime(delay:Float, callback:() -> Void):Void {
		final t = delayed(delay, callback);
		timeCallbacks.push(t);
	}

	override function onShow() {
		ui.show();
	}

	function loadMap(id:Int):Tilemap {
		final lvl = ldtk.levels[id];
		final bgTiles = LdtkUtils.getTilesLayer(lvl.l_Bg.json);
		final tiles = LdtkUtils.getTilesLayer(lvl.l_Tiles.json);
		tilemap.loadJson({
			w: tiles.w,
			h: tiles.h,
			layers: [
				bgTiles.map,
				tiles.map
			],
			objects: []
		});
		remove(tilemap);
		add(tilemap);

		checkpoints.resize(0);
		for (obj in lvl.l_Entities.all_Checkpoint) {
			final checkpoint = new Checkpoint();
			checkpoint.x = obj.cx * tilemap.tileW + tilemap.tileW / 2;
			checkpoint.y = obj.cy * tilemap.tileH + tilemap.tileH / 2;
			checkpoints.push(checkpoint);
			tilemap.add(checkpoint);
		}

		platforms.resize(0);
		for (obj in lvl.l_Entities.all_Platform) {
			final movePoints:Array<IPoint> = obj.f_Point.map(p -> {x: p.cx, y: p.cy});
			movePoints.unshift({x: obj.cx, y: obj.cy});
			final platform = new Platform(obj.iid, this, tilemap, {
				sizeW: Std.int(obj.width / tilemap.tileW),
				sizeH: Std.int(obj.height / tilemap.tileH),
				movePoints: movePoints,
				speed: obj.f_speed,
				pathMode: obj.f_isOneDir ? OneDir : YoyoDir,
				hasPathLine: obj.f_hasPathLine,
			});
			platforms.push(platform);
			tilemap.add(platform);
		}

		tileButtons.resize(0);
		for (obj in lvl.l_Entities.all_Button) {
			final btnPlatforms:Array<Platform> = [];
			for (ref in obj.f_Entity_ref) {
				final platform = platforms.find(platform -> platform.iid == ref.entityIid);
				btnPlatforms.push(platform);
			}
			// final refs = obj.f_Entity_ref.entityIid;
			// final platform = platforms.find(platform -> platform.iid == ref);
			final tileButton = new TileButton(this, tilemap, obj.cx, obj.cy, btnPlatforms);
			tileButtons.push(tileButton);
			tilemap.add(tileButton);
		}

		elevators.resize(0);
		for (obj in lvl.l_Entities.all_Elevator) {
			final elevator = new Elevator(this, tilemap, obj.cx, obj.cy, obj.f_LevelId, obj.f_used);
			elevators.push(elevator);
			tilemap.addAt(2, elevator);
		}

		Utils._randomSeed = 100;
		tilemap.everyTile(0, tile -> {
			final state = tile.getState();
			final ix = tile.ix;
			final iy = tile.iy;
			final v = Math.round(Utils.seedRandom() * 100);
			if (tile.id == 0 || tile.id == 20) {
				switch v {
					case 1: tilemap.setTileId(0, ix, iy, 21);
					case 2: tilemap.setTileId(0, ix, iy, 22);
					case _:
						final v = Math.round(Utils.seedRandom() * 400);
						switch v {
							case 3: tilemap.setTileId(0, ix, iy, 23);
							case 4: tilemap.setTileId(0, ix, iy, 30);
							case 5: tilemap.setTileId(0, ix, iy, 31);
							case 6: tilemap.setTileId(0, ix, iy, 37);
							case _:
						}
				}
			}
		});

		tilemap.everyTile(1, tile -> {
			final state = tile.getState();
			final type = tile.getType();
			switch type {
				case MovingBlockInactive:
					new MovingBlock(this, tilemap, tile.ix, tile.iy, Inactive);
				case MovingBlock:
					new MovingBlock(this, tilemap, tile.ix, tile.iy, Active);
				case JumpPadInactive:
					new JumpPad(this, tilemap, tile.ix, tile.iy, Inactive);
				case JumpPadActive:
					new JumpPad(this, tilemap, tile.ix, tile.iy, Active);
				case PlatformCallerInactive:
				case PlatformCallerActive:
				case ElevatorBlock:
					tile.obj.visible = false;
			}
		});

		return tilemap;
	}

	public function restart() {
		loadMap(save.level);
		spawnPlayer(save.level);
		player.reset();
		ui.flashScreen();
		selector.reset();
		// ui.cursor.exitSelectionMode(false);
		exitSelection();
	}

	var debugObj:Sprite = null;

	public function activateElevator(elevator:Elevator) {
		if (elevator.isActive || elevator.isUsed) return;
		player.lock();
		elevator.activate();

		save.checkpointI = -1;
		save.level = elevator.levelId;
		Settings.set(save);

		if (save.level == 2) {
			onWin();
			return;
		}

		delayed(3000, () -> {
			final prevElevator = elevator;
			loadMap(save.level);
			final elevatorI = elevators.findIndex(e -> e.isUsed);
			final elevator = elevators[elevatorI];
			elevator.visible = false;
			final f = platforms.remove(elevator.platform);

			final platform = prevElevator.platform;
			addPlatform(platform);
			elevators[elevatorI] = prevElevator;
			tilemap.add(prevElevator);
			tilemap.add(player.obj);

			final iy = tilemap.h + prevElevator.moveH;
			platform.movePoints[0].y = iy;
			platform.movePoints[1].y = tilemap.h - 2;
			final sy = platform.speed.y;
			final oldPlatformY = platform.y;
			platform.update([]);
			final addY = platform.y - oldPlatformY;
			player.obj.y += addY;
			prevElevator.y += addY;
			platform.onUpdate();

			final offX = lastCameraOffset.x;
			final offY = lastCameraOffset.y;
			moveCameraOnPlayer(0);
			camera.y -= offY + sy * 5;
			camera.x -= offX;

			platform.onComplete = () -> {
				player.unlock();
			}
		});
	}

	public function moveCameraOnPlayer(delay = 10):Void {
		final pos = player.obj.toLocal(this);
		moveCameraOn({
			x: pos.x + player.speed.x * 3,
			y: pos.y + player.speed.y * 3,
		}, delay);
	}

	override function onUpdate() {
		final min = Math.min(width, height);
		final max = Math.max(width, height);
		final minScale = isPortrait ? max / 250 : min / 250;
		tilemap.scale = Std.int(minScale).clamp(2, 10);
		// tilemap.scale = 1;

		final isUp = App.keys.get(W) || App.keys.get(Up) || App.keys.get(Space);
		final isDown = App.keys.get(S) || App.keys.get(Down);
		final isLeft = App.keys.get(A) || App.keys.get(Left);
		final isRight = App.keys.get(D) || App.keys.get(Right);
		player.isMoving = isLeft || isRight;
		if (inSelectionMode) player.isMoving = false;

		if (inSelectionMode) {
			final off = selCameraOffset;
			if (isUp) off.y -= 8;
			if (isDown) off.y += 8;
			if (isLeft) off.x -= 8;
			if (isRight) off.x += 8;
			off.x = off.x.clamp(-300, 300);
			off.y = off.y.clamp(-300, 300);
			final pos = player.obj.toLocal(this);
			moveCameraOn({
				x: pos.x + player.speed.x * 3 + off.x,
				y: pos.y + player.speed.y * 3 + off.y,
			}, 10);
		} else {
			moveCameraOnPlayer();
		}

		if (!inSelectionMode) {
			if ((player.onGround || player.isCanFly) && isUp) {
				player.jump();
			}
			if (isLeft && isRight) player.isMoving = false;
			if (isLeft) {
				player.speed.x -= player.accelerationSpeed;
				player.limitRunSpeed();
			}
			if (isRight) {
				player.speed.x += player.accelerationSpeed;
				player.limitRunSpeed();
			}
		}

		for (platform in platforms) {
			platform.update([]);
		}

		player.update();
		ui.cursor.update();
		selector.update();

		final playerRect = player.obj.getGlobalPosition();

		for (i => checkpoint in checkpoints) {
			if (i == save.checkpointI) continue;
			final r = checkpoint.getGlobalPosition();
			if (!Collision.aabb(playerRect, r)) continue;
			checkpoint.activate();
			save.checkpointI = i;
			Settings.set(save);
		}
	}

	function spawnPlayer(levelId:Int):Void {
		final lvl = ldtk.levels[levelId];
		var start = lvl.l_Entities.all_Start[0];
		final tileW = tilemap.tileW;
		final tileH = tilemap.tileH;
		var startPos:Point = {
			x: start.cx * tileW + tileW / 2,
			y: start.cy * tileH + tileH / 2,
		}
		if (save.checkpointI != -1) {
			final checkpoint = checkpoints[save.checkpointI];
			if (checkpoint != null) startPos = {
				x: checkpoint.x,
				y: checkpoint.y,
			}
			else {
				trace("Checkpoint not found", save.checkpointI);
				save.checkpointI = -1;
				Settings.set(save);
			}
		}
		tilemap.addAt(3, player.obj);
		player.obj.x = startPos.x;
		player.obj.y = startPos.y;
		// player.obj.x = 100;
		// player.obj.y = 100;
	}

	var lastCameraOffset:Point = {x: 0, y: 0};

	function moveCameraOn(pos:Point, delay = 10) {
		delay = delay.clamp(1, 100);
		var newX = pos.x;
		var newY = pos.y;
		camera.x += (newX - camera.x) / delay;
		camera.y += (newY - camera.y) / delay;
		camera.x = Math.round(camera.x);
		camera.y = Math.round(camera.y);
		lastCameraOffset.x = newX - camera.x;
		lastCameraOffset.y = newY - camera.y;
	}

	var audioFilter:Filter = null;
	var audioTween:Tween = null;

	function addAudioEffect():Void {
		if (audioFilter != null) return;
		audioFilter = new Filter(LOWPASS, 50000, -50000);
		final filter = music.addEffect(audioFilter);
		audioTween?.stop();
		audioTween = this.addTween({
			duration: 500,
			onUpdate: e -> {
				final ratio = e.ratio;
				music.rate = ratio.lerp(1, 0.5);
				filter.cutoff = ratio.lerp(50000, 500);
				filter.resonance = ratio.lerp(-50000, -500);
			}
		});
	}

	function removeAudioEffect():Void {
		if (audioFilter == null) return;
		final filter = audioFilter;
		audioFilter = null;
		audioTween?.stop();
		audioTween = this.addTween({
			duration: 0,
			onUpdate: e -> {
				final ratio = 1 - e.ratio;
				music.rate = ratio.lerp(1, 0.5);
				filter.cutoff = ratio.lerp(50000, 500);
				filter.resonance = ratio.lerp(-50000, -500);
			},
			onComplete: () -> {
				music.removeAllEffects();
			}
		});
	}

	function enterSelection() {
		inSelectionMode = true;
		addAudioEffect();
		timeRatio = 0.1;

		for (tween in timeCallbacks) {
			tween.time /= timeRatio;
			tween.duration /= timeRatio;
		}

		final keys:Array<khm.KeyCode> = [W, A, S, D, Left, Right, Up, Down, Space];
		for (code in keys) App.keys.remove(code);
		selCameraOffset.x = 0;
		selCameraOffset.y = 0;
		ui.cursor.enterSelectionMode();
		ui.selectionOutline.removeTweens();
		ui.selectionOutline.opacityAnim(0.2, 500);
	}

	public function exitSelection():Void {
		if (!ui.cursor.inSelectionMode) return;
		removeAudioEffect();
		inSelectionMode = false;
		for (tween in timeCallbacks) {
			tween.time *= timeRatio;
			tween.duration *= timeRatio;
		}
		timeRatio = 1;
		ui.selectionOutline.removeTweens();
		ui.selectionOutline.hideAnim(300);

		if (selector.timeline.length == 0) {
			ui.cursor.exitSelectionMode(false);
		} else {
			ui.cursor.playAnim();
			selector.playTimeline(() -> {
				// for timeout anim transition
				final withPlay = ui.cursor.anim.frameType == "play";
				ui.cursor.exitSelectionMode(withPlay);
			});
		}
	}

	public function exitSelectionWithError():Void {
		if (!ui.cursor.inSelectionMode) return;
		exitSelection();
		ui.cursor.errorAnim();
	}

	public function addPlatform(platform:Platform):Void {
		platforms.push(platform);
		tilemap.add(platform);
	}

	// override function render(g:Graphics) {
	// 	super.render(g);
	// 	#if debug
	// 	if (debugObj != null) {
	// 		trace(debugObj.y);
	// 		g.font = App.debugFont;
	// 		g.drawString(Std.int(debugObj.y) + "", 0, 0);
	// 	}
	// 	#end
	// }
}
