package game;

import game.TileState.TileMoveState;
import khm.Container;
import khm.IGameObject;
import khm.Lazyr;
import khm.MathLib;
import khm.Particles;
import khm.Res.res;
import khm.Res;
import khm.Sprite;
import khm.Types.IPoint;
import khm.Types.Point;
import khm.Types.Rect;
import khm.tilemap.Tile;
import khm.tilemap.Tilemap;
import khm.utils.Collision;
import khm.utils.DebugPanel;

class JumpPad extends Container {
	final game:Game;
	final tilemap:Tilemap;
	final tile:Tile;

	public final initialState:TileMoveState;
	public var moveState:TileMoveState;

	public function new(game:Game, tilemap:Tilemap, x:Int, y:Int, initialState:TileMoveState) {
		super();
		this.game = game;
		this.tilemap = tilemap;
		setSize(16, 16);
		final layer = tilemap.tileset.layers[1];
		tile = tilemap.getTile(1, x, y);
		tile.id = TileType.JumpPadInactive;
		tile.getState().jumpPad = this;
		this.initialState = initialState;
		moveState = initialState;

		if (initialState == Active) activate(true);
	}

	final toggleDuration = 500.0;

	public function activate(instant:Bool):Float {
		final tileC = tile.toContainer();
		DebugPanel.ignoreObject(tileC);
		final state = tile.getState();
		final layer = tilemap.tileset.layers[1];
		final activeSprite = Sprite.fromTile(res.tileset, layer[7].rect);
		tileC.add(activeSprite);
		activeSprite.alpha = 0;
		activeSprite.showAnim(instant ? 0 : 300);

		moveState = Activation;
		tile.setIdOnly(TileType.JumpPadActive);
		tileC.delayed(toggleDuration, () -> {
			moveState = Active;
		});
		return toggleDuration;
	}

	public function deactivate(instant:Bool):Float {
		final tileC = tile.toContainer();
		DebugPanel.ignoreObject(tileC);
		final state = tile.getState();
		final activeSprite:Sprite = cast tileC.last;
		activeSprite.hideAnim(instant ? 0 : toggleDuration);

		moveState = Deativation;
		tile.setIdOnly(TileType.JumpPadInactive);
		tileC.delayed(toggleDuration, () -> {
			moveState = Inactive;
		});
		return toggleDuration;
	}
}
