package game;

import game.TileState.TileMoveState;
import khm.Container;
import khm.IGameObject;
import khm.Lazyr;
import khm.MathLib;
import khm.Particles;
import khm.Res.res;
import khm.Res;
import khm.Sprite;
import khm.Types.IPoint;
import khm.Types.Point;
import khm.Types.Rect;
import khm.tilemap.Tile;
import khm.tilemap.Tilemap;
import khm.utils.Collision;
import khm.utils.DebugPanel;

class MovingBlock extends Container {
	final game:Game;
	final tilemap:Tilemap;
	final tile:Tile;

	public final initialState:TileMoveState;

	public var moveState:TileMoveState;

	var tileShadowC:Container = null;

	public function new(game:Game, tilemap:Tilemap, x:Int, y:Int, initialState:TileMoveState) {
		super();
		this.game = game;
		this.tilemap = tilemap;
		setSize(16, 16);
		final layer = tilemap.tileset.layers[1];
		tile = tilemap.getTile(1, x, y);
		tile.id = TileType.MovingBlock;
		tile.getState().movingBlock = this;
		this.initialState = initialState;

		if (initialState == Inactive) deactivate(true);
		moveState = initialState;
	}

	final toggleDuration = 500.0;

	public function activate(instant:Bool):Float {
		final state = tile.getState();
		if (tileShadowC == null) throw "no shadow inited";
		// final tileShadowC = movedShadow ?? return 0;
		final topShadow:Sprite = tileShadowC.state.topShadow;

		tile.obj.opacityAnim(1, toggleDuration);

		topShadow.addTween({
			duration: toggleDuration,
			ease: SineIn,
			displayHeight: 0,
		});
		tileShadowC.hideAnim(toggleDuration);

		moveState = Activation;
		tileShadowC.delayed(toggleDuration, () -> {
			moveState = Active;
		});
		return toggleDuration;
	}

	public function deactivate(instant:Bool):Float {
		final state = tile.getState();
		tileShadowC ??= createTileShadowContainer(tile);
		DebugPanel.ignoreObject(tileShadowC);

		switch moveState {
			case Active, Activation:
			case Deativation, Inactive:
				return 0;
		}

		tileShadowC.alpha = 0;
		final duration = instant ? 0 : toggleDuration;
		tile.obj.opacityAnim(0.5, duration);
		tileShadowC.showAnim(duration);
		final topShadow:Sprite = cast tileShadowC.children.find(obj ->
			obj.state.name == "topShadow");
		final ix = tile.ix;
		final iy = tile.iy;
		final topTile = tilemap.getTile(1, ix, iy - 1);
		var hasShadow = !topTile.isActivatableType();
		if (topTile.getState()?.movingBlock?.initialState == Active) {
			hasShadow = true;
		}

		final dh = hasShadow ? 7 : 0;
		topShadow.addTween({
			duration: duration,
			ease: SineOut,
			displayHeight: dh,
		});
		final state = tile.getState();
		moveState = Deativation;
		tileShadowC.delayed(duration, () -> {
			moveState = Inactive;
		});
		return toggleDuration;
	}

	function createTileShadowContainer(tile:Tile):Container {
		final tileC = tile.toContainer();
		DebugPanel.ignoreObject(tileC);
		// final tileShadowC = tilemap.map.layers[1].addContainer();
		final tileShadowC = tileC.addContainer();
		final layer = tilemap.tileset.layers[1];
		final shadow = Sprite.fromTile(res.tileset, layer[39].rect);
		final topShadow = Sprite.fromTile(res.tileset, layer[39].rect);
		topShadow.state.name = "topShadow";
		topShadow.originY = 0;
		topShadow.y = tile.obj.topLeftLocal().y;
		topShadow.displayHeight = 0;
		final border = Sprite.fromTile(res.tileset, layer[40].rect);
		final border2 = Sprite.fromTile(res.tileset, layer[41].rect);
		tileShadowC.add(shadow);
		tileShadowC.shrink();
		tileShadowC.add(border);
		tileShadowC.add(border2);
		tileShadowC.add(topShadow);
		tileShadowC.state.topShadow = topShadow;
		return tileShadowC;
	}
}
