package game;

import game.TileState.TileMoveState;
import khm.Container;
import khm.IGameObject;
import khm.Lazyr;
import khm.MathLib;
import khm.Particles;
import khm.Res.res;
import khm.Res;
import khm.Sprite;
import khm.Types.IPoint;
import khm.Types.Point;
import khm.Types.Rect;
import khm.tilemap.Tile;
import khm.tilemap.Tilemap;
import khm.utils.Collision;
import khm.utils.DebugPanel;

class TileButton extends Container {
	final game:Game;
	final tilemap:Tilemap;
	final tile:Tile;
	final platforms:Array<Platform>;

	public final initialState:TileMoveState;
	public var moveState:TileMoveState;

	public function new(game:Game, tilemap:Tilemap, x:Int, y:Int, platforms:Array<Platform>) {
		super();
		this.game = game;
		this.tilemap = tilemap;
		this.platforms = platforms;
		setSize(16, 16);
		final layer = tilemap.tileset.layers[1];
		tile = tilemap.getTile(1, x, y);
		tile.id = TileType.PlatformCallerInactive;
		tile.getState().tileButton = this;
		initialState = Inactive;
		moveState = Inactive;
	}

	public function activate(instant:Bool):Float {
		final tileC = tile.toContainer();
		DebugPanel.ignoreObject(tileC);
		final state = tile.getState();
		final layer = tilemap.tileset.layers[1];
		final activeSprite = Sprite.fromTile(res.tileset, layer[TileType.PlatformCallerActive].rect);
		tileC.add(activeSprite);
		activeSprite.alpha = 0;
		activeSprite.showAnim(instant ? 0 : 300);

		var maxDuration = 0.0;
		for (platform in platforms) {
			platform.isActive = true;
			final toggleDuration = platform.getTotalDuration();
			if (maxDuration < toggleDuration) maxDuration = toggleDuration;
		}
		moveState = Activation;
		tile.setIdOnly(TileType.PlatformCallerActive);
		tileC.delayed(maxDuration, () -> {
			moveState = Active;
		});
		return maxDuration;
	}

	public function deactivate(instant:Bool):Float {
		final tileC = tile.toContainer();
		DebugPanel.ignoreObject(tileC);
		final state = tile.getState();
		final activeSprite:Sprite = cast tileC.last;
		final toggleDuration = 300;
		activeSprite.hideAnim(instant ? 0 : toggleDuration);

		// platform.isActive = false;
		moveState = Deativation;
		tile.setIdOnly(TileType.PlatformCallerInactive);
		tileC.delayed(toggleDuration, () -> {
			moveState = Inactive;
		});
		return toggleDuration;
	}
}
